package scala.annotation.effects

import annotation.{StaticAnnotation, TypeConstraint}

trait Effect extends StaticAnnotation with TypeConstraint{
 
  def lte(b: Effect): Boolean = {
	b.getClass.isAssignableFrom(this.getClass)
  }
  /*def lte(b: Effect, getClassNameByString: String => String): Boolean = {
    lte(b)
  }*/
  def lte(b: Effect, getClassNameByString: String => Class[_]): Boolean = {
    lte(b)
  }
  
}

trait EffectsWithGeneric extends Effect{
  val privilegeName: String = "@TODO"
  val tps: List[List[String]]
  def lteCovariant(aTp: Class[_], bTp: Class[_]): Boolean = {
    bTp.isAssignableFrom(aTp)
  }
  def lteContravariant(aTp: Class[_], bTp: Class[_]): Boolean = {
    aTp.isAssignableFrom(bTp)
  }
  def lteInvariant(aTp: Class[_], bTp: Class[_]): Boolean = {
    aTp.equals(bTp)
  }
  def lteCoContravariant(aTp: Class[_], bTp: Class[_]): Boolean = {
    aTp.isAssignableFrom(bTp) || bTp.isAssignableFrom(aTp)
  }

  def leqFunc(index: Int, a: Class[_], b: Class[_]): Boolean = {
    b.isAssignableFrom(a)
  }
  def lteOneTps[T](aTp: T, bTps: List[T], lteFunc: (T, T) => Boolean) = {
    bTps.exists(bTp => lteFunc(aTp, bTp))
  }

  def getClassesAt(index: Int)(getClassNameByString: String => String): List[Class[_]] = {
    tps.lift(index).map{ltp => ltp.map{tp => Class.forName(getClassNameByString(tp))}}.getOrElse(List())
  }
  def getClassesByStringAt(index: Int)(getClassByString: String => Class[_]): List[Class[_]] = {
    tps.lift(index).map{ltp => ltp.map{tp => getClassByString(tp)}}.getOrElse(List())
  }
}
