package scala.tools.nsc.effects
package example
import scala.tools.nsc.Global


abstract class ExampleDomain extends EffectDomain {
  import global._

  // needs to be a lazy val - otherwise the `global` field of the lattice is set to `null`!
  lazy val lattice = new ExampleLattice {
    val global: ExampleDomain.this.global.type = ExampleDomain.this.global
  }
  import lattice._


  lazy val annotationClasses: List[Symbol] = List(readeClass,writeeClass,alloceClass,emptyRefeClass,testClass, subClass, subwriteeClass)
  lazy val annotationWithoutGenericClasses: List[Symbol] = List(readeClass,emptyRefeClass,subClass)

  lazy val barWriteeTrait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$writee"))
  lazy val barAlloce1Trait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$alloce1"))
  lazy val barAlloce2Trait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$alloce2"))
  lazy val barTestTrait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$test"))
  lazy val barSubwriteeTrait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$subwritee"))

  //override val uncheckedFunctions: List[String] = List[String]("print*")

  def parseAnnotationInfos(annots: List[AnnotationInfo], default: => Effect): Effect = {
    def exceptionsOf(tp: Type): List[Type] = tp match {
      case TypeRef(_, _, args) if args.size > 0 =>
        args.flatMap(exceptionsOf)
      case tp =>
        List(tp)
    }

    val anns = annots.filter(annotationClasses contains _.atp.typeSymbol)
    if (anns.isEmpty) {
      default
    } else {
      ((bottom) /: anns)((eff, annot) => {
        val effect =
          if(annot.atp.typeSymbol == readeClass) new ExampleReade()
          else if(annot.atp.typeSymbol == writeeClass) {
            val TypeRef(_, _, List(arg)) = annot.atp
            val tps = exceptionsOf(arg)
            //println(arg, "=>",tps)
            new ExampleWritee(tps)
          }
          else if(annot.atp.typeSymbol == alloceClass) {
            val TypeRef(_, _, List(arg1, arg2)) = annot.atp
            val tps = (exceptionsOf(arg1), exceptionsOf(arg2))
            new ExampleAlloce(tps)
          }
          else if(annot.atp.typeSymbol == emptyRefeClass) new ExampleEmptyRefe()
          else if(annot.atp.typeSymbol == testClass) {
            val TypeRef(_, _, List(arg)) = annot.atp
            val tps = exceptionsOf(arg)
            new ExampleTest(tps)
          }
          else if(annot.atp.typeSymbol == subClass) new ExampleSub()
          else{
            val TypeRef(_, _, List(arg)) = annot.atp
            val tps = exceptionsOf(arg)
            //println(arg, "=>",tps)
            new ExampleSubwritee(tps)
          }

        join(eff, ExampleEffect(List(effect), default.unknown))
      })
    }
  }



  def toAnnotation(eff: Effect): List[AnnotationInfo] = {

    def toAnnotationInfo(eff: ExampleType): AnnotationInfo = eff match {
      /** Sub must go  before Reade because Sub <: Reade **/
      case _: ExampleSub => AnnotationInfo(subClass.tpe, Nil, Nil)
      case _: ExampleReade => AnnotationInfo(readeClass.tpe, Nil, Nil)
      case ExampleWritee(tps) => AnnotationInfo(typeRef(writeeClass.tpe.prefix, writeeClass, List(toType(tps, barWriteeTrait))), Nil, Nil)
      case ExampleAlloce(tps) => AnnotationInfo(typeRef(alloceClass.tpe.prefix, alloceClass, List(toType(tps._1, barAlloce1Trait),toType(tps._2, barAlloce2Trait))), Nil, Nil)
      case _: ExampleEmptyRefe => AnnotationInfo(emptyRefeClass.tpe, Nil, Nil)
      case ExampleTest(tps) => AnnotationInfo(typeRef(testClass.tpe.prefix, testClass, List(toType(tps, barTestTrait))), Nil, Nil)
      case ExampleSubwritee(tps) => AnnotationInfo(typeRef(subwriteeClass.tpe.prefix, subwriteeClass, List(toType(tps, barSubwriteeTrait))), Nil, Nil)
      
    }
    def toType(eff: List[Type], barTrait: Symbol): Type = eff match {
      case Nil => nothingType
      case x :: Nil => x
      case x :: xs => typeRef(barTrait.tpe.prefix, barTrait, List(x, toType(xs, barTrait)))
    }



    val l = eff.tps.map{
      e => toAnnotationInfo(e)
    }
    if(eff.unknown)
      l ++ List(AnnotationInfo(unknownClass.tpe, Nil, Nil))
    else l
  }

  override def computeEffectImpl(tree: Tree, ctx: EffectContext): Effect = {
      neededEffect(tree).getOrElse(super.computeEffectImpl(tree, ctx))
  
  }

  override def neededEffect(tree: Tree): Option[lattice.Effect] = {

    tree match{
      
      case Apply(fun, args) => 
        
       /*call write*[T]() => check @writee[T] @reade*/
       val effs1: List[Effect] = (if("""(foo\.bar\..*)write.*$""".r.findAllIn(buildPath(fun)).hasNext){
          //for write[T] (type parameters)
          fun match{
            case TypeApply(f, as) => 
              List(ExampleEffect(new ExampleWritee(as.map{_.tpe})) u ExampleEffect(new ExampleReade()))
            case _ => List()
          }
        } else List()) 


        /*call alloc[T <: Object, X](V) => check @alloce[T, V] @test[T]*/
        val effs2: List[Effect] = (if("""walloc$""".r.findAllIn(buildPath(fun)).hasNext){
          fun match{
            case TypeApply(f, as) =>
              val params = fun.symbol.paramss.flatten.map{t => t.tpe}
              val genericParams = as.map{_.tpe}
              //println(genericParams.lift(0), rootMirror.getClassByName(newTypeName("foo.bar.base")).tpe )
              (for{
                g1 <- genericParams.lift(0) if g1 <:< rootMirror.getClassByName(newTypeName("foo.bar.base")).tpe //if g1 <:< definitions.ObjectClass.tpe // T
                g2 <- genericParams.lift(1) // X 
                p1 <- params.lift(0)
              } yield List(ExampleEffect(List(
                new ExampleAlloce((List(g1),List(p1))), 
                new ExampleTest(List(g1))
                )))).getOrElse(List())
            case _ => List()
          }
          
        } else List()) 
        //println(tree, effs2)

        /*call scala.*.print*(T <: Throwable) => check @test[T]*/
        val effs3: List[Effect] = (if("""scala.*\.print.*$""".r.findAllIn(buildPath(fun)).hasNext){
          val params = fun.symbol.paramss.flatten.map{t => t.tpe}
          (for{
            g1 <- params.lift(0) if g1 <:< definitions.ThrowableClass.tpe
            //p2 <- params.lift(1) if p2 ??? definitions.???.tpe
            //...
          } yield List(ExampleEffect(new ExampleTest(List(g1))))).getOrElse(List())
        } else List())

        val effs = effs1 ++ effs2 ++ effs3
        if(effs.length>0){
          Some((bottom /: effs)((feff, eff) =>
            join(eff, feff)))
          
        } else None

      case _ => None
    }
  }
}

abstract class ExampleLattice extends EffectLattice {
  val global: Global
  import global._

  lazy val readeClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.reade"))
  lazy val writeeClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.writee"))
  lazy val alloceClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.alloce"))
  lazy val emptyRefeClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.emptyRefe"))
  lazy val testClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.test"))
  lazy val subClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.sub"))
  lazy val subwriteeClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.subwritee"))

  lazy val nothingType = definitions.NothingClass.tpe

  type Effect = ExampleEffect
  //type privilege = scala.annotation.effects.Effect
  val ctEvidence = implicitly[scala.reflect.ClassTag[Effect]]

  abstract class ExampleType{
    def lte(b: ExampleType) = {
      b.getClass.isAssignableFrom(this.getClass)
    }
    def lteCovariant(aTp: Type, bTp: Type): Boolean = {
      aTp <:< bTp
    }
    def lteContravariant(aTp: Type, bTp: Type): Boolean = {
      bTp <:< aTp
    }
    def lteInvariant(aTp: Type, bTp: Type): Boolean = {
      aTp =:= bTp
    }
    def lteCoContravariant(aTp: Type, bTp: Type): Boolean = {
      bTp <:< aTp || aTp <:< bTp
    }

  }

  def lteOneTps[T](aTp: T, bTps: List[T], lteFunc: (T, T) => Boolean) = {
    bTps.exists(bTp => lteFunc(aTp, bTp))
  }

  class ExampleReade() extends ExampleType{
    override def toString(): String = {
      return "reade"
    }
  }

  class ExampleWritee(val tps: List[Type]) extends ExampleType{
    //def lte(b: ExampleType{ val tps: List[Type]}) = {
    override def lte(b: ExampleType) = b match {
      case bt: ExampleWritee => tps.forall(lteOneTps(_, bt.tps, lteCovariant))

      case _ => false
    }
    override def toString(): String = {
      return "writee[("+(tps.map{tp => tp.toString}.mkString(","))+")]"
    }
    /* for writee[+T]
    def lteOneOneTpsWritee(aTp: Type, bTp: Type): Boolean = {
      aTp <:< bTp
    }*/
  }
  object ExampleWritee{ def apply(tps: List[Type]) = new ExampleWritee(tps); def unapply(e: ExampleWritee): Option[List[Type]] = Some(e.tps)}

  class ExampleAlloce(val tps: (List[Type], List[Type])) extends ExampleType{
    //def lte(b: ExampleType{ val tps: (List[Type], List[Type])}) = {
    override def lte(b: ExampleType) =  b match {
      case bt: ExampleAlloce =>
          tps._1.forall(lteOneTps(_, bt.tps._1, lteCovariant)) &&
            tps._2.forall(lteOneTps(_, bt.tps._2, lteContravariant))

      case _ => false
    }
    override def toString(): String = {
      return "alloce[("+(tps._1.map{tp => tp.toString}.mkString(","))+"),("+(tps._2.map{tp => tp.toString}.mkString(","))+")]"
    }
    /* for alloce[+T, -V]
    def lteOneOneTpsAlloce(aTp: (Type, Type), bTp: (Type, Type)) = {
      (aTp._1 <:< bTp._1 && bTp._2 <:< aTp._2)
    }*/
  }
  object ExampleAlloce{ def apply(tps: (List[Type], List[Type])) = new ExampleAlloce(tps); def unapply(e: ExampleAlloce): Option[(List[Type], List[Type])] = Some(e.tps)}

  class ExampleEmptyRefe() extends ExampleType{
    override def lte(b: ExampleType) =  true /** BOTTOM! **/
    override def toString(): String = {
      return "emptyRefee"
    }
  }

  class ExampleTest(val tps: List[Type]) extends ExampleType{
    //def lte(b: ExampleType{ val tps: List[Type]}) = {
    override def lte(b: ExampleType) = b match {
      case bt: ExampleTest => tps.forall(lteOneTps(_, bt.tps, lteContravariant))
        
      case _ => false
    }
    override def toString(): String = {
      return "test[("+(tps.map{tp => tp.toString}.mkString(","))+")]"
    }
    /*for test[-T]
    def lteOneOneTpsTest(aTp: Type, bTp: Type) = {
      bTp <:< aTp
    }*/
  }
  object ExampleTest{ def apply(tps: List[Type]) = new ExampleTest(tps); def unapply(e: ExampleTest): Option[List[Type]] = Some(e.tps)}

  class ExampleSub() extends ExampleType{
    override def lte(b: ExampleType) = b match {
      case bt: ExampleReade => true
      case _ => false
    }
    override def toString(): String = {
      return "sub"
    }
  }

  class ExampleSubwritee(val tps: List[Type]) extends ExampleType{
    override def lte(b: ExampleType) = b match {
      case bt: ExampleWritee => tps.forall(lteOneTps(_, bt.tps, lteCovariant))
      case bt: ExampleSubwritee => tps.forall(lteOneTps(_, bt.tps, lteCovariant))   
      case _ => false
    }
    override def toString(): String = {
      return "subwritee[("+(tps.map{tp => tp.toString}.mkString(","))+")]"
    }
  }
  object ExampleSubwritee{ def apply(tps: List[Type]) = new ExampleSubwritee(tps); def unapply(e: ExampleSubwritee): Option[List[Type]] = Some(e.tps)}

  case class ExampleEffect(tps: List[ExampleType] = Nil, override val unknown: Boolean = false) extends WithUnknownEffect {
    def this(tp: ExampleType) = this(List(tp))
    override def toString() = {
      tps.map{tp => "@"+tp.toString()}.mkString(" ") + (if(unknown) "+¿" else "")
    }
  }
  object ExampleEffect {
    def apply(tp: ExampleType) = new ExampleEffect(tp)
  }


  lazy val top: Effect    = ExampleEffect(List(new ExampleReade,
    new ExampleWritee(List(definitions.ObjectClass.tpe)),
    new ExampleAlloce((List(definitions.AnyClass.tpe), List(definitions.NothingClass.tpe))),
    new ExampleTest(List(definitions.ThrowableClass.tpe))
  ))

  lazy val bottom: Effect = ExampleEffect(List(new ExampleEmptyRefe))
  lazy val unknown: Effect = new ExampleEffect(List(new ExampleEmptyRefe), true)
  def default = bottom

  def lte(a: Effect, b: Effect): Boolean = {
    val v = lteUnknown(a,b) ||
    a.tps.forall(lteOne(_, b))
    //if(a.toString.contains("A1") || b.toString.contains("A1"))
    //println(a,"<",b,"=", v, "?=", lteUnknown(a,b))
    v
    }

  // test for emptyRef, that way List(emptyRef) conforms to List()
  def lteOne(aTp: ExampleType, b: Effect): Boolean = aTp match{
    case _: ExampleEmptyRefe => true
    case _ => b.tps.exists(bTp => aTp.lte(bTp))      
  }

  def lteOneOne(aTp: ExampleType, bTp: ExampleType): Boolean = aTp.lte(bTp)

  /*(aTp, bTp) match{
    case (ExampleWritee(aTps), ExampleWritee(bTps)) => aTps.forall(lteOneTps(_, bTps,lteOneOneTpsWritee))
    case (ExampleAlloce(aTps), ExampleAlloce(bTps)) => aTps.forall(lteOneTps(_, bTps,lteOneOneTpsAlloce))
    case (ExampleTest(aTps), ExampleTest(bTps)) => aTps.forall(lteOneTps(_, bTps,lteOneOneTpsTest))
    case (a , b) => b.getClass.isAssignableFrom(a.getClass)
  }*/



  /*def elimSub(tps: List[Type], lteFunc: ): List[Type] = tps match {
    case Nil => Nil
    case t :: ts =>
      val rest = elimSub(ts.filter(t1 => !(t1 <:< t)))
      if (rest.exists(t1 => t <:< t1)) rest
      else t :: rest
  }*/
  /*def elimSub[T](tps: List[T], lteFunc: (T,T) => Boolean): List[T] = tps match{
    case Nil => Nil
    case t :: ts =>
      val rest = elimSub(ts.filter(t1 => !lteFunc(t1, t)), lteFunc)
      if (rest.exists(t1 => lteFunc(t, t1))) rest
      else t :: rest
  }

  def extractSameType[T](list: List[ExampleType]): List[T] = list.flatMap{ l =>
    l match {
      case a : T => List(a)
      case _ => List()
    }

  }*/
  def elimSub(aTps: List[ExampleType]): List[ExampleType] = aTps match {
    case Nil => Nil
    case t :: ts =>
      val (subtypes, others) = ts.partition(t1 => lteOneOne(t1, t))
      /*t match{
        case a: ExampleWritee =>
          ExampleWritee(elimSub(extractSameType[ExampleWritee](subtypes).flatMap{_.tps}, lteOneOneTpsWritee))
        case a: ExampleAlloce =>
          ExampleAlloce(elimSub(extractSameType[ExampleAlloce](subtypes).flatMap{_.tps}, lteOneOneTpsAlloce))
        case a: ExampleTest => a.tps
          ExampleTest(elimSub(extractSameType[ExampleTest](subtypes).flatMap{_.tps}, lteOneOneTpsTest))
        case _ => t
      }*/

      val rest = elimSub(others)
      if (rest.exists(t1 => lteOneOne(t, t1))) rest
      else t :: rest
  }

  def revoke(aTps: List[ExampleType], tp: ExampleType): List[ExampleType] = {
    aTps.filterNot(t => lteOneOne(t,tp))
  }

  def join(a: Effect, b: Effect): Effect = {
    val tps = if(a <= bottom) b.tps
    else if(b <= bottom) a.tps
    else a.tps ++ b.tps

    ExampleEffect(elimSub(tps), joinUnknown(a,b))
  }

  /**
   * think of 'allowed' exceptions
   *   - a allows E1 and E2
   *   - b allows E1 and E3
   *   - the meet(a,b) allows everything which is allowed by both, i.e. E1
   */
  def meet(a: Effect, b: Effect): Effect = {
    val keepA = a.tps.filter(lteOne(_, b))
    val keepB = b.tps.filter(lteOne(_, a))
    // remove duplicates. if a and b both contain some exception type E1,
    // that E1 will end up in both keepA and keepB
    val aNotB = keepA.filterNot(aTp => keepB.exists(bTp => aTp.lte(bTp)))
    ExampleEffect(aNotB ++ keepB, meetUnknown(a,b))
  }
  /* Usable for static effects */
  def diff(a: Effect, b: Effect): Effect = {
    val valid = a.tps.filter(!lteOne(_, b))
    ExampleEffect(elimSub(valid))
  }

  def getStrictPrivileges(eff: Effect): Effect = {
    ExampleEffect(eff.tps)
  }

  
}
