package scala.tools.nsc.effects
package myio
import scala.tools.nsc.Global


abstract class MyIODomain extends EffectDomain {
  import global._


  lazy val lattice = new MyIOLattice {
    val global: MyIODomain.this.global.type = MyIODomain.this.global
  }
  import lattice._

  lazy val annotationClasses: List[Symbol] = List(myNoIoClass, myIoClass, mySubClass, mySmallestClass, inputClass, outputClass)

  lazy val barMyIoTrait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$myIo"))
  lazy val barMySub1Trait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$mySub1"))
  lazy val barMySub2Trait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$mySub2"))
  lazy val barMySub3Trait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$mySub3"))
  lazy val barMySmallest1Trait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$mySmallest1"))
  lazy val barMySmallest2Trait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$mySmallest2"))
  lazy val barInputTrait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$input"))
  lazy val barOutputTrait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$output"))


  def parseAnnotationInfos(annots: List[AnnotationInfo], default: => Effect): Effect = {
     def exceptionsOf(tp: Type): List[Type] = tp match {
      case TypeRef(_, _, args) if args.size > 0 =>
        args.flatMap(exceptionsOf)
      case tp =>
        List(tp)
    }
    val anns = annots.filter(annotationClasses contains _.atp.typeSymbol)
    if (anns.isEmpty) {
      default
    } else {
      ((bottom) /: anns)((eff, annot) => {
        val effect =
          if(annot.atp.typeSymbol == myNoIoClass){
            new MyIOMyNoIo()
          }
          else if(annot.atp.typeSymbol == myIoClass){
            val TypeRef(_, _, List(arg)) = annot.atp
            val tps = exceptionsOf(arg)
            new MyIOMyIo(tps)
          }
          else if(annot.atp.typeSymbol == mySubClass){
            val TypeRef(_, _, List(arg1,arg2,arg3)) = annot.atp
            val tps = (exceptionsOf(arg1),exceptionsOf(arg2),exceptionsOf(arg3))
            new MyIOMySub(tps)
          }
          else if(annot.atp.typeSymbol == mySmallestClass){
            val TypeRef(_, _, List(arg1,arg2)) = annot.atp
            val tps = (exceptionsOf(arg1),exceptionsOf(arg2))
            new MyIOMySmallest(tps)
          }
          else if(annot.atp.typeSymbol == inputClass){
            val TypeRef(_, _, List(arg)) = annot.atp
            val tps = exceptionsOf(arg)
            new MyIOInput(tps)
          }
          else{
            val TypeRef(_, _, List(arg)) = annot.atp
            val tps = exceptionsOf(arg)
            new MyIOOutput(tps)
          }
        join(eff, MyIOEffect(List(effect), default.unknown))
      })
    }
  }

  def toAnnotation(eff: Effect): List[AnnotationInfo] = {
    def toAnnotationInfo(eff: MyIOType): AnnotationInfo = eff match {
      case _: MyIOMyNoIo => AnnotationInfo(myNoIoClass.tpe, Nil, Nil)
      case MyIOMySmallest(tps) => AnnotationInfo(typeRef(mySmallestClass.tpe.prefix, mySmallestClass, List(toType(tps._1, barMySmallest1Trait), toType(tps._2, barMySmallest2Trait))), Nil, Nil)
      case MyIOMySub(tps) => AnnotationInfo(typeRef(mySubClass.tpe.prefix, mySubClass, List(toType(tps._1, barMySub1Trait), toType(tps._2, barMySub2Trait), toType(tps._3, barMySub3Trait))), Nil, Nil)
      case MyIOMyIo(tps) => AnnotationInfo(typeRef(myIoClass.tpe.prefix, myIoClass, List(toType(tps, barMyIoTrait))), Nil, Nil)
      case MyIOInput(tps) => AnnotationInfo(typeRef(inputClass.tpe.prefix, inputClass, List(toType(tps, barInputTrait))), Nil, Nil)
      case MyIOOutput(tps) => AnnotationInfo(typeRef(outputClass.tpe.prefix, outputClass, List(toType(tps, barOutputTrait))), Nil, Nil)
    }

    def toType(eff: List[Type], barTrait: Symbol): Type = eff match {
      case Nil => nothingType
      case x :: Nil => x
      case x :: xs => typeRef(barTrait.tpe.prefix, barTrait, List(x, toType(xs, barTrait)))
    }



    val l = eff.tps.map{
      e => toAnnotationInfo(e)
    }
    if(eff.unknown)
      l ++ List(AnnotationInfo(unknownClass.tpe, Nil, Nil))
    else l
  }

  
    override def updateEffect(tree: Tree, eff: lattice.Effect, rtpe: Type): lattice.Effect = {
      val resultType = removeAllEffectAnnotations(rtpe)
      def obtainApply(tree: Tree): lattice.Effect = {
        val treeInfo.Applied(fun, tparams, vparamss) = tree
        val targs = tparams.map{t => t.tpe}
        val argss = vparamss.flatten.map{t => t.tpe}
        obtain("call", tree, fun, targs, argss)
      }
      def obtainDef(tree: Tree): lattice.Effect = {
        val targs = tree.symbol.typeParams.map{_.tpe}
        val DefDef(mods, name, tparams, vparamss, tpt, rhs) = tree
        val argss = vparamss.flatten.map{ _.symbol.tpe }
        obtain("def", tree, tree, targs, argss)
      }
      def obtain(operation: String, tree: Tree, fun: Tree, genericParams: List[Type], params: List[Type]): lattice.Effect = {
        val eff0: lattice.Effect = (if(operation == "call" && """scala\.Predef\.read[^.]*$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          (for{
                r <- Some(resultType)
          }
          yield  MyIOEffect(List(new MyIOInput(List(r.typeSymbol.tpe))))
          ).getOrElse(eff)
        } else eff) 

        val eff1: lattice.Effect = (if(operation == "call" && """println$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          (for{
                p1 <- params.lift(0) if p1 <:< definitions.StringClass.tpe
          }
          yield  MyIOEffect(List(new MyIOOutput(List(p1.typeSymbol.tpe))))
          ).getOrElse(eff0)
        } else eff0) 

        val eff2: lattice.Effect = (if(operation == "call" && """readM$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
           MyIOEffect(List(new MyIOInput(List(definitions.IntClass.tpe))))
        } else eff1) 

        val eff3: lattice.Effect = (if(operation == "def" && """scala\.Predef\.read[^.]*$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          (for{
                p1 <- params.lift(0)
                r <- Some(resultType)
          }
          yield  MyIOEffect(List(new MyIOInput(List(r.typeSymbol.tpe))))
          ).getOrElse(eff2)
        } else eff2) 

        val eff4: lattice.Effect = (if(operation == "call" && """models\..*\.insert$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          (for{
                p1 <- params.lift(0)
                g1 <- genericParams.lift(0)
          }
          yield eff3 u MyIOEffect(List(new MyIOInput(List(p1.typeSymbol.tpe)), new MyIOOutput(List(g1.typeSymbol.tpe))))
          ).getOrElse(eff3)
        } else eff3) 

        val eff5: lattice.Effect = (if(operation == "call" && """models\..*\.[^.]*$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
           MyIOEffect(List(new MyIOMyNoIo()))
        } else eff4) 

        val eff6: lattice.Effect = (if(operation == "def" && """.*\.Action$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
           MyIOEffect(List(new MyIOMyNoIo()))
        } else eff5) 
        eff6
      }
      tree match{
        case _: Apply | _: TypeApply => obtainApply(tree)
        case _: Ident | _: Select if tree.symbol != null && tree.symbol.isMethod => obtainApply(tree)
        case _: Ident | _: Select if tree.symbol != null && (tree.symbol.isModule && !tree.symbol.isPackage) => obtainApply(tree)
        case _: DefDef => obtainDef(tree)
        case _ => eff
      }
    }
    

  override def updateEffectDef(tree: Tree, effect: lattice.Effect, rtpe: Type): Option[lattice.Effect] = {
    def obtain(tree: Tree): Option[lattice.Effect] = {
      val fun = tree
      val targs = tree.symbol.typeParams
      val DefDef(mods, name, tparams, vparamss, tpt, rhs) = tree
        val effs1: List[Effect] = (if("""views\.html\.dummy\.apply$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          List(MyIOEffect(List(new MyIOMyNoIo())))
        }else List())
        val effs2: List[Effect] = (if("""views\.html\.foo\.apply$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          val genericParams = targs.map{_.tpe}
          (for{
                g1 <- genericParams.lift(0)
          }
              yield List(MyIOEffect(List(            new MyIOMyNoIo()))
          )).getOrElse(List())
        } else List()) 
        val effs3: List[Effect] = (if(""".*\.EffectController\.[^.]*$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          List(MyIOEffect(List(new MyIOMyNoIo())))
        }else List())
        val effs = effs1 ++ effs2 ++ effs3
        if(effs.length>0){
          Some((bottom /: effs)((feff, eff) =>
            join(eff, feff)))
        } else None
      }
      tree match{
        case _: DefDef => obtain(tree)
        case _ => None
      }
  }

  override def neededEffect(tree: Tree): Option[lattice.Effect] = {
    def obtain(tree: Tree): Option[lattice.Effect] = {
      val treeInfo.Applied(fun, targs, argss) = tree
        val effs1: List[Effect] = (if("""scala\.slick\.backend\..*\.[^.]*$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          List(MyIOEffect(List(new MyIOOutput(List(definitions.AnyClass.tpe)), new MyIOInput(List(definitions.AnyClass.tpe)))))
        }else List())
        val effs2: List[Effect] = (if("""scala\.slick\..*\.insert$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          val params = argss.flatten.map{t => t.tpe}
          (for{
                p1 <- params.lift(0)
          }
              yield List(MyIOEffect(List(            new MyIOInput(List(p1.typeSymbol.tpe))))
          )).getOrElse(List())
        } else List()) 
        val effs3: List[Effect] = (if(""".*\.Action$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          List(MyIOEffect(List(new MyIOMyNoIo())))
        }else List())
        val effs = effs1 ++ effs2 ++ effs3
        if(effs.length>0){
          Some((bottom /: effs)((feff, eff) =>
            join(eff, feff)))
        } else None
      }
      tree match{
        case _: Apply | _: TypeApply => obtain(tree)
        case _: Ident | _: Select if tree.symbol != null && tree.symbol.isMethod => obtain(tree)
        case _: Ident | _: Select if tree.symbol != null && (tree.symbol.isModule && !tree.symbol.isPackage) => obtain(tree)
        case _ => None
      }
  }
}

abstract class MyIOLattice extends EffectLattice {
  val global: Global
  import global._

  lazy val myNoIoClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.myNoIo"))
  lazy val myIoClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.myIo"))
  lazy val mySubClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.mySub"))
  lazy val mySmallestClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.mySmallest"))
  lazy val inputClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.input"))
  lazy val outputClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.output"))


  lazy val nothingType = definitions.NothingClass.tpe

  type Effect = MyIOEffect

  val ctEvidence = implicitly[scala.reflect.ClassTag[Effect]]

  abstract class MyIOType{
    def lte(b: MyIOType) = {
      b.getClass.isAssignableFrom(this.getClass)
    }
    def lteCovariant(aTp: Type, bTp: Type): Boolean = {
      aTp <:< bTp
    }
    def lteContravariant(aTp: Type, bTp: Type): Boolean = {
      bTp <:< aTp
    }
    def lteInvariant(aTp: Type, bTp: Type): Boolean = {
      aTp =:= bTp
    }
    def lteCoContravariant(aTp: Type, bTp: Type): Boolean = {
      aTp <:< bTp || bTp <:< aTp
    }

  }

  def lteOneTps[T](aTp: T, bTps: List[T], lteFunc: (T, T) => Boolean) = {
    bTps.exists(bTp => lteFunc(aTp, bTp))
  }
  class MyIOMyNoIo() extends MyIOType{
    override def lte(b: MyIOType) =  true
    override def toString(): String = {
      return "myNoIo"
    }
  }
  object MyIOMyNoIo{ def apply() = new MyIOMyNoIo(); def unapply(a: MyIOMyNoIo) = true}

  class MyIOMyIo(val tps: List[Type]) extends MyIOType{
    override def lte(b: MyIOType) = b match {
      case bt: MyIOMyIo =>
        tps.forall(lteOneTps(_, bt.tps, lteCovariant))
      case _ => false
    }
    override def toString(): String = {return "myIo["+(tps.map{tp => tp.toString}.mkString("|"))+"]"}
  }
  object MyIOMyIo{ def apply(tps: List[Type]) = new MyIOMyIo(tps); def unapply(e: MyIOMyIo): Option[List[Type]] = Some(e.tps)}

  class MyIOMySub(val tps: (List[Type], List[Type], List[Type])) extends MyIOType{
    override def lte(b: MyIOType) = b match {
      case bt: MyIOMySub =>
          tps._1.forall(lteOneTps(_, bt.tps._1, lteCovariant))  && tps._2.forall(lteOneTps(_, bt.tps._2, lteContravariant))  && tps._3.forall(lteOneTps(_, bt.tps._3, lteContravariant)) 
      case bt: MyIOMyIo =>
          tps._2.forall(lteOneTps(_, bt.tps, lteCoContravariant)) 

      case _ => false
    }
    override def toString(): String = {return "mySub["+(tps._1.map{tp => tp.toString}.mkString("|"))+", "+(tps._2.map{tp => tp.toString}.mkString("|"))+", "+(tps._3.map{tp => tp.toString}.mkString("|"))+"]"}
  }
  object MyIOMySub{ def apply(tps: (List[Type], List[Type], List[Type])) = new MyIOMySub(tps); def unapply(e: MyIOMySub): Option[(List[Type], List[Type], List[Type])] = Some(e.tps)}

  class MyIOMySmallest(val tps: (List[Type], List[Type])) extends MyIOType{
    override def lte(b: MyIOType) = b match {
      case bt: MyIOMySmallest =>
          tps._1.forall(lteOneTps(_, bt.tps._1, lteCovariant))  && tps._2.forall(lteOneTps(_, bt.tps._2, lteCovariant)) 
      case bt: MyIOMySub =>
          tps._2.forall(lteOneTps(_, bt.tps._2, lteCoContravariant)) 
      case bt: MyIOMyIo =>
          tps._2.forall(lteOneTps(_, bt.tps, lteCovariant)) 

      case _ => false
    }
    override def toString(): String = {return "mySmallest["+(tps._1.map{tp => tp.toString}.mkString("|"))+", "+(tps._2.map{tp => tp.toString}.mkString("|"))+"]"}
  }
  object MyIOMySmallest{ def apply(tps: (List[Type], List[Type])) = new MyIOMySmallest(tps); def unapply(e: MyIOMySmallest): Option[(List[Type], List[Type])] = Some(e.tps)}

  class MyIOInput(val tps: List[Type]) extends MyIOType{
    override def lte(b: MyIOType) = b match {
      case bt: MyIOInput =>
        tps.forall(lteOneTps(_, bt.tps, lteCovariant))
      case _ => false
    }
    override def toString(): String = {return "input["+(tps.map{tp => tp.toString}.mkString("|"))+"]"}
  }
  object MyIOInput{ def apply(tps: List[Type]) = new MyIOInput(tps); def unapply(e: MyIOInput): Option[List[Type]] = Some(e.tps)}

  class MyIOOutput(val tps: List[Type]) extends MyIOType{
    override def lte(b: MyIOType) = b match {
      case bt: MyIOOutput =>
        tps.forall(lteOneTps(_, bt.tps, lteCovariant))
      case _ => false
    }
    override def toString(): String = {return "output["+(tps.map{tp => tp.toString}.mkString("|"))+"]"}
  }
  object MyIOOutput{ def apply(tps: List[Type]) = new MyIOOutput(tps); def unapply(e: MyIOOutput): Option[List[Type]] = Some(e.tps)}

  case class MyIOEffect(tps: List[MyIOType] = Nil, override val unknown: Boolean = false) extends WithUnknownEffect {
    def this(tp: MyIOType) = this(List(tp))
    override def toString() = {
      tps.map{tp => "@"+tp.toString()}.mkString(" ") + (if(unknown) "+?" else "")
    }
  }

  object MyIOEffect {
    def apply(tp: MyIOType) = new MyIOEffect(tp)
  }

  lazy val top: Effect=
  MyIOEffect(List(new MyIOMyIo(List(definitions.AnyClass.tpe)), new MyIOInput(List(definitions.AnyClass.tpe)), new MyIOOutput(List(definitions.AnyClass.tpe))))

  lazy val bottom: Effect = MyIOEffect(List(new MyIOMyNoIo()))
  lazy val unknown: Effect = new MyIOEffect(List(new MyIOMyNoIo()), true)
  def default = bottom

  def lte(a: Effect, b: Effect): Boolean = {
    lteUnknown(a,b) ||
    a.tps.forall(lteOne(_, b))
  }

  def lteOne(aTp: MyIOType, b: Effect): Boolean = aTp match{
    case _: MyIOMyNoIo => true
    case _ => b.tps.exists(bTp => aTp.lte(bTp))
  }

  def lteOneOne(aTp: MyIOType, bTp: MyIOType): Boolean = aTp.lte(bTp)

  def elimSub(aTps: List[MyIOType]): List[MyIOType] = aTps match {
    case Nil => Nil
    case t :: ts =>
      val (subtypes, others) = ts.partition(t1 => lteOneOne(t1, t))
      val rest = elimSub(others)
      if (rest.exists(t1 => lteOneOne(t, t1))) rest
      else t :: rest
  }

  def revoke(aTps: List[MyIOType], tps: List[MyIOType]): List[MyIOType] = {
    (aTps /: tps)((atps, tp) => revoke(atps, tp))
  }

  def revoke(aTps: List[MyIOType], tp: MyIOType): List[MyIOType] = {
    aTps.filterNot(t => lteOneOne(t,tp))
  }

  def join(a: Effect, b: Effect): Effect = {
    val tps = if(a <= bottom) b.tps
    else if(b <= bottom) a.tps
    else a.tps ++ b.tps
    MyIOEffect(elimSub(tps), joinUnknown(a,b))
  }

  def meet(a: Effect, b: Effect): Effect = {
    val keepA = a.tps.filter(lteOne(_, b))
    val keepB = b.tps.filter(lteOne(_, a))
    val aNotB = keepA.filterNot(aTp => keepB.exists(bTp => aTp.lte(bTp)))
    MyIOEffect(aNotB ++ keepB, meetUnknown(a,b))
  }

  def diff(a: Effect, b: Effect): Effect = {
    val valid = a.tps.filter(!lteOne(_, b))
    MyIOEffect(elimSub(valid), diffUnknown(a,b))
  }

  def getStrictPrivileges(eff: Effect): Effect = {
    MyIOEffect(eff.tps)
  }
}
