package scala.tools.nsc.effects
package simpleio
import scala.tools.nsc.Global


abstract class SimpleIODomain extends EffectDomain {
  import global._


  lazy val lattice = new SimpleIOLattice {
    val global: SimpleIODomain.this.global.type = SimpleIODomain.this.global
  }
  import lattice._

  lazy val annotationClasses: List[Symbol] = List(simpleNoIOClass, simpleOutputClass, simpleInputClass)



  def parseAnnotationInfos(annots: List[AnnotationInfo], default: => Effect): Effect = {
     def exceptionsOf(tp: Type): List[Type] = tp match {
      case TypeRef(_, _, args) if args.size > 0 =>
        args.flatMap(exceptionsOf)
      case tp =>
        List(tp)
    }
    val anns = annots.filter(annotationClasses contains _.atp.typeSymbol)
    if (anns.isEmpty) {
      default
    } else {
      ((bottom) /: anns)((eff, annot) => {
        val effect =
          if(annot.atp.typeSymbol == simpleNoIOClass){
            new SimpleIOSimpleNoIO()
          }
          else if(annot.atp.typeSymbol == simpleOutputClass){
            new SimpleIOSimpleOutput()
          }
          else{
            new SimpleIOSimpleInput()
          }
        join(eff, SimpleIOEffect(List(effect), default.unknown))
      })
    }
  }

  def toAnnotation(eff: Effect): List[AnnotationInfo] = {
    def toAnnotationInfo(eff: SimpleIOType): AnnotationInfo = eff match {
      case _: SimpleIOSimpleNoIO => AnnotationInfo(simpleNoIOClass.tpe, Nil, Nil)
      case _: SimpleIOSimpleOutput => AnnotationInfo(simpleOutputClass.tpe, Nil, Nil)
      case _: SimpleIOSimpleInput => AnnotationInfo(simpleInputClass.tpe, Nil, Nil)
    }

    def toType(eff: List[Type], barTrait: Symbol): Type = eff match {
      case Nil => nothingType
      case x :: Nil => x
      case x :: xs => typeRef(barTrait.tpe.prefix, barTrait, List(x, toType(xs, barTrait)))
    }



    val l = eff.tps.map{
      e => toAnnotationInfo(e)
    }
    if(eff.unknown)
      l ++ List(AnnotationInfo(unknownClass.tpe, Nil, Nil))
    else l
  }

  
    override def updateEffect(tree: Tree, eff: lattice.Effect, rtpe: Type): lattice.Effect = {
      val resultType = removeAllEffectAnnotations(rtpe)
      def obtainApply(tree: Tree): lattice.Effect = {
        val treeInfo.Applied(fun, tparams, vparamss) = tree
        val targs = tparams.map{t => t.tpe}
        val argss = vparamss.flatten.map{t => t.tpe}
        obtain("call", tree, fun, targs, argss)
      }
      def obtainDef(tree: Tree): lattice.Effect = {
        val targs = tree.symbol.typeParams.map{_.tpe}
        val DefDef(mods, name, tparams, vparamss, tpt, rhs) = tree
        val argss = vparamss.flatten.map{ _.symbol.tpe }
        obtain("def", tree, tree, targs, argss)
      }
      def obtain(operation: String, tree: Tree, fun: Tree, genericParams: List[Type], params: List[Type]): lattice.Effect = {
                      val eff0: lattice.Effect = (if(operation == "def" && """views\.html\.dummy\.apply$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
           SimpleIOEffect(List(new SimpleIOSimpleNoIO()))
        } else eff) 

        val eff1: lattice.Effect = (if(operation == "def" && """views\.html\.foo\.apply$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          (for{
                g1 <- genericParams.lift(0)
          }
          yield  SimpleIOEffect(List(new SimpleIOSimpleNoIO()))
          ).getOrElse(eff0)
        } else eff0) 

        val eff2: lattice.Effect = (if(operation == "call" && """scala\.Predef\.read[^.]*$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          (for{
                r <- Some(resultType)
          }
          yield  SimpleIOEffect(List(new SimpleIOSimpleInput()))
          ).getOrElse(eff1)
        } else eff1) 

        val eff3: lattice.Effect = (if(operation == "call" && """fakePrint$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          (for{
                p1 <- params.lift(0) if p1 <:< definitions.StringClass.tpe
          }
          yield  SimpleIOEffect(List(new SimpleIOSimpleOutput()))
          ).getOrElse(eff2)
        } else eff2) 

        val eff4: lattice.Effect = (if(operation == "call" && """readM$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
           SimpleIOEffect(List(new SimpleIOSimpleInput()))
        } else eff3) 

        val eff5: lattice.Effect = (if(operation == "def" && """scala\.Predef\.read[^.]*$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          (for{
                p1 <- params.lift(0)
                r <- Some(resultType)
          }
          yield  SimpleIOEffect(List(new SimpleIOSimpleInput()))
          ).getOrElse(eff4)
        } else eff4) 

        val eff6: lattice.Effect =  eff5
        eff6
      }
      tree match{
        case _: Apply | _: TypeApply => obtainApply(tree)
        case _: Ident | _: Select if tree.symbol != null && tree.symbol.isMethod => obtainApply(tree)
        case _: Ident | _: Select if tree.symbol != null && (tree.symbol.isModule && !tree.symbol.isPackage) => obtainApply(tree)
        case _: DefDef => obtainDef(tree)
        case _ => eff
      }
    }
                    

  
    override def annotateEffect(tree: Tree): Option[lattice.Effect] = {
      def obtainApply(tree: Tree): Option[lattice.Effect] = {
        val treeInfo.Applied(fun, tparams, vparamss) = tree
        val targs = tparams.map{t => t.tpe}
        val argss = vparamss.flatten.map{t => t.tpe}
        obtain("call", tree, fun, targs, argss)
      }
      def obtainDef(tree: Tree): Option[lattice.Effect] = {
        val targs = tree.symbol.typeParams.map{_.tpe}
        val DefDef(mods, name, tparams, vparamss, tpt, rhs) = tree
        val argss = vparamss.flatten.map{ _.symbol.tpe }
        obtain("def", tree, tree, targs, argss)
      }
      def obtain(operation: String, tree: Tree, fun: Tree, genericParams: List[Type], params: List[Type]): Option[lattice.Effect] = {
        val eff0: Option[lattice.Effect] = (if(operation == "call" && """fff$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
           Some(SimpleIOEffect(List(new SimpleIOSimpleInput())))
        } else None) 
        eff0
      }
      tree match{
        case _: Apply | _: TypeApply => obtainApply(tree)
        case _: Ident | _: Select if tree.symbol != null && tree.symbol.isMethod => obtainApply(tree)
        case _: Ident | _: Select if tree.symbol != null && (tree.symbol.isModule && !tree.symbol.isPackage) => obtainApply(tree)
        case _: DefDef => obtainDef(tree)
        case _ => None
      }
    }
    

  override def updateEffectDef(tree: Tree, effect: lattice.Effect, rtpe: Type): Option[lattice.Effect] = {
    def obtain(tree: Tree): Option[lattice.Effect] = {
      val fun = tree
      val targs = tree.symbol.typeParams
      val DefDef(mods, name, tparams, vparamss, tpt, rhs) = tree
        val effs = List()
        if(effs.length>0){
          Some((bottom /: effs)((feff, eff) =>
            join(eff, feff)))
        } else None
      }
      tree match{
        case _: DefDef => obtain(tree)
        case _ => None
      }
  }

  override def neededEffect(tree: Tree): Option[lattice.Effect] = {
    def obtain(tree: Tree): Option[lattice.Effect] = {
      val treeInfo.Applied(fun, targs, argss) = tree
        val effs = List()
        if(effs.length>0){
          Some((bottom /: effs)((feff, eff) =>
            join(eff, feff)))
        } else None
      }
      tree match{
        case _: Apply | _: TypeApply => obtain(tree)
        case _: Ident | _: Select if tree.symbol != null && tree.symbol.isMethod => obtain(tree)
        case _: Ident | _: Select if tree.symbol != null && (tree.symbol.isModule && !tree.symbol.isPackage) => obtain(tree)
        case _ => None
      }
  }
}

abstract class SimpleIOLattice extends EffectLattice {
  val global: Global
  import global._

  lazy val simpleNoIOClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.simpleNoIO"))
  lazy val simpleOutputClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.simpleOutput"))
  lazy val simpleInputClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.simpleInput"))


  lazy val nothingType = definitions.NothingClass.tpe

  type Effect = SimpleIOEffect

  val ctEvidence = implicitly[scala.reflect.ClassTag[Effect]]

  abstract class SimpleIOType{
    def lte(b: SimpleIOType) = {
      b.getClass.isAssignableFrom(this.getClass)
    }
    def lteCovariant(aTp: Type, bTp: Type): Boolean = {
      aTp <:< bTp
    }
    def lteContravariant(aTp: Type, bTp: Type): Boolean = {
      bTp <:< aTp
    }
    def lteInvariant(aTp: Type, bTp: Type): Boolean = {
      aTp =:= bTp
    }
    def lteCoContravariant(aTp: Type, bTp: Type): Boolean = {
      aTp <:< bTp || bTp <:< aTp
    }

  }

  def lteOneTps[T](aTp: T, bTps: List[T], lteFunc: (T, T) => Boolean) = {
    bTps.exists(bTp => lteFunc(aTp, bTp))
  }
  class SimpleIOSimpleNoIO() extends SimpleIOType{
    override def lte(b: SimpleIOType) =  true
    override def toString(): String = {
      return "simpleNoIO"
    }
  }
  object SimpleIOSimpleNoIO{ def apply() = new SimpleIOSimpleNoIO(); def unapply(a: SimpleIOSimpleNoIO) = true}

  class SimpleIOSimpleOutput() extends SimpleIOType{
    override def toString(): String = {
      return "simpleOutput"
    }
  }
object SimpleIOSimpleOutput{ def apply() = new SimpleIOSimpleOutput(); def unapply(a: SimpleIOSimpleOutput) = true}

  class SimpleIOSimpleInput() extends SimpleIOType{
    override def toString(): String = {
      return "simpleInput"
    }
  }
object SimpleIOSimpleInput{ def apply() = new SimpleIOSimpleInput(); def unapply(a: SimpleIOSimpleInput) = true}

  case class SimpleIOEffect(tps: List[SimpleIOType] = Nil, override val unknown: Boolean = false) extends WithUnknownEffect {
    def this(tp: SimpleIOType) = this(List(tp))
    override def toString() = {
      tps.map{tp => "@"+tp.toString()}.mkString(" ") + (if(unknown) "+?" else "")
    }
  }

  object SimpleIOEffect {
    def apply(tp: SimpleIOType) = new SimpleIOEffect(tp)
  }

  lazy val top: Effect=
  SimpleIOEffect(List(new SimpleIOSimpleOutput(), new SimpleIOSimpleInput()))

  lazy val bottom: Effect = SimpleIOEffect(List(new SimpleIOSimpleNoIO()))
  lazy val unknown: Effect = new SimpleIOEffect(List(new SimpleIOSimpleNoIO()), true)
  def default = bottom

  def lte(a: Effect, b: Effect): Boolean = {
    lteUnknown(a,b) ||
    a.tps.forall(lteOne(_, b))
  }

  def lteOne(aTp: SimpleIOType, b: Effect): Boolean = aTp match{
    case _: SimpleIOSimpleNoIO => true
    case _ => b.tps.exists(bTp => aTp.lte(bTp))
  }

  def lteOneOne(aTp: SimpleIOType, bTp: SimpleIOType): Boolean = aTp.lte(bTp)

  def elimSub(aTps: List[SimpleIOType]): List[SimpleIOType] = aTps match {
    case Nil => Nil
    case t :: ts =>
      val (subtypes, others) = ts.partition(t1 => lteOneOne(t1, t))
      val rest = elimSub(others)
      if (rest.exists(t1 => lteOneOne(t, t1))) rest
      else t :: rest
  }

  def revoke(aTps: List[SimpleIOType], tps: List[SimpleIOType]): List[SimpleIOType] = {
    (aTps /: tps)((atps, tp) => revoke(atps, tp))
  }

  def revoke(aTps: List[SimpleIOType], tp: SimpleIOType): List[SimpleIOType] = {
    aTps.filterNot(t => lteOneOne(t,tp))
  }

  def join(a: Effect, b: Effect): Effect = {
    val tps = if(a <= bottom) b.tps
    else if(b <= bottom) a.tps
    else a.tps ++ b.tps
    SimpleIOEffect(elimSub(tps), joinUnknown(a,b))
  }

  def meet(a: Effect, b: Effect): Effect = {
    val keepA = a.tps.filter(lteOne(_, b))
    val keepB = b.tps.filter(lteOne(_, a))
    val aNotB = keepA.filterNot(aTp => keepB.exists(bTp => aTp.lte(bTp)))
    SimpleIOEffect(aNotB ++ keepB, meetUnknown(a,b))
  }

  def diff(a: Effect, b: Effect): Effect = {
    val valid = a.tps.filter(!lteOne(_, b))
    SimpleIOEffect(elimSub(valid), diffUnknown(a,b))
  }

  def getStrictPrivileges(eff: Effect): Effect = {
    SimpleIOEffect(eff.tps)
  }
}
