package scala.tools.nsc.effects
package simpleseio
import scala.tools.nsc.Global


abstract class SimpleSEIODomain extends EffectDomain {
  import global._


  lazy val lattice = new SimpleSEIOLattice {
    val global: SimpleSEIODomain.this.global.type = SimpleSEIODomain.this.global
  }
  import lattice._

  lazy val annotationClasses: List[Symbol] = List(simpleSENoIOClass, simpleSEOutputClass, simpleSEInputClass, simpleSESubinputClass)



  def parseAnnotationInfos(annots: List[AnnotationInfo], default: => Effect): Effect = {
     def exceptionsOf(tp: Type): List[Type] = tp match {
      case TypeRef(_, _, args) if args.size > 0 =>
        args.flatMap(exceptionsOf)
      case tp =>
        List(tp)
    }
    val anns = annots.filter(annotationClasses contains _.atp.typeSymbol)
    if (anns.isEmpty) {
      default
    } else {
      ((bottom) /: anns)((eff, annot) => {
        val effect =
          if(annot.atp.typeSymbol == simpleSENoIOClass){
            new SimpleSEIOSimpleSENoIO()
          }
          else if(annot.atp.typeSymbol == simpleSEOutputClass){
            new SimpleSEIOSimpleSEOutput()
          }
          else if(annot.atp.typeSymbol == simpleSEInputClass){
            new SimpleSEIOSimpleSEInput()
          }
          else{
            new SimpleSEIOSimpleSESubinput()
          }
        join(eff, SimpleSEIOEffect(List(effect), default.unknown))
      })
    }
  }

  def toAnnotation(eff: Effect): List[AnnotationInfo] = {
    def toAnnotationInfo(eff: SimpleSEIOType): AnnotationInfo = eff match {
      case _: SimpleSEIOSimpleSENoIO => AnnotationInfo(simpleSENoIOClass.tpe, Nil, Nil)
      case _: SimpleSEIOSimpleSEOutput => AnnotationInfo(simpleSEOutputClass.tpe, Nil, Nil)
      case _: SimpleSEIOSimpleSESubinput => AnnotationInfo(simpleSESubinputClass.tpe, Nil, Nil)
      case _: SimpleSEIOSimpleSEInput => AnnotationInfo(simpleSEInputClass.tpe, Nil, Nil)
    }

    def toType(eff: List[Type], barTrait: Symbol): Type = eff match {
      case Nil => nothingType
      case x :: Nil => x
      case x :: xs => typeRef(barTrait.tpe.prefix, barTrait, List(x, toType(xs, barTrait)))
    }



    val l = eff.tps.map{
      e => toAnnotationInfo(e)
    }
    if(eff.unknown)
      l ++ List(AnnotationInfo(unknownClass.tpe, Nil, Nil))
    else l
  }

  
    override def updateEffect(tree: Tree, eff: lattice.Effect, rtpe: Type): lattice.Effect = {
      val resultType = removeAllEffectAnnotations(rtpe)
      def obtainApply(tree: Tree): lattice.Effect = {
        val treeInfo.Applied(fun, tparams, vparamss) = tree
        val targs = tparams.map{t => t.tpe}
        val argss = vparamss.flatten.map{t => t.tpe}
        obtain("call", tree, fun, targs, argss)
      }
      def obtainDef(tree: Tree): lattice.Effect = {
        val targs = tree.symbol.typeParams.map{_.tpe}
        val DefDef(mods, name, tparams, vparamss, tpt, rhs) = tree
        val argss = vparamss.flatten.map{ _.symbol.tpe }
        obtain("def", tree, tree, targs, argss)
      }
      def obtain(operation: String, tree: Tree, fun: Tree, genericParams: List[Type], params: List[Type]): lattice.Effect = {
        val eff0: lattice.Effect = (if(operation == "call" && """scala\.Predef\.read[^.]*$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          (for{
                r <- Some(resultType)
          }
          yield  SimpleSEIOEffect(List(new SimpleSEIOSimpleSEInput()))
          ).getOrElse(eff)
        } else eff) 

        val eff1: lattice.Effect = (if(operation == "call" && """fakePrint$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          (for{
                p1 <- params.lift(0) if p1 <:< definitions.StringClass.tpe
          }
          yield  SimpleSEIOEffect(List(new SimpleSEIOSimpleSEOutput()))
          ).getOrElse(eff0)
        } else eff0) 

        val eff2: lattice.Effect = (if(operation == "call" && """readM$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
           SimpleSEIOEffect(List(new SimpleSEIOSimpleSEInput()))
        } else eff1) 

        val eff3: lattice.Effect = (if(operation == "def" && """scala\.Predef\.read[^.]*$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          (for{
                p1 <- params.lift(0)
                r <- Some(resultType)
          }
          yield  SimpleSEIOEffect(List(new SimpleSEIOSimpleSEInput()))
          ).getOrElse(eff2)
        } else eff2) 
        eff3
      }
      tree match{
        case _: Apply | _: TypeApply => obtainApply(tree)
        case _: Ident | _: Select if tree.symbol != null && tree.symbol.isMethod => obtainApply(tree)
        case _: Ident | _: Select if tree.symbol != null && (tree.symbol.isModule && !tree.symbol.isPackage) => obtainApply(tree)
        case _: DefDef => obtainDef(tree)
        case _ => eff
      }
    }
    

  override def updateEffectDef(tree: Tree, effect: lattice.Effect, rtpe: Type): Option[lattice.Effect] = {
    def obtain(tree: Tree): Option[lattice.Effect] = {
      val fun = tree
      val targs = tree.symbol.typeParams
      val DefDef(mods, name, tparams, vparamss, tpt, rhs) = tree
        val effs1: List[Effect] = (if("""views\.html\.dummy\.apply$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          List(SimpleSEIOEffect(List(new SimpleSEIOSimpleSENoIO())))
        }else List())
        val effs2: List[Effect] = (if("""views\.html\.foo\.apply$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          val genericParams = targs.map{_.tpe}
          (for{
                g1 <- genericParams.lift(0)
          }
              yield List(SimpleSEIOEffect(List(            new SimpleSEIOSimpleSENoIO()))
          )).getOrElse(List())
        } else List()) 
        val effs = effs1 ++ effs2
        if(effs.length>0){
          Some((bottom /: effs)((feff, eff) =>
            join(eff, feff)))
        } else None
      }
      tree match{
        case _: DefDef => obtain(tree)
        case _ => None
      }
  }

  override def neededEffect(tree: Tree): Option[lattice.Effect] = {
    def obtain(tree: Tree): Option[lattice.Effect] = {
      val treeInfo.Applied(fun, targs, argss) = tree
        val effs = List()
        if(effs.length>0){
          Some((bottom /: effs)((feff, eff) =>
            join(eff, feff)))
        } else None
      }
      tree match{
        case _: Apply | _: TypeApply => obtain(tree)
        case _: Ident | _: Select if tree.symbol != null && tree.symbol.isMethod => obtain(tree)
        case _: Ident | _: Select if tree.symbol != null && (tree.symbol.isModule && !tree.symbol.isPackage) => obtain(tree)
        case _ => None
      }
  }
}

abstract class SimpleSEIOLattice extends EffectLattice {
  val global: Global
  import global._

  lazy val simpleSENoIOClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.simpleSENoIO"))
  lazy val simpleSEOutputClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.simpleSEOutput"))
  lazy val simpleSEInputClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.simpleSEInput"))
  lazy val simpleSESubinputClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.simpleSESubinput"))


  lazy val nothingType = definitions.NothingClass.tpe

  type Effect = SimpleSEIOEffect

  val ctEvidence = implicitly[scala.reflect.ClassTag[Effect]]

  abstract class SimpleSEIOType{
    def lte(b: SimpleSEIOType) = {
      b.getClass.isAssignableFrom(this.getClass)
    }
    def lteCovariant(aTp: Type, bTp: Type): Boolean = {
      aTp <:< bTp
    }
    def lteContravariant(aTp: Type, bTp: Type): Boolean = {
      bTp <:< aTp
    }
    def lteInvariant(aTp: Type, bTp: Type): Boolean = {
      aTp =:= bTp
    }
    def lteCoContravariant(aTp: Type, bTp: Type): Boolean = {
      aTp <:< bTp || bTp <:< aTp
    }

  }

  def lteOneTps[T](aTp: T, bTps: List[T], lteFunc: (T, T) => Boolean) = {
    bTps.exists(bTp => lteFunc(aTp, bTp))
  }
  class SimpleSEIOSimpleSENoIO() extends SimpleSEIOType{
    override def lte(b: SimpleSEIOType) =  true
    override def toString(): String = {
      return "simpleSENoIO"
    }
  }
  object SimpleSEIOSimpleSENoIO{ def apply() = new SimpleSEIOSimpleSENoIO(); def unapply(a: SimpleSEIOSimpleSENoIO) = true}

  class SimpleSEIOSimpleSEOutput() extends SimpleSEIOType{
    override def toString(): String = {
      return "simpleSEOutput"
    }
  }
object SimpleSEIOSimpleSEOutput{ def apply() = new SimpleSEIOSimpleSEOutput(); def unapply(a: SimpleSEIOSimpleSEOutput) = true}

  class SimpleSEIOSimpleSEInput() extends SimpleSEIOType{
    override def toString(): String = {
      return "simpleSEInput"
    }
  }
object SimpleSEIOSimpleSEInput{ def apply() = new SimpleSEIOSimpleSEInput(); def unapply(a: SimpleSEIOSimpleSEInput) = true}

  class SimpleSEIOSimpleSESubinput() extends SimpleSEIOType{
    override def lte(b: SimpleSEIOType) = b match {
      case bt: SimpleSEIOSimpleSESubinput => true
      case bt: SimpleSEIOSimpleSEInput => true
      case _ => false
    }
    override def toString(): String = {
      return "simpleSESubinput"
    }
  }
object SimpleSEIOSimpleSESubinput{ def apply() = new SimpleSEIOSimpleSESubinput(); def unapply(a: SimpleSEIOSimpleSESubinput) = true}

  case class SimpleSEIOEffect(tps: List[SimpleSEIOType] = Nil, override val unknown: Boolean = false) extends WithUnknownEffect {
    def this(tp: SimpleSEIOType) = this(List(tp))
    override def toString() = {
      tps.map{tp => "@"+tp.toString()}.mkString(" ") + (if(unknown) "+?" else "")
    }
  }

  object SimpleSEIOEffect {
    def apply(tp: SimpleSEIOType) = new SimpleSEIOEffect(tp)
  }

  lazy val top: Effect=
  SimpleSEIOEffect(List(new SimpleSEIOSimpleSEOutput(), new SimpleSEIOSimpleSEInput()))

  lazy val bottom: Effect = SimpleSEIOEffect(List(new SimpleSEIOSimpleSENoIO()))
  lazy val unknown: Effect = new SimpleSEIOEffect(List(new SimpleSEIOSimpleSENoIO()), true)
  def default = bottom

  def lte(a: Effect, b: Effect): Boolean = {
    lteUnknown(a,b) ||
    a.tps.forall(lteOne(_, b))
  }

  def lteOne(aTp: SimpleSEIOType, b: Effect): Boolean = aTp match{
    case _: SimpleSEIOSimpleSENoIO => true
    case _ => b.tps.exists(bTp => aTp.lte(bTp))
  }

  def lteOneOne(aTp: SimpleSEIOType, bTp: SimpleSEIOType): Boolean = aTp.lte(bTp)

  def elimSub(aTps: List[SimpleSEIOType]): List[SimpleSEIOType] = aTps match {
    case Nil => Nil
    case t :: ts =>
      val (subtypes, others) = ts.partition(t1 => lteOneOne(t1, t))
      val rest = elimSub(others)
      if (rest.exists(t1 => lteOneOne(t, t1))) rest
      else t :: rest
  }

  def revoke(aTps: List[SimpleSEIOType], tps: List[SimpleSEIOType]): List[SimpleSEIOType] = {
    (aTps /: tps)((atps, tp) => revoke(atps, tp))
  }

  def revoke(aTps: List[SimpleSEIOType], tp: SimpleSEIOType): List[SimpleSEIOType] = {
    aTps.filterNot(t => lteOneOne(t,tp))
  }

  def join(a: Effect, b: Effect): Effect = {
    val tps = if(a <= bottom) b.tps
    else if(b <= bottom) a.tps
    else a.tps ++ b.tps
    SimpleSEIOEffect(elimSub(tps), joinUnknown(a,b))
  }

  def meet(a: Effect, b: Effect): Effect = {
    val keepA = a.tps.filter(lteOne(_, b))
    val keepB = b.tps.filter(lteOne(_, a))
    val aNotB = keepA.filterNot(aTp => keepB.exists(bTp => aTp.lte(bTp)))
    SimpleSEIOEffect(aNotB ++ keepB, meetUnknown(a,b))
  }

  def diff(a: Effect, b: Effect): Effect = {
    val valid = a.tps.filter(!lteOne(_, b))
    SimpleSEIOEffect(elimSub(valid), diffUnknown(a,b))
  }

  def getStrictPrivileges(eff: Effect): Effect = {
    SimpleSEIOEffect(eff.tps)
  }
}
