package scala.tools.nsc.effects
package simpletpio
import scala.tools.nsc.Global


abstract class SimpleTPIODomain extends EffectDomain {
  import global._


  lazy val lattice = new SimpleTPIOLattice {
    val global: SimpleTPIODomain.this.global.type = SimpleTPIODomain.this.global
  }
  import lattice._

  lazy val annotationClasses: List[Symbol] = List(simpleTPNoIOClass, simpleTPOutputClass, simpleTPInputClass, simpleTPSubinputClass)

  lazy val barSimpleTPInputTrait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$simpleTPInput"))
  lazy val barSimpleTPSubinputTrait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$simpleTPSubinput"))


  def parseAnnotationInfos(annots: List[AnnotationInfo], default: => Effect): Effect = {
     def exceptionsOf(tp: Type): List[Type] = tp match {
      case TypeRef(_, _, args) if args.size > 0 =>
        args.flatMap(exceptionsOf)
      case tp =>
        List(tp)
    }
    val anns = annots.filter(annotationClasses contains _.atp.typeSymbol)
    if (anns.isEmpty) {
      default
    } else {
      ((bottom) /: anns)((eff, annot) => {
        val effect =
          if(annot.atp.typeSymbol == simpleTPNoIOClass){
            new SimpleTPIOSimpleTPNoIO()
          }
          else if(annot.atp.typeSymbol == simpleTPOutputClass){
            new SimpleTPIOSimpleTPOutput()
          }
          else if(annot.atp.typeSymbol == simpleTPInputClass){
            val TypeRef(_, _, List(arg)) = annot.atp
            val tps = exceptionsOf(arg)
            new SimpleTPIOSimpleTPInput(tps)
          }
          else{
            val TypeRef(_, _, List(arg)) = annot.atp
            val tps = exceptionsOf(arg)
            new SimpleTPIOSimpleTPSubinput(tps)
          }
        join(eff, SimpleTPIOEffect(List(effect), default.unknown))
      })
    }
  }

  def toAnnotation(eff: Effect): List[AnnotationInfo] = {
    def toAnnotationInfo(eff: SimpleTPIOType): AnnotationInfo = eff match {
      case _: SimpleTPIOSimpleTPNoIO => AnnotationInfo(simpleTPNoIOClass.tpe, Nil, Nil)
      case _: SimpleTPIOSimpleTPOutput => AnnotationInfo(simpleTPOutputClass.tpe, Nil, Nil)
      case SimpleTPIOSimpleTPSubinput(tps) => AnnotationInfo(typeRef(simpleTPSubinputClass.tpe.prefix, simpleTPSubinputClass, List(toType(tps, barSimpleTPSubinputTrait))), Nil, Nil)
      case SimpleTPIOSimpleTPInput(tps) => AnnotationInfo(typeRef(simpleTPInputClass.tpe.prefix, simpleTPInputClass, List(toType(tps, barSimpleTPInputTrait))), Nil, Nil)
    }

    def toType(eff: List[Type], barTrait: Symbol): Type = eff match {
      case Nil => nothingType
      case x :: Nil => x
      case x :: xs => typeRef(barTrait.tpe.prefix, barTrait, List(x, toType(xs, barTrait)))
    }



    val l = eff.tps.map{
      e => toAnnotationInfo(e)
    }
    if(eff.unknown)
      l ++ List(AnnotationInfo(unknownClass.tpe, Nil, Nil))
    else l
  }

  
    override def updateEffect(tree: Tree, eff: lattice.Effect, rtpe: Type): lattice.Effect = {
      val resultType = removeAllEffectAnnotations(rtpe)
      def obtainApply(tree: Tree): lattice.Effect = {
        val treeInfo.Applied(fun, tparams, vparamss) = tree
        val targs = tparams.map{t => t.tpe}
        val argss = vparamss.flatten.map{t => t.tpe}
        obtain("call", tree, fun, targs, argss)
      }
      def obtainDef(tree: Tree): lattice.Effect = {
        val targs = tree.symbol.typeParams.map{_.tpe}
        val DefDef(mods, name, tparams, vparamss, tpt, rhs) = tree
        val argss = vparamss.flatten.map{ _.symbol.tpe }
        obtain("def", tree, tree, targs, argss)
      }
      def obtain(operation: String, tree: Tree, fun: Tree, genericParams: List[Type], params: List[Type]): lattice.Effect = {
        val eff0: lattice.Effect = (if(operation == "call" && """scala\.Predef\.read[^.]*$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          (for{
                r <- Some(resultType)
          }
          yield  SimpleTPIOEffect(List(new SimpleTPIOSimpleTPInput(List(r.typeSymbol.tpe))))
          ).getOrElse(eff)
        } else eff) 

        val eff1: lattice.Effect = (if(operation == "call" && """fakePrint$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          (for{
                p1 <- params.lift(0) if p1 <:< definitions.StringClass.tpe
          }
          yield  SimpleTPIOEffect(List(new SimpleTPIOSimpleTPOutput()))
          ).getOrElse(eff0)
        } else eff0) 

        val eff2: lattice.Effect = (if(operation == "call" && """readM$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
           SimpleTPIOEffect(List(new SimpleTPIOSimpleTPInput(List(definitions.IntClass.tpe))))
        } else eff1) 

        val eff3: lattice.Effect = (if(operation == "def" && """scala\.Predef\.read[^.]*$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          (for{
                p1 <- params.lift(0)
                r <- Some(resultType)
          }
          yield  SimpleTPIOEffect(List(new SimpleTPIOSimpleTPInput(List(r.typeSymbol.tpe))))
          ).getOrElse(eff2)
        } else eff2) 
        eff3
      }
      tree match{
        case _: Apply | _: TypeApply => obtainApply(tree)
        case _: Ident | _: Select if tree.symbol != null && tree.symbol.isMethod => obtainApply(tree)
        case _: Ident | _: Select if tree.symbol != null && (tree.symbol.isModule && !tree.symbol.isPackage) => obtainApply(tree)
        case _: DefDef => obtainDef(tree)
        case _ => eff
      }
    }
    

  override def updateEffectDef(tree: Tree, effect: lattice.Effect, rtpe: Type): Option[lattice.Effect] = {
    def obtain(tree: Tree): Option[lattice.Effect] = {
      val fun = tree
      val targs = tree.symbol.typeParams
      val DefDef(mods, name, tparams, vparamss, tpt, rhs) = tree
        val effs1: List[Effect] = (if("""views\.html\.dummy\.apply$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          List(SimpleTPIOEffect(List(new SimpleTPIOSimpleTPNoIO())))
        }else List())
        val effs2: List[Effect] = (if("""views\.html\.foo\.apply$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          val genericParams = targs.map{_.tpe}
          (for{
                g1 <- genericParams.lift(0)
          }
              yield List(SimpleTPIOEffect(List(            new SimpleTPIOSimpleTPNoIO()))
          )).getOrElse(List())
        } else List()) 
        val effs = effs1 ++ effs2
        if(effs.length>0){
          Some((bottom /: effs)((feff, eff) =>
            join(eff, feff)))
        } else None
      }
      tree match{
        case _: DefDef => obtain(tree)
        case _ => None
      }
  }

  override def neededEffect(tree: Tree): Option[lattice.Effect] = {
    def obtain(tree: Tree): Option[lattice.Effect] = {
      val treeInfo.Applied(fun, targs, argss) = tree
        val effs = List()
        if(effs.length>0){
          Some((bottom /: effs)((feff, eff) =>
            join(eff, feff)))
        } else None
      }
      tree match{
        case _: Apply | _: TypeApply => obtain(tree)
        case _: Ident | _: Select if tree.symbol != null && tree.symbol.isMethod => obtain(tree)
        case _: Ident | _: Select if tree.symbol != null && (tree.symbol.isModule && !tree.symbol.isPackage) => obtain(tree)
        case _ => None
      }
  }
}

abstract class SimpleTPIOLattice extends EffectLattice {
  val global: Global
  import global._

  lazy val simpleTPNoIOClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.simpleTPNoIO"))
  lazy val simpleTPOutputClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.simpleTPOutput"))
  lazy val simpleTPInputClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.simpleTPInput"))
  lazy val simpleTPSubinputClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.simpleTPSubinput"))


  lazy val nothingType = definitions.NothingClass.tpe

  type Effect = SimpleTPIOEffect

  val ctEvidence = implicitly[scala.reflect.ClassTag[Effect]]

  abstract class SimpleTPIOType{
    def lte(b: SimpleTPIOType) = {
      b.getClass.isAssignableFrom(this.getClass)
    }
    def lteCovariant(aTp: Type, bTp: Type): Boolean = {
      aTp <:< bTp
    }
    def lteContravariant(aTp: Type, bTp: Type): Boolean = {
      bTp <:< aTp
    }
    def lteInvariant(aTp: Type, bTp: Type): Boolean = {
      aTp =:= bTp
    }
    def lteCoContravariant(aTp: Type, bTp: Type): Boolean = {
      aTp <:< bTp || bTp <:< aTp
    }

  }

  def lteOneTps[T](aTp: T, bTps: List[T], lteFunc: (T, T) => Boolean) = {
    bTps.exists(bTp => lteFunc(aTp, bTp))
  }
  class SimpleTPIOSimpleTPNoIO() extends SimpleTPIOType{
    override def lte(b: SimpleTPIOType) =  true
    override def toString(): String = {
      return "simpleTPNoIO"
    }
  }
  object SimpleTPIOSimpleTPNoIO{ def apply() = new SimpleTPIOSimpleTPNoIO(); def unapply(a: SimpleTPIOSimpleTPNoIO) = true}

  class SimpleTPIOSimpleTPOutput() extends SimpleTPIOType{
    override def toString(): String = {
      return "simpleTPOutput"
    }
  }
object SimpleTPIOSimpleTPOutput{ def apply() = new SimpleTPIOSimpleTPOutput(); def unapply(a: SimpleTPIOSimpleTPOutput) = true}

  class SimpleTPIOSimpleTPInput(val tps: List[Type]) extends SimpleTPIOType{
    override def lte(b: SimpleTPIOType) = b match {
      case bt: SimpleTPIOSimpleTPInput =>
        tps.forall(lteOneTps(_, bt.tps, lteCovariant))
      case _ => false
    }
    override def toString(): String = {return "simpleTPInput["+(tps.map{tp => tp.toString}.mkString("|"))+"]"}
  }
  object SimpleTPIOSimpleTPInput{ def apply(tps: List[Type]) = new SimpleTPIOSimpleTPInput(tps); def unapply(e: SimpleTPIOSimpleTPInput): Option[List[Type]] = Some(e.tps)}

  class SimpleTPIOSimpleTPSubinput(val tps: List[Type]) extends SimpleTPIOType{
    override def lte(b: SimpleTPIOType) = b match {
      case bt: SimpleTPIOSimpleTPSubinput =>
          tps.forall(lteOneTps(_, bt.tps, lteCovariant)) 
      case bt: SimpleTPIOSimpleTPInput =>
          tps.forall(lteOneTps(_, bt.tps, lteCovariant)) 

      case _ => false
    }
    override def toString(): String = {return "simpleTPSubinput["+(tps.map{tp => tp.toString}.mkString("|"))+"]"}
  }
  object SimpleTPIOSimpleTPSubinput{ def apply(tps: List[Type]) = new SimpleTPIOSimpleTPSubinput(tps); def unapply(e: SimpleTPIOSimpleTPSubinput): Option[List[Type]] = Some(e.tps)}

  case class SimpleTPIOEffect(tps: List[SimpleTPIOType] = Nil, override val unknown: Boolean = false) extends WithUnknownEffect {
    def this(tp: SimpleTPIOType) = this(List(tp))
    override def toString() = {
      tps.map{tp => "@"+tp.toString()}.mkString(" ") + (if(unknown) "+?" else "")
    }
  }

  object SimpleTPIOEffect {
    def apply(tp: SimpleTPIOType) = new SimpleTPIOEffect(tp)
  }

  lazy val top: Effect=
  SimpleTPIOEffect(List(new SimpleTPIOSimpleTPOutput(), new SimpleTPIOSimpleTPInput(List(definitions.AnyClass.tpe))))

  lazy val bottom: Effect = SimpleTPIOEffect(List(new SimpleTPIOSimpleTPNoIO()))
  lazy val unknown: Effect = new SimpleTPIOEffect(List(new SimpleTPIOSimpleTPNoIO()), true)
  def default = bottom

  def lte(a: Effect, b: Effect): Boolean = {
    lteUnknown(a,b) ||
    a.tps.forall(lteOne(_, b))
  }

  def lteOne(aTp: SimpleTPIOType, b: Effect): Boolean = aTp match{
    case _: SimpleTPIOSimpleTPNoIO => true
    case _ => b.tps.exists(bTp => aTp.lte(bTp))
  }

  def lteOneOne(aTp: SimpleTPIOType, bTp: SimpleTPIOType): Boolean = aTp.lte(bTp)

  def elimSub(aTps: List[SimpleTPIOType]): List[SimpleTPIOType] = aTps match {
    case Nil => Nil
    case t :: ts =>
      val (subtypes, others) = ts.partition(t1 => lteOneOne(t1, t))
      val rest = elimSub(others)
      if (rest.exists(t1 => lteOneOne(t, t1))) rest
      else t :: rest
  }

  def revoke(aTps: List[SimpleTPIOType], tps: List[SimpleTPIOType]): List[SimpleTPIOType] = {
    (aTps /: tps)((atps, tp) => revoke(atps, tp))
  }

  def revoke(aTps: List[SimpleTPIOType], tp: SimpleTPIOType): List[SimpleTPIOType] = {
    aTps.filterNot(t => lteOneOne(t,tp))
  }

  def join(a: Effect, b: Effect): Effect = {
    val tps = if(a <= bottom) b.tps
    else if(b <= bottom) a.tps
    else a.tps ++ b.tps
    SimpleTPIOEffect(elimSub(tps), joinUnknown(a,b))
  }

  def meet(a: Effect, b: Effect): Effect = {
    val keepA = a.tps.filter(lteOne(_, b))
    val keepB = b.tps.filter(lteOne(_, a))
    val aNotB = keepA.filterNot(aTp => keepB.exists(bTp => aTp.lte(bTp)))
    SimpleTPIOEffect(aNotB ++ keepB, meetUnknown(a,b))
  }

  def diff(a: Effect, b: Effect): Effect = {
    val valid = a.tps.filter(!lteOne(_, b))
    SimpleTPIOEffect(elimSub(valid), diffUnknown(a,b))
  }

  def getStrictPrivileges(eff: Effect): Effect = {
    SimpleTPIOEffect(eff.tps)
  }
}
