package effectchecker
import scala.tools.nsc.Global

abstract class BiEffectDomain extends EffectDomain {
  import global._

  val d1: EffectDomain{
    val global: BiEffectDomain.this.global.type
  }
  val d2: EffectDomain{
    val global: BiEffectDomain.this.global.type
  }



  val lattice = new BiLattice {
    val global: BiEffectDomain.this.global.type = BiEffectDomain.this.global
    val l1: d1.lattice.type = d1.lattice
    val l2: d2.lattice.type = d2.lattice
  }

  import lattice._

  val annotationClasses: List[Symbol] = d1.annotationClasses ++ d2.annotationClasses

  def parseAnnotationInfos(annots: List[AnnotationInfo], default: => Effect): Effect =
    BiEffect(d1.parseAnnotationInfos(annots, default.e1), d2.parseAnnotationInfos(annots, default.e2))

  override def updateDontAdjust(tree: Tree, bool: Boolean = true) = {
    dontAdjust += (tree -> bool)
    d1.updateDontAdjust(tree, bool)
    d2.updateDontAdjust(tree, bool)
  }

  override def adjust(tree: Tree): Tree = d2.adjust(d1.adjust(tree))

  override def check(tree: Tree): List[String] = {
    d1.check(tree) ++ d2.check(tree)
  }

  override def checkDefault(tree: Tree): List[String] = {
    d1.checkDefault(tree) ++ d2.checkDefault(tree)
  }

  override def delta(tree: Tree): lattice.Effect = BiEffect(d1.delta(tree), d2.delta(tree))

  override def getEffect(tree: Tree): lattice.Effect  = {
    //println("d1 :"+d1.getEffect(tree)+" d2: "+d2.getEffect(tree))
    BiEffect(d1.getEffect(tree).asInstanceOf[lattice.l1.Effect], d2.getEffect(tree).asInstanceOf[lattice.l2.Effect])
  }

  override def adjust(tree: Tree, xi: lattice.Effect): Unit  = {
    d1.adjust(tree, xi.e1)
    d2.adjust(tree, xi.e2)
  }

  override def adjust(tree: Tree, xi: lattice.Effect, rel: RelEffectList): Unit  = {
    d1.adjust(tree, xi.e1)
    d2.adjust(tree, xi.e2)
    adjustRel(tree, rel)
  }
  override def adjustChildren(tree: Tree, xi: lattice.Effect): Unit  = {
    //println("BIEFFECT adjusting children of "+tree+" d1: " + d1, "e1: "+xi.e1)
    //println("BIEFFECT adjusting children of "+tree+" d2: " + d2, "e1: "+xi.e2)
    d1.adjustChildren(tree, xi.e1)
    d2.adjustChildren(tree, xi.e2)
  }
  override def adjustChildren(tree: Tree, xi: lattice.Effect, rel: RelEffectList): lattice.Effect  = {
    //println("adjusting children of "+tree+" d1: " + d1, "e1: "+xi.e1)
    //println("adjusting children of "+tree+" d2: " + d2, "e1: "+xi.e2)
    d1.adjustChildren(tree, xi.e1)
    d2.adjustChildren(tree, xi.e2)
    adjustChildrenRel(tree, rel)
    xi
  }
  override def adjustChildrenRec(tree: Tree, xi: lattice.Effect): Unit  = {
    d1.adjustChildrenRec(tree, xi.e1)
    d2.adjustChildrenRec(tree, xi.e2)
  }
  override def adjustChildrenRec(tree: Tree, xi: lattice.Effect, rel: RelEffectList): lattice.Effect  = {
    d1.adjustChildrenRec(tree, xi.e1)
    d2.adjustChildrenRec(tree, xi.e2)
    adjustChildrenRelRec(tree, rel)
    xi
  }

  override def overrideNeededEffects(tree: Tree, xi: lattice.Effect): Unit = {
    d1.overrideNeededEffects(tree, xi.e1)
    d2.overrideNeededEffects(tree, xi.e2)
  }

  override def updateEffect(tree: Tree, effect: lattice.Effect, rtpe: Type): lattice.Effect = {
    BiEffect(d1.updateEffect(tree, effect.e1, rtpe),
    d2.updateEffect(tree, effect.e2, rtpe))
  } 


  override def preAdjust(tree: Tree): Unit = {
    d1.preAdjust(tree)
    d2.preAdjust(tree)
  }


}
trait BiLattice extends EffectLattice {

  import global._
  val l1: EffectLattice
  val l2: EffectLattice

  type Effect = BiEffect

  val ctEvidence = implicitly[scala.reflect.ClassTag[Effect]]

  def top: Effect = BiEffect(l1.top, l2.top)
  def bottom: Effect = BiEffect(l1.bottom, l2.bottom)
  def unknown: Effect = new BiEffect(l1.unknown, l2.unknown)

  case class BiEffect(val e1: l1.Effect, val e2: l2.Effect)  extends WithUnknownEffect{
    override def toString = s"BiEffect($e1, $e2)"+(if(unknown) "+¿" else "")
    override val unknown = e1.unknown || e2.unknown
  }
  def default = BiEffect(l1.default, l2.default)

  def join(a: Effect, b: Effect): Effect = BiEffect(l1.join(a.e1, b.e1), l2.join(a.e2, b.e2))
  def meet(a: Effect, b: Effect): Effect = BiEffect(l1.meet(a.e1, b.e1), l2.meet(a.e2, b.e2))
  def lte(a: Effect, b: Effect): Boolean = l1.lte(a.e1, b.e1) && l2.lte(a.e2, b.e2)
  def diff(a: Effect, b: Effect): Effect = BiEffect(l1.diff(a.e1, b.e1), l2.diff(a.e2, b.e2))

  def getStrictPrivileges(eff: Effect): Effect = {
    BiEffect(l1.getStrictPrivileges(eff.e1), l2.getStrictPrivileges(eff.e2))
  }

  def getStrictPrivilegesTree(eff: Effect): List[scala.tools.nsc.Global#Tree] = l1.getStrictPrivilegesTree(eff.e1) ++ l2.getStrictPrivilegesTree(eff.e2)

  type OverrideEffect = NeededEffect
  val ctEvidenceOverrideEffect = implicitly[scala.reflect.ClassTag[OverrideEffect]]
  case class NeededEffect(e: Effect)

  def constructNeededEffect(e: Effect): OverrideEffect = NeededEffect(e)
}
