package effectchecker
package dbaccess
import scala.tools.nsc.Global


abstract class DBAccessDomain extends EffectDomain {
  import global._


  lazy val lattice = new DBAccessLattice {
    val global: DBAccessDomain.this.global.type = DBAccessDomain.this.global
  }
  import lattice._

  lazy val annotationClasses: List[Symbol] = List(accessClass, insertClass, noAccessClass)

  lazy val barInsertTrait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$insert"))


  def parseAnnotationInfos(annots: List[AnnotationInfo], default: => Effect): Effect = {
     def exceptionsOf(tp: Type): List[Type] = tp match {
      case TypeRef(_, _, args) if args.size > 0 =>
        args.flatMap(exceptionsOf)
      case tp =>
        List(tp)
    }
    val anns = annots.filter(annotationClasses contains _.atp.typeSymbol)
    if (anns.isEmpty) {
      default
    } else {
      ((bottom) /: anns)((eff, annot) => {
        val effect =
          if(annot.atp.typeSymbol == accessClass){
            new DBAccessAccess()
          }
          else if(annot.atp.typeSymbol == insertClass){
            val TypeRef(_, _, List(arg)) = annot.atp
            val tps = exceptionsOf(arg)
            new DBAccessInsert(tps)
          }
          else{
            new DBAccessNoAccess()
          }
        join(eff, DBAccessEffect(List(effect), default.unknown))
      })
    }
  }

  def toAnnotation(eff: Effect): List[AnnotationInfo] = {
    def toAnnotationInfo(eff: DBAccessType): AnnotationInfo = eff match {
      case _: DBAccessAccess => AnnotationInfo(accessClass.tpe, Nil, Nil)
      case DBAccessInsert(tps) => AnnotationInfo(typeRef(insertClass.tpe.prefix, insertClass, List(toType(tps, barInsertTrait))), Nil, Nil)
      case _: DBAccessNoAccess => AnnotationInfo(noAccessClass.tpe, Nil, Nil)
    }

    def toType(eff: List[Type], barTrait: Symbol): Type = eff match {
      case Nil => nothingType
      case x :: Nil => x
      case x :: xs => typeRef(barTrait.tpe.prefix, barTrait, List(x, toType(xs, barTrait)))
    }



    val l = eff.tps.map{
      e => toAnnotationInfo(e)
    }
    if(eff.unknown)
      l ++ List(AnnotationInfo(unknownClass.tpe, Nil, Nil))
    else l
  }

  
    override def updateEffect(tree: Tree, eff: lattice.Effect, rtpe: Type): lattice.Effect = {
      val resultType = removeAllEffectAnnotations(rtpe)
      def obtainApply(tree: Tree): lattice.Effect = {
        val treeInfo.Applied(fun, tparams, vparamss) = tree
        val targs = tparams.map{t => t.tpe}
        val argss = vparamss.flatten.map{t => t.tpe}
        obtain("call", tree, fun, targs, argss)
      }
      def obtainDef(tree: Tree): lattice.Effect = {
        val targs = tree.symbol.typeParams.map{_.tpe}
        val DefDef(mods, name, tparams, vparamss, tpt, rhs) = tree
        val argss = vparamss.flatten.map{ _.symbol.tpe }
        obtain("def", tree, tree, targs, argss)
      }
      def obtain(operation: String, tree: Tree, fun: Tree, genericParams: List[Type], params: List[Type]): lattice.Effect = {
                      val eff0: lattice.Effect = (if(operation == "def" && """views\.html\..*\.apply$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
           DBAccessEffect(List(new DBAccessNoAccess()))
        } else eff) 

        val eff1: lattice.Effect = (if(operation == "call" && """scala\.slick\.backend\..*\.[^.]*$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
           DBAccessEffect(List(new DBAccessAccess()))
        } else eff0) 

        val eff2: lattice.Effect = (if(operation == "def" && """models\.[^.]*$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
           DBAccessEffect(List(new DBAccessNoAccess()))
        } else eff1) 

        val eff3: lattice.Effect = (if(operation == "call" && """models\..*\.[^.]*$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
           DBAccessEffect(List(new DBAccessAccess()))
        } else eff2) 

        val eff4: lattice.Effect = (if(operation == "call" && """views\.html\..*\.main$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
           DBAccessEffect(List(new DBAccessNoAccess()))
        } else eff3) 

        val eff5: lattice.Effect = (if(operation == "call" && """models\..*\.insert$""".r.findAllIn(buildPath(fun.symbol)).hasNext){
          (for{
                p1 <- params.lift(0)
          }
          yield  DBAccessEffect(List(new DBAccessInsert(List(p1.typeSymbol.tpe))))
          ).getOrElse(eff4)
        } else eff4) 

        val eff6: lattice.Effect =  eff5
        eff6
      }
      tree match{
        case _: Apply | _: TypeApply => obtainApply(tree)
        case _: Ident | _: Select if tree.symbol != null && tree.symbol.isMethod => obtainApply(tree)
        case _: Ident | _: Select if tree.symbol != null && (tree.symbol.isModule && !tree.symbol.isPackage) => obtainApply(tree)
        case _: DefDef => obtainDef(tree)
        case _ => eff
      }
    }
                    

  
    override def annotateEffect(tree: Tree): Option[lattice.Effect] = {
      def obtainApply(tree: Tree): Option[lattice.Effect] = {
        val treeInfo.Applied(fun, tparams, vparamss) = tree
        val targs = tparams.map{t => t.tpe}
        val argss = vparamss.flatten.map{t => t.tpe}
        obtain("call", tree, fun, targs, argss)
      }
      def obtainDef(tree: Tree): Option[lattice.Effect] = {
        val targs = tree.symbol.typeParams.map{_.tpe}
        val DefDef(mods, name, tparams, vparamss, tpt, rhs) = tree
        val argss = vparamss.flatten.map{ _.symbol.tpe }
        obtain("def", tree, tree, targs, argss)
      }
      def obtain(operation: String, tree: Tree, fun: Tree, genericParams: List[Type], params: List[Type]): Option[lattice.Effect] = {
        val eff0: Option[lattice.Effect] = (if(operation == "call" && """models\..*\.insert$""".r.findAllIn(buildPath(fun.symbol)).hasNext & (within(tree, """controllers\.EffectController"""))){
          (for{
                p1 <- params.lift(0)
          }
          yield  DBAccessEffect(List(new DBAccessAccess(), new DBAccessInsert(List(rootMirror.getClassByName(newTypeName("models.User")).tpe))))
          )
        } else None) 
        eff0
      }
      tree match{
        case _: Apply | _: TypeApply => obtainApply(tree)
        case _: Ident | _: Select if tree.symbol != null && tree.symbol.isMethod => obtainApply(tree)
        case _: Ident | _: Select if tree.symbol != null && (tree.symbol.isModule && !tree.symbol.isPackage) => obtainApply(tree)
        case _: DefDef => obtainDef(tree)
        case _ => None
      }
    }
    

  override def adjust(tree: Tree): Tree = {
    super.adjust(tree)
    val xi = getEffect(tree)
    tree match{
      case DefDef(mods, name, tparams, vparamss, tpt, rhs) =>
      case _ =>
    }
    tree
  }

  override def neededEffect(tree: Tree): Option[lattice.Effect] = {
    def obtain(tree: Tree): Option[lattice.Effect] = {
      val treeInfo.Applied(fun, targs, argss) = tree
        val effs = List()
        if(effs.length>0){
          Some((bottom /: effs)((feff, eff) =>
            join(eff, feff)))
        } else None
      }
      tree match{
        case _: Apply | _: TypeApply => obtain(tree)
        case _: Ident | _: Select if tree.symbol != null && tree.symbol.isMethod => obtain(tree)
        case _: Ident | _: Select if tree.symbol != null && (tree.symbol.isModule && !tree.symbol.isPackage) => obtain(tree)
        case _ => None
      }
  }
}

abstract class DBAccessLattice extends EffectLattice {
  val global: Global
  import global._

  lazy val accessClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.access"))
  lazy val insertClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.insert"))
  lazy val noAccessClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.noAccess"))


  lazy val nothingType = definitions.NothingClass.tpe

  type Effect = DBAccessEffect

  val ctEvidence = implicitly[scala.reflect.ClassTag[Effect]]

  abstract class DBAccessType{
    def lte(b: DBAccessType) = {
      b.getClass.isAssignableFrom(this.getClass)
    }
    def lteCovariant(aTp: Type, bTp: Type): Boolean = {
      aTp <:< bTp
    }
    def lteContravariant(aTp: Type, bTp: Type): Boolean = {
      bTp <:< aTp
    }
    def lteInvariant(aTp: Type, bTp: Type): Boolean = {
      aTp =:= bTp
    }
    def lteCoContravariant(aTp: Type, bTp: Type): Boolean = {
      aTp <:< bTp || bTp <:< aTp
    }

  }

  def lteOneTps[T](aTp: T, bTps: List[T], lteFunc: (T, T) => Boolean) = {
    bTps.exists(bTp => lteFunc(aTp, bTp))
  }
  class DBAccessAccess() extends DBAccessType{
    override def toString(): String = {
      return "access"
    }
  }
object DBAccessAccess{ def apply() = new DBAccessAccess(); def unapply(a: DBAccessAccess) = true}

  class DBAccessInsert(val tps: List[Type]) extends DBAccessType{
    override def lte(b: DBAccessType) = b match {
      case bt: DBAccessInsert =>
        tps.forall(lteOneTps(_, bt.tps, lteCovariant))
      case _ => false
    }
    override def toString(): String = {return "insert["+(tps.map{tp => tp.toString}.mkString("|"))+"]"}
  }
  object DBAccessInsert{ def apply(tps: List[Type]) = new DBAccessInsert(tps); def unapply(e: DBAccessInsert): Option[List[Type]] = Some(e.tps)}

  class DBAccessNoAccess() extends DBAccessType{
    override def lte(b: DBAccessType) =  true
    override def toString(): String = {
      return "noAccess"
    }
  }
  object DBAccessNoAccess{ def apply() = new DBAccessNoAccess(); def unapply(a: DBAccessNoAccess) = true}

  case class DBAccessEffect(tps: List[DBAccessType] = Nil, override val unknown: Boolean = false) extends WithUnknownEffect {
    def this(tp: DBAccessType) = this(List(tp))
    override def toString() = {
      tps.map{tp => "@"+tp.toString()}.mkString(" ") + (if(unknown) "+?" else "")
    }
  }

  object DBAccessEffect {
    def apply(tp: DBAccessType) = new DBAccessEffect(tp)
  }

  lazy val top: Effect=
  DBAccessEffect(List(new DBAccessAccess(), new DBAccessInsert(List(definitions.AnyClass.tpe))))

  lazy val bottom: Effect = DBAccessEffect(List(new DBAccessNoAccess()))
  lazy val unknown: Effect = new DBAccessEffect(List(new DBAccessNoAccess()), true)
  def default = bottom

  def lte(a: Effect, b: Effect): Boolean = {
    lteUnknown(a,b) ||
    a.tps.forall(lteOne(_, b))
  }

  def lteOne(aTp: DBAccessType, b: Effect): Boolean = aTp match{
    case _: DBAccessNoAccess => true
    case _ => b.tps.exists(bTp => aTp.lte(bTp))
  }

  def lteOneOne(aTp: DBAccessType, bTp: DBAccessType): Boolean = aTp.lte(bTp)

  def elimSub(aTps: List[DBAccessType]): List[DBAccessType] = aTps match {
    case Nil => Nil
    case t :: ts =>
      val (subtypes, others) = ts.partition(t1 => lteOneOne(t1, t))
      val rest = elimSub(others)
      if (rest.exists(t1 => lteOneOne(t, t1))) rest
      else t :: rest
  }

  def revoke(aTps: List[DBAccessType], tps: List[DBAccessType]): List[DBAccessType] = {
    (aTps /: tps)((atps, tp) => revoke(atps, tp))
  }

  def revoke(aTps: List[DBAccessType], tp: DBAccessType): List[DBAccessType] = {
    aTps.filterNot(t => lteOneOne(t,tp))
  }

  def join(a: Effect, b: Effect): Effect = {
    val tps = if(a <= bottom) b.tps
    else if(b <= bottom) a.tps
    else a.tps ++ b.tps
    DBAccessEffect(elimSub(tps), joinUnknown(a,b))
  }

  def meet(a: Effect, b: Effect): Effect = {
    val keepA = a.tps.filter(lteOne(_, b))
    val keepB = b.tps.filter(lteOne(_, a))
    val aNotB = keepA.filterNot(aTp => keepB.exists(bTp => aTp.lte(bTp)))
    DBAccessEffect(aNotB ++ keepB, meetUnknown(a,b))
  }

  def diff(a: Effect, b: Effect): Effect = {
    val valid = a.tps.filter(!lteOne(_, b))
    DBAccessEffect(elimSub(valid), diffUnknown(a,b))
  }

  def getStrictPrivileges(eff: Effect): Effect = {
    DBAccessEffect(eff.tps)
  }
  def getStrictPrivilegesTree(eff: Effect): List[scala.tools.nsc.Global#Tree] = {
    eff.tps.flatMap{ tp => tp match{

      case _: DBAccessAccess => List(reify{scala.annotation.effects.access()}.tree)

      case DBAccessInsert(tps) =>
        val lists = List(
          Apply(Ident(reify{List}.tree.symbol), tps.map{tp => Literal(Constant(tp.toString()))})
        )
        if(eff.tps.size>0)
          List(Apply(Select(Ident(reify{scala.annotation.effects.runtimeInsert}.tree.symbol), newTermName("apply")),
            List(Apply(Ident(reify{List}.tree.symbol), lists))))
        else List[Tree]()

      case _: DBAccessNoAccess => List(reify{scala.annotation.effects.noAccess()}.tree)
      }
    }.toList
  }
    type OverrideEffect = NeededEffect
    val ctEvidenceOverrideEffect = implicitly[scala.reflect.ClassTag[OverrideEffect]]
    case class NeededEffect(e: Effect)
    def constructNeededEffect(e: Effect): OverrideEffect = NeededEffect(e)
}
