package effectchecker
import scala.reflect.ClassTag
import scala.tools.nsc._


abstract class EffectDomain extends RelEffects with Infer with DefaultEffects{
  self =>
  val global: Global
  import global._

  val lattice: EffectLattice
  import lattice._

  val annotationClasses: List[Symbol]

  val uncheckedFunctions: List[String] = List[String]()

  def adjust(tree: Tree, xi: lattice.Effect): Unit = {
    tree.updateAttachment[lattice.Effect](xi)
  }
  def adjustRel(tree: Tree, rel: RelEffectList): Unit = {
    tree.updateAttachment[RelEffectList](rel)
  }

  //def check(tree: Tree, eff: lattice.Effect): Boolean = lattice.check(tree, eff)
  def getEffect(tree: Tree): lattice.Effect  = {
    tree.attachments.get[lattice.Effect].getOrElse(lattice.default)
  }
  def getRelEffects(tree: Tree): RelEffectList  = {
    tree.attachments.get[RelEffectList].getOrElse(RelEffectList(List()))
  }

  def adjustChildren(tree: Tree, xi: lattice.Effect): Unit = {
    tree.children.map{ child => adjust(child, xi)}
  }
  def adjustChildrenRel(tree: Tree, rel: RelEffectList): Unit = {
    tree.children.map{ child => adjustRel(child, rel)}
  }
  def adjustChildren(tree: Tree, xi: lattice.Effect, rel: RelEffectList): lattice.Effect  = {
    //print("adjusting childs of ", tree.symbol,": ", tree.children.mkString("/"),xi)
    adjustChildren(tree, xi)
    adjustChildrenRel(tree, rel)
    xi
  }
  def adjustChildren(tree: Tree): lattice.Effect  = {
    val xi = getEffect(tree)
    adjustChildren(tree, xi)
    adjustChildrenRel(tree, getRelEffects(tree))
    xi
  }
  def adjustChildrenRec(tree: Tree, xi: lattice.Effect): Unit  = {
    tree.children.map{ child =>
      adjust(child, xi)
      //print("recursiv", child)
      adjustChildrenRec(child, xi)
    }
  }
  def adjustChildrenRelRec(tree: Tree, rel: RelEffectList): Unit  = {
    tree.children.map{ child =>
      adjustRel(child, rel)
      adjustChildrenRelRec(child, rel)
    }
  }
  def adjustChildrenRec(tree: Tree, xi: lattice.Effect, rel: RelEffectList): lattice.Effect  = {
    tree.children.map{ child =>
      adjustChildrenRec(child, xi)
      adjustChildrenRelRec(child, rel)
    }
    xi
  }

  def adjust(tree: Tree, xi: lattice.Effect, rel: RelEffectList): Unit = {
    //print("seteando permisos de ", tree.symbol, "a", xi, ". Antes tenia ", tree.attachments.get[lattice.Effect])
    adjust(tree, xi)
    adjustRel(tree, rel)
    //print("confirmando que ", tree.symbol, "tenga los permisos de", xi)
  }

  def updateEffect(tree: Tree, effect: lattice.Effect, rtpe: Type): lattice.Effect = effect

  def annotateEffect(tree: Tree): Option[lattice.Effect] = None
  //case class NeededEffects(e: Effect)
  def overrideNeededEffects(tree: Tree, xi: lattice.Effect): Unit = {
    tree.updateAttachment[lattice.OverrideEffect](lattice.constructNeededEffect(xi))
  }

  def getNeededEffect(tree: Tree): Option[lattice.Effect]  = {

    //println("getting for "+annotationClasses+" "+lattice.NeededEffect)
    tree.attachments.get[lattice.OverrideEffect].map{n => Some(n.e)}.getOrElse(None)
  }

  case class AnnProcessed()
  def annProcessed(tree: Tree) = {
    tree.updateAttachment[AnnProcessed](AnnProcessed())
  }
  def isAnnProcessed(tree: Tree) = {
    tree.attachments.get[AnnProcessed].isDefined
  }
  /* this function is originally recursivelly, but this brake the following (eff : @unknown) : @pure. Because it wouldn't insert the appropiated restrict pure */
  def collectAnnots(typed: Typed): (List[AnnotationInfo], Tree) = {
    val annots = typed.tpt.tpe.annotations
    //(annots, typed.expr)
    
    typed.expr match {
      case typed1: Typed =>
        val (restAnnots, expr) = collectAnnots(typed1)
        (annots ::: restAnnots, expr)
      case expr =>
        (annots, expr)
    }
  }


  /*
  * if it doesnt have annotations, maybe is a anonymous function
  * h(gRead _) se traduce a: h((x:Int) => gRead(x))
  * y la función anónima no tiene efectos por ende es bottom y hace fallar gRead
  * */
  def infersEffects(tree: Tree): lattice.Effect = {
    val childrenEffects = tree.children.map{ child =>
      if(child.symbol!=null){
        if(child.symbol.isAnonymousFunction)
          infersEffects(child)
        else
          fromAnnotationList(child.symbol.info.finalResultType.annotations)
      } else bottom
    }
    (bottom /: childrenEffects)((effects, eff) => {
      //val TypeRef(_, _, List(arg)) = annot.atp
      //println("class of ", annot, "is",annot.getClass)
      join(effects, eff)
    })
  }

  def typedDontAdjust(typed: Typed): Unit = {
    typed.expr match {
      case typed1: Typed =>
        updateDontAdjust(typed1)
        typedDontAdjust(typed1)
      case expr =>
    }
  }

  def preAdjust(tree: Tree): Unit = {
    val annEffs = annotateEffect(tree)
    annEffs.map{a => println(tree, a)}
    if(!isAnnProcessed(tree)){
      annEffs.map{ newExpected =>
      
        
        
        val rel = getRelEffects(tree)
       
        
        
        println("what to do now?")

        


        //adding the necessary has
        println("adding has?")

        val E1 = getStrictPrivileges(newExpected)
        val E = getStrictPrivileges(getEffect(tree))

      
        if(!dontAdjust.getOrElse(tree, false))
          insertRestrictLazy(tree, newExpected) /* do transform the tree, because we are not inserting this restrict in that tree */

        //printlnln("settung OT", expr, tree)
        tree.updateAttachment[OriginalTree](OriginalTree(tree))
        val deltas = diff(E1, E)

        if(!lte(deltas, bottom)){ /* if deltas is not bottom */
          //println("transformed", deltas)
          //setTransformed(tree) //not yet
          //transformed += (tree -> true)
          insertHasTree(tree, deltas)
        } 


        println("adjusting to "+newExpected)
        adjust(tree, newExpected)
        adjustRel(tree, rel)

        annProcessed(tree)

      }
    }
  }

  case class Parent(tree: Tree) 

  def setParent(tree: Tree) = {
    tree.children.map{child => 
      child.updateAttachment[Parent](Parent(tree))
    }
  }
  def getParent(tree: Tree): Option[Parent] = {
    tree.attachments.get[Parent]
  }

  def within(tree: Tree, path: String): Boolean = {
    if(tree.symbol != null){
      /*println("trying "+buildPath(tree.symbol))
      println("has?", (path.r.findAllIn(buildPath(tree.symbol)).hasNext))*/

      (path.r.findAllIn(buildPath(tree.symbol)).hasNext || 
      getParent(tree).map(p => within(p.tree, path)).getOrElse(false))
    } else false  
  }

  def adjust(tree: Tree): Tree = {
    //print("--------------------------")
    //print(tree)
    //println(Debugger.identyfySymbol(tree))
    //println("the effects: ", fromAnnotationList(tree.symbol.info.finalResultType.annotations))
    //println("0000000000000000000000000000")*/
    //if its a lambda, then adjust the xi to the infered one
    val sym = tree.symbol
   
    tree match{
      case typed: Typed => /* f() : @alloc @unknown */
        //print(tree,"... is typed...")
        /* Collect annotations */
        val (annots, expr) = collectAnnots(typed)
        val hasEffAnnots = existsEffectAnnotation(annots)
        //println("new expected:", annots, hasEffAnnots)
        //print("raw", showRaw(tree))
        /* If it has annotation about effects (can be others),
        * then adjust the children with the effect of the annotations.
        * Otherwise use the same as the parent */
        if (hasEffAnnots) {

          val newExpected = fromAnnotationList(annots)
          val rel = RelEffectList(relFromAnnotationList(annots))
          //println(tree, "has typed annotations", newExpected)

          /* new effects, we have to insert a modification of the runtime privileges*/
          
          typedDontAdjust(typed)
          
          if(!dontAdjust.getOrElse(tree, false))
            adjustChildren(tree, newExpected, rel)
          else
            adjustChildren(tree)
        }
        else{
          adjustChildren(tree)
        }
      

      case _: TypeDef | _: DefDef | _: ClassDef | _: ModuleDef | _: Function => /* its a lambda */
        /*if(tree.symbol.toString contains "main"){
          println("++++++++++++++++++++++++++++++")
          println("soy un lambda?: ", tree.symbol, tree)
          println("from annotations",fromAnnotationList(tree.symbol.info.finalResultType.annotations))
        }*/



        if(tree.symbol.isAnonymousFunction){ /*Not going to be annotated! => its going to propagate no effects for the children. We don't want that! */
          //println("an function "+tree,tree.symbol.tpe)
          /* what effect? the effect of the children combined (we have to do a lite-inference) */
          /* new effects, we have to insert a modification of the runtime privileges*/
          //println(tree.attachments.get[Boolean].getOrElse(false))
          //println(dontAdjust.get(tree))
          if(dontAdjust.getOrElse(tree, false)){ /* If this is a cast generated lambda, then we keep the current privileges */
            //println("dont ajust the privileges, asign the same privilege as the parent wich is:", getEffect(tree))
            adjustChildren(tree)
          }
          else{
            //println(dontAdjust)
            //println("...............-------------", tree, infersEffects(tree))
            //Assign the rel effect of the parent in case a function is used inside
            val relEffects = getRelEffects(tree)
            adjustChildren(tree, infersEffects(tree), relEffects)
          }
        }
        else{

          /* If it is a synthetic function definition, then is inserted by a compilator step. We don't want to generate errors from this. Let's adjust the body with Top */
          val annotatedEffects = if(tree.symbol.isSynthetic) top else fromAnnotationList(tree.symbol.info.finalResultType.annotations)
          val rel = RelEffectList(relFromAnnotationList(tree.symbol.info.finalResultType.annotations))
          val updatedEffects = updateEffect(tree, annotatedEffects, tree.symbol.info.finalResultType)

          /*if(tree.symbol.toString() contains "productElement"){
            println(tree,annotatedEffects)
          }*/
          //println("mis efectos relativos son ", rel)
          //println("mis efectos son: ", annotatedEffects)
          /* new effects, we have to insert a modification of the runtime privileges*/
          /*if(rel.rel.length>0){
            println(rel)
            println("adjusting children of ", tree, annotatedEffects, rel)
            println("children", tree.children)
            tree.children.map{ c =>
              println("c", c.attachments.all)
            }
          }*/
          //if(annotatedEffects.unknown)
          //println("adjusting children of "+tree+"to", annotatedEffects, updatedEffects)

          adjustChildren(tree, updatedEffects, rel)
          /*if(tree.symbol.toString contains "main"){
            tree.children.map{
              child => println("adjusting child "+child+"to "+ annotatedEffects)
            }
          }*/

          // fix para que no den problemas los argumentos con default => give all privileges to ValDefs
          fixPermissionsOfDefaultArguments(tree, rel)



        }

        /*
        this override is necesary because when checking because of this
        testPoly[String](45) where call testPoly[T] => check myIo[T]
        but def testPoly[T](x: Int): Int has no effect so the infer plugin says its Top
        This work for testPoly[String](45) but when checking the sub expressions it fail when 'testPoly' is checked because it doesn't match
        with the "call testPoly[T] ..."
        */
      case Apply(fun, args) =>
        neededEffect(tree).map{e => overrideNeededEffects(fun, e) }
        adjustChildren(tree)
      case TypeApply(fun, args) =>
        neededEffect(tree).map{e => overrideNeededEffects(fun, e) }
        adjustChildren(tree)

      /*case (_: Ident | _: Select) if sym!=null && sym.isMethod && isUncheckedFun(tree, uncheckedFunctions) => /* application */
        println("CASO RARO!!!!", tree)
        val relEffects = getRelEffects(tree)
        adjustChildren(tree, top, relEffects)*/
      case _ =>
        adjustChildren(tree)
    }
    tree
  }

  def fixPermissionsOfDefaultArguments(tree: Tree, rel: RelEffectList) = {
    tree match{
      case _: DefDef =>
        tree.children.map{ child =>
          child match {
            case _ : ValDef =>
            adjust(child, top, rel)
            case _ =>
          }

        }
      case _ =>
    }
  }
  val errors = List[String]()

  /* Return the required effect given an effect. None if no effect is required */
  /* Return the required effect given an effect. None if no effect is required */
  def neededEffect(tree: Tree): Option[lattice.Effect] = None

  def check(tree: Tree): List[String] = {
    val xi = getEffect(tree)
    //val needed = getNeededEffect(tree).map{ e => Some(e)}.getOrElse{neededEffect(tree)}

    //lets look for the needed effect 'e', then lets look for a pointcut that update that effect 'e'. If symbol is null we can not get the info about the result type.
    val updatedEffect = if(tree.symbol != null)
      updateEffect(tree, neededEffect(tree).getOrElse(bottom), tree.symbol.info.resultType)
    else neededEffect(tree).getOrElse(bottom)
    val n = getNeededEffect(tree).getOrElse{updatedEffect}
    //val needed = neededEffect(tree)

    //if(tree.toString contains "foo"){
      //println("the needed of "+tree+" es "+ needed+" y el overrided "+getNeededEffect(tree))
    //}

    val needed = if(n <= bottom) None else Some(n)


    needed.map{ eff =>
      //println("needed effect "+getNeededEffect(tree))
      //println("for tree "+tree+" "+annotationClasses+" checking if "+eff+" <= "+xi)
      if(eff <= xi){
        //checkDefault(tree)
        errors
      }
      else{
        println("for tree "+tree+" "+annotationClasses+" checking if "+eff+" <= "+xi)
        ("Expression does not have enough privileges to do that operation. \nGranted priv.:"+xi+"\nRequired priv.:"+eff) :: errors
      }
    }.getOrElse{checkDefault(tree)}
  }
  def checkDefault(tree: Tree): List[String] = {
    //Debugger.identyfySymbol(tree)
    val sym = tree.symbol
    tree match {
      case (_: Ident | _: Select) if sym.isMethod => /* application */



      if(sym.isPrimaryConstructor && sym.toString == "constructor Object" || isUncheckedFun(sym, uncheckedFunctions) || buildPath(sym) == "<error>"){
          errors
        } else {

        //println("checking "+tree)
        //Debugger.identyfySymbol(tree)
        //println(tree, sym, sym.isMethod, fromAnnotationList(tree.symbol.info.finalResultType.annotations), sym.isPrimaryConstructor, sym.isConstructor)
        //case Select(qual, _) =>
          //println("---------------------------------")
          //println("checking ",sym)
          //println("effects: ",tree.symbol.info.finalResultType.annotations)
          //que pasa con los terminos anotados?

          val privileges = getEffect(tree)


          //From annotations is not going to return the desirable result. For example if a method is defined on an abstract class, then it will probably will be typed as top given that there is not enough information.
          //The solution would be to update the fromAnnotationList effect as well
          val annotatedEffects = fromAnnotationList(tree.symbol.info.finalResultType.annotations)
          val updatedEffects = updateEffect(tree, annotatedEffects, tree.symbol.info.finalResultType)


          //println("checking needed",tree,  annotatedEffects, privileges)

          //println(E1)
          //rootMirror.getClassByName(newTypeName("hello.E1"))
          //println(rootMirror.getClassByName(newTypeName("E1")).tpe)

          if(!(updatedEffects <= privileges) && !hasRelativeEffect(tree) && !isTransformed(tree)) {
            updateEffect(tree, annotatedEffects, tree.symbol.info.finalResultType)
            //println("la magia esta aca:",annotatedEffects, "<=", privileges)
            //println("pero no esta actualizando "+ tree.symbol)
            //println("has?",hasRelativeEffect(tree))
            //println("rels: ",getRelEffects(tree))

            val relEffects = getRelEffects(tree)

            //println(tree, sym, annotatedEffects, isUncheckedFun(tree, uncheckedFunctions))

            ("-Expression "+tree+"-"+symbol+"-"+buildPath(sym)+"-"+uncheckedFunctions+"-"+updatedEffects+"- does not have enough privileges to do that operation. \nGranted priv.:"+privileges+"\nRequired priv.:"+updatedEffects) :: errors
          }
          else{ errors }
        }
      case This(qual) =>
        //println("checking ",qual)
        //println(tree.symbol.info.finalResultType.annotations)
        errors
      case _ => errors
    }
  }



  /** function to cast <T21 -E2-> T22 <= T11 -E1-> T12>e1, i.e., T11 -> T12 @E1 to T21 -> T22 @E2
  * we need originalTree to get the original effects of the application and re adjust the new trees accordingly
  */

  //def getE2Sym(e2Params)
    def isFunctionX(sym: Symbol, tpe: Type): Boolean = {
      import definitions.FunctionClass
      /*
      rootMirror.getClassByName(newTypeName("scala.Function1")).tpe <:< tpe.typeSymbol.tpe ||
        rootMirror.getClassByName(newTypeName("scala.Function2")).tpe <:< tpe.typeSymbol.tpe ||
        rootMirror.getClassByName(newTypeName("scala.Function3")).tpe <:< tpe.typeSymbol.tpe ||
        rootMirror.getClassByName(newTypeName("scala.Function4")).tpe <:< tpe.typeSymbol.tpe ||
        rootMirror.getClassByName(newTypeName("scala.Function5")).tpe <:< tpe.typeSymbol.tpe
        */
      //tpe.member(newTermName("apply")).exists && (tpe.member(newTermName("curried")).exists || tpe.member(newTermName("andThen")).exists)
      //TODO: Nothing results, FIX!!!!
      tpe match{
        case TypeRef(_, sym, refArgs) if refArgs.length - 1>0 && sym == FunctionClass(refArgs.length - 1) => true
        case refType @ RefinedType(parents, decls) => true
        case _ => false
      }
      tpe.toString.contains(" => ")

    }

  case class OriginalTree(tree: Tree)

  def insertCast(e1: Tree, E1: Effect, E2: Effect, e2s: List[List[Tree]], originalTree: Tree, currentOwner: Symbol, targs: List[Tree]): Tree = {
    val pos = e1.pos
    //printlnln("inserting Cast into ", e1, E2+"<-"+E1)

    def castAsLambda(e1: Tree, E1: Effect, E2: Effect, e1ApplySym: Symbol, e2ListType: List[List[Type]], lvl: Int, e1FinalType: Type, targs: List[Tree], inverted: Boolean = true, castToPoly: List[List[Boolean]] = List(), insertHas: Boolean = true): Tree = {

      val sE2 = getStrictPrivileges(E2)
      val sE1 = getStrictPrivileges(E1)
      val sE1_E2: Effect = diff(sE1, sE2)
      //val paramsE1 = e1ApplySym.paramss.flatten
      /* fix to get the types of trait Function1 */
      val paramsE1 = e1ApplySym.typeSignatureIn(e1FinalType).paramss

      //println("<"+E2+"<="+E1+">"+e1+"")
      //println("e2type", e2ListType)
      val nextInsertHas = true
      //println("params ?? ",paramsE1)
      //println("cast as lambda", e1, e2ListType)

      /*calculate if we need to insert the has in the next cast*/

      def buildApply(func: Tree, argss: List[List[Symbol]], argTypess: List[List[Type]]) = {
        //println("PPPPPPPPPP", func, argss.headOption.getOrElse(List()), argTypess.headOption.getOrElse(List()), lvl)

        val base = Apply(func, listofIdent(argss.headOption.getOrElse(List()), argTypess.headOption.getOrElse(List()), lvl,0, inverted, castToPoly.headOption.getOrElse(List())))
        var from = argss.headOption.getOrElse(List()).length
        if(argss.length>0){ /* case for curryfication*/
          val rest = argss.tail zip argTypess.tail zip castToPoly
          //println("OOOOOOOOOO", base, rest)
          val q = (base /: rest){(apply, argsPoly) =>
            val oldFrom = from
            val args = argsPoly._1
            from += args._1.length
            Apply(apply, listofIdent(args._1, args._2, lvl, oldFrom, inverted, argsPoly._2))}
          //println("QQQQQQSSS", q)
          q
        } else base
      }


      val applyFun = e1 match{
        case (_: Ident | _: Select ) if  e1ApplySym.typeSignatureIn(e1FinalType).paramss.length<=0 => e1
        //case _: TypeApply => TypeApply(e1, )
        case _ =>
          val fun = if(targs.length>0)
            TypeApply(e1, targs)
          else e1
          //println("buildApply", fun, paramsE1, e2ListType)
          buildApply(fun, paramsE1, e2ListType)
          //Apply(e1, listofIdent(paramsE1.flatten, e2ListType.flatten, lvl))
      }
      //println("EL APPLY FUN", applyFun)


      //println("<<"+E2+"<="+E1+">>"+e1+"")
      //println("RESTRICT ", E2)
      //if(insertHas)
      //  println("HAS ", sE1_E2)

      //println("sE1_E2",sE1_E2)
      val hasLambda = if(insertHas) insertHasLazy(applyFun, sE1_E2) else applyFun
      applyFun.updateAttachment[OriginalTree](OriginalTree(originalTree)) /* for blame */
      //println("inserting restrict of ", E2)

      /* TODO: restricted must consider the latent effects */
      val restricted = insertRestrictLazy(hasLambda, E2)

      /*Function(
        List(ValDef(Modifiers(PARAM), TermName("x"), Ident(TypeName("Int")), EmptyTree)),
        Function(
          List(ValDef(Modifiers(PARAM), TermName("y"), Ident(TypeName("Int")), EmptyTree)),
          Apply(Select(Ident(TermName("x")), TermName("$times")), List(Ident(TermName("y"))))
        )
      )*/

      def buildLambda(symbolss: List[List[Symbol]], lvl: Int, from: Int, then: Tree): Tree = {
        symbolss match{
          case Nil => then
          case x :: xs =>
            val lambda = Function(listofValDef(x, lvl, from, e2ListType.flatten), buildLambda(xs, lvl,from+x.length, then))
            updateDontAdjust(lambda)
            lambda
        }
      }
      //println("building lambda for ", paramsE1, lvl, 0, restricted)
      val lambda =
        if(paramsE1.length>0) buildLambda(paramsE1, lvl, 0, restricted)//Function(listofValDef(paramsE1, lvl), restricted)
        //In case its a function that receive no arguments
        else Function(List(), restricted)



      /* Don't adjust the typedLambda tree, or it is going to try to infer it given that is a lambda */
      updateDontAdjust(lambda)
      /* set the applyFun transformed so it won't enter in a infinite loop */
      //println("set transformed to", applyFun)
      setTransformed(applyFun)

      /*e1 match{
        case _: Ident | _: Select => setTransformed(e1)
        case _ =>
      }*/

      setTransformedRec(restricted)

      lambda
    }

    type t = scala.Predef.Function[Int,Int]



    def listofIdent(params: List[Symbol], e2ParamTypes: List[Type], lvl: Int = 0, from: Int, inverted: Boolean, castToPoly: List[Boolean]): List[Tree] = {
      params.zipWithIndex.map{
        case (symbol,index) => {
          val x = Ident(newTermName("x$"+lvl+"$"+(index+from)))
          val e1Sym = symbol.tpe.member(newTermName("apply"))
          //println("E1SYM", symbol, symbol.tpe, e1Sym)
          //println("E2SYM", e2ParamTypes(index))
          //val e1Sym = symbol.tpe.finalResultType.computeMemberType(e1Sym_)
          val e2Sym = e2ParamTypes(index).finalResultType.member(newTermName("apply"))

          //println(" P P P P P",symbol.tpe ,symbol.info, symbol.info.typeSymbol, symbol.tpe.isStructuralRefinement, symbol.info.typeSymbol.isAnonOrRefinementClass)
          //println(" O O O O", symbol.info.baseTypeSeq, symbol.tpe.baseTypeSeq)
          //println(" members", e2ParamTypes(index).finalResultType.typeSymbol,e2ParamTypes(index).finalResultType.typeSymbol.tpe, e2ParamTypes(index).finalResultType.members)

          //
          //println("is functionX?", isFunctionX(symbol, symbol.tpe), symbol.tpe, symbol.tpe.dealias)

          //if(e1Sym.exists && e2Sym.exists){
          if(isFunctionX(symbol, symbol.tpe.dealias)/* && isFunctionX(e2ParamTypes(index).finalResultType)*/){
            val (newe1,newe2) = if(inverted){
              (e2Sym, e1Sym)
            } else{
              (e1Sym, e2Sym)
            }
            val poly = isPolymorphic(newe2)
            //println("next poly information : "+poly)
            val E11 = fromAnnotationList(newe1.tpe.finalResultType.annotations)
            val E21 = fromAnnotationList(newe2.tpe.finalResultType.annotations)

            val E2latent = latent(newe2, Map(), Map(), getRelEffects(originalTree))
            //println("will introduce a cast lambda in",x, e1Sym, e2Sym, E21+"<-"+E11)
            val insertHas = !castToPoly.lift(index).getOrElse(false)
            //println("next cast will introduce a has? : "+insertHas)
            castAsLambda(x, E11, E2latent, e1Sym, e2Sym.paramss.map{l => l.map{ll => ll.tpe}}, lvl+1, symbol.tpe.finalResultType, List(), !inverted, poly, insertHas)

          }
          else x
        }
      }
    }
    def listofValDef(params: List[Symbol], lvl: Int = 0, from: Int = 0, e2ListType: List[Type]): List[ValDef] = {
      params.zipWithIndex.map{
        case (symbol,index) =>

          val typeTree = symbol.tpe.finalResultType
          //println("THE TREE:", symbol, typeTree, e2ListType(index))
          ValDef(Modifiers(scala.reflect.internal.Flags.PARAM | scala.reflect.internal.Flags.SYNTHETIC), newTermName("x$"+lvl+"$"+(index+from)), TypeTree(e2ListType(index)), EmptyTree)
      }
    }

    //println("inserting cast on ",e1,  E2, "<-", E1)

    /*val E2 = (bottom /: unaryArgs.toList)((effects, arg) => {
      join(effects, infersEffects(arg._2))
    })*/
    /*val sE2 = getStrictPrivileges(E2)
    val sE1 = getStrictPrivileges(E1)
    val sE1_E2 = sE1.diff(sE2)

    val params = e1.symbol.paramss.flatten

    val applyFun = Apply(e1, listofIdent(params, e2s.map{_.symbol}))
    println("the apply fun is", applyFun)
    println("has: ",sE1_E2)
    val hasLambda = insertHasLazy(applyFun, sE1_E2)
    println("restrict: ",E2)
    val restricted = insertRestrictLazy(hasLambda, E2)
    //println("restricted", restricted)
    */
    //val lambda = Function(listofValDef(params),restricted)
    //val lambda = castAsLambda(e1, E1, E2, e1.symbol, e2s.map{_.tpe.finalResultType.member(newTermName("apply"))},0)

    val lambda = castAsLambda(e1, E1, E2, e1.symbol, e2s.map{l => l.map{ll => ll.tpe}},0, e1.symbol.tpe.finalResultType, targs)



    /* set the applyFun transformed so it won't enter in a infinite loop */
    //transformed += (applyFun -> true)

    //println(lambda)
    //println(showRaw(lambda))
    val typedLambda = lambda//typer.typed(lambda)
    //dontCheck += (applyFun -> true)
    //dontCheck += (lambda -> true)



    typedLambda
    /*
    //val castedE2 = insertCast(e2s)
    println(typedLambda, "applied to ", e2s_)
    val typedTree = typer.typed(Apply(Select(typedLambda, newTermName("apply")), e2s_))


    //println("raw",showRaw(typedTree))
    adjust(typedTree, getEffect(originalTree))
    //println("typedTree", getEffect(tree))
    /* adjust the lambda so when typechecking won't throw new errors */
    adjustChildren(typedTree, getEffect(originalTree))
    //print("changing owner from", typedLambda.symbol.owner)
    //typedTree.changeOwner((typedLambda.symbol.owner, typer.context.owner))
    //typedTree.changeOwner((typedLambda.symbol.owner, tree.symbol.owner))
    typedTree.changeOwner((typedLambda.symbol.owner, currentOwner))//symbol))
    //println(" to", symbol)
    //println("lambda symbol: ", tree.attachments.get[Symbol])

    //println("symbols: ",fun.symbol.owner, fun.symbol.owner.owner, tree.symbol.owner, tree.symbol.owner.owner.moduleClass)
    //println(typedTree)
    typedTree*/
  }


  /* Why not call deltaImpl for each implementation?. In case an ascription would do more than one thing depending on the domain */
  def delta(tree: Tree): Effect = {
    val granted = getStrictPrivileges(getEffect(tree))
    /*if(tree.symbol !=null && tree.symbol.toString().contains("functionThatReads"))
    //println("delta of ", tree, granted)*/
    val needed: Effect = neededEffect(tree).map{ eff =>
      getStrictPrivileges(eff)
    }.getOrElse(bottom)
    /* the deltas of the parent plus the one of the domain */
    diff(needed, granted)
  }
  /*
  def delta(tree: Tree): Set[privilege] = {
    //TODO: simular app
    Set[privilege]()
  }*/

  def isE1Unk(e1ApplySym: Symbol, E_1: Effect, E_2: Effect, e2ListType: List[List[Type]], e1FinalType: Type): Boolean ={
    if(e1ApplySym!=null && e1ApplySym.exists){

    val paramsE1 = e1ApplySym.typeSignatureIn(e1FinalType).paramss
    val e2ParamTypes = e2ListType.flatten

    if(e2ParamTypes.size<=0) false
    else
    E_1.unknown ||  E_2.unknown ||
    paramsE1.flatten.zipWithIndex.foldLeft(false)((isUnk,pair) => {
      val (symbol, index) = pair
      val e1Sym = symbol.tpe.member(newTermName("apply"))
      if(!e2ParamTypes.isDefinedAt(index)) //implicit or default args
        true
      else{
        val e2Sym = e2ParamTypes(index).finalResultType.member(newTermName("apply"))
        val E1 = fromAnnotationList(e1Sym.tpe.finalResultType.annotations)
        val E2 = fromAnnotationList(e2Sym.tpe.finalResultType.annotations)
        if(isFunctionX(symbol, symbol.tpe))
          isUnk || isE1Unk(e1Sym, E1, E2, e2Sym.paramss.map{l => l.map{ll => ll.tpe}}, symbol.tpe.finalResultType)
        else isUnk
      }
    })
    } else false
  }

  var transformed =  Map[Tree, Boolean]()
  var dontCheck =  Map[Tree, Boolean]()
  var dontAdjust =  Map[Tree, Boolean]()
  var lambdaEffectForce =  Map[Tree, Effect]()
  var symbol: Symbol = null
  def insertHas(tree: Tree, currentOwner: Symbol): Tree = {
    //if(tree.toString().contains("foo"))
    //println(transformed, tree)
    //if(!transformed.getOrElse(tree, false)){
    /*tree match{
      case typed: Typed => println("found tree typed", tree, isTransformed(tree))
      case _ =>
    }*/
    if(!isTransformed(tree)){

      /** insert restrict
        * We can not modify the subtree of "tree" because when we re-typecheck it, it fails in all the runtimePrivileges code
        * Beause of that, we do that delaying the transformation of the tree to when "tree" is actually the tree to be modified
        * */
      tree match {
        case _: TypeDef | _: DefDef | _: ClassDef | _: ModuleDef | _: Function => /* its a lambda */
          if(tree.symbol.toString() contains "method main"){


            val DefDef(mods, name, tparams, vparamss, tpt, rhs) = tree
            val validChildren = tree.children.filter{c => c match { case _ : ValDef => false; case _ => true}}
            val xi = validChildren.headOption.map{ child => getEffect(validChildren(0)) }.getOrElse(getEffect(tree))


            //println("lazing inserting restrict on", rhs.symbol, xi)
            insertRestrictLazy(rhs, xi)
          }
          //println("updating children", tree.symbol)
          symbol = tree.symbol
        case _ =>
      }


      /* cast for application */



      def buildCastedTree(tree: Tree) = {
        val treeInfo.Applied(fun, targs, argss) = tree
        //println("==============",tree, "=================")

        //println("---- ", fun, targs, argss, "------")
        //Debugger.identyfySymbol(tree)

        val e1 = fun/*tree match{
          case Apply(e_1,e_2) => e_1
          case TypeApply(e_1,e_2) => e_1
          case _ => fun
        }*/
        val e2 = argss

        val (unaryArgs, repeatedArgs) = splitRepeatedParamss(fun.symbol.paramss, argss)
        val E = getEffect(tree)
        val sE = getStrictPrivileges(E)

        /* Compute latent effects in case e1 has relative effects. For example e1 can have the effects pure(someFunction) => we need to compute te effects that might throw someFunction */
        //TODO: aplicar aca el update
        val E1An = fromAnnotation(fun.symbol.info) u computeLatentEffect(tree)
        //println("E1An:", E1An, "latent:",computeLatentEffect(tree))
        val E1 = updateEffect(tree, E1An, fun.symbol.info.finalResultType)


        //println("PRIMERO",fun.symbol.info.finalResultType, fromAnnotation(fun.symbol.info.finalResultType), computeLatentEffect(tree))
        //println("VEAMOS,", E1An, fun.symbol.info.finalResultType)

        //println("latent: de ",tree, computeLatentEffect(tree))


        val sE1 = getStrictPrivileges(E1)


        val E2 = (bottom /: unaryArgs.toList)((effects, arg) => {
          //printlnln(arg,"->", infersEffects(arg._2))
          join(effects, infersEffects(arg._2))
        })
        //println(tree)
        //println("E:",E, "E1:",E1, "E2:", E2)
        val sE2 = getStrictPrivileges(E2)

        val sE1_E = diff(sE1, sE)

        //println("calculating isE1Unk of tree", tree, e1, e2, e1.tpe)
        val isUnkRec = isE1Unk(e1.symbol, E1, E2, e2.map{l => l.map{ll => ll.tpe}}, e1.symbol.tpe.finalResultType)


        val unknown = E.unknown || E1.unknown || isUnkRec
        val params = fun.symbol.paramss.flatten



        //println("partiendo", e2, e2.map{l => l.map{ll => ll.tpe}})

        //println(e1.symbol, e1.symbol.tpe.finalResultType, isUnkRec)
        /*if one argument, and it need to check! */
        //printlnln("para entrar: ",tree, tree.symbol,unknown,sE1_E, E.unknown, params.size, argss.flatten.length)
        //println("E1:", E1, "E2:", E2, "E:", E)
        val cTree = if(params.size == argss.flatten.length && (unknown && (!lte(sE1_E, bottom) || E.unknown || E1.unknown || isUnkRec))){
          /*if(tree.toString contains "insert"){
            //println("E1", tree,  E1, E2, e2)
            e2.flatten.map{ ee =>
              //println(buildPath(ee.symbol))
            }
          }*/
          //println("\n"*4)
          //println("casting "+tree)
          //println("e1 "+e1+" "+e1.symbol.tpe)
          //println("e2 "+e2)
          unaryArgs.toList.map{arg =>
            //println("casting from "+fun.symbol.info+" into "+arg._1.info+" E="+E)
          }

          val rel = relFromAnnotation(fun.symbol.info)


          val latent = computeLatentEffect(tree)
          //println(" == typing Lambda", tree, "==")
          //println(Debugger.identyfySymbol(tree))
          //println(tree.symbol.isSynthetic)
          //println("inserting cast from ", E1, "to ", E)
          val typedLambda = insertCast(e1, E1, E, e2, tree, currentOwner, targs)

          def buildApply(select: Tree, argss: List[List[Tree]]): Tree = {
            val base = Apply(select, argss.headOption.getOrElse(List()))
            setTransformed(base)
            if(argss.length>0){
              (base /: argss.tail){(apply, args) =>
                val a = Apply(apply, args)
                setTransformed(a)
                a
              }
            } else base
          }

          //println("original tree" , tree)
          //println("new Tree", buildApply(Select(typedLambda, newTermName("apply")), e2))

          val typedTree = typer.typed(buildApply(Select(typedLambda, newTermName("apply")), e2))
          //println("--------------------")
          //println("typedTree: ",typedTree)
          val treeInfo.Applied(funx, targsx, argssx) = typedTree
          //println("decomposed: ", funx, targsx, argssx)
          //Debugger.identyfySymbol(typedTree)
          typedTree match{
            case Apply(s, _) => setTransformed(s)
          }
          setTransformed(typedTree)



          typedTree.updateAttachment[OriginalTree](OriginalTree(tree)) /* for blame */
          //println("typed Tree: ",typedTree)
          //println("TYPED TREE",typedTree, getEffect(tree))
          adjust(typedTree, getEffect(tree), getRelEffects(tree))

          //println(typedTree, getEffect(tree), getEffect(typedTree))
          //println("attachments:", typedTree.attachments.all)

          //println("typedTree", getEffect(tree))
          /* adjust the lambda so when typechecking won't throw new errors */
          //println("chantandole a los hijos estos efectos:", getEffect(tree))
          adjustChildren(typedTree, getEffect(tree), getRelEffects(tree))
          //print("changing owner from", typedLambda.symbol.owner)
          //typedTree.changeOwner((typedLambda.symbol.owner, typer.context.owner))
          //typedTree.changeOwner((typedLambda.symbol.owner, tree.symbol.owner))
          typedTree.changeOwner((typedLambda.symbol.owner, currentOwner))
          typedTree

        } else tree

        cTree
      }


      val sym = tree.symbol
      val castedTree = tree match {
        case _ : Apply  => buildCastedTree(tree)
        case (_: Ident | _: Select) if sym.isMethod  => buildCastedTree(tree)

        case _ => tree
      }


      /* actually insert the restrict by modifing the tree node */
      val restrictedTree = tree.attachments.get[RestrictLazy].map{newPrivileges =>
        //println("inserting restrict in ", castedTree, newPrivileges)
        tree.removeAttachment[RestrictLazy]
        insertRestrictTree(castedTree, newPrivileges, tree)
      }.getOrElse{castedTree}




      /** insert the has! **/
      val deltas: Effect = tree match{
        /*C-Eff => f() : @alloc @unknown */
        case typed: Typed =>

          val (annots, expr) = collectAnnots(typed)
          val hasEffAnnots = existsEffectAnnotation(annots)
          //printlnln("is typed", tree, expr, hasEffAnnots, annots)
          if (hasEffAnnots) {
            //printlnln("calculando new Expected")
            val newExpected = fromAnnotationList(annots)

            //printlnln("terminado")

            val E1 = getStrictPrivileges(newExpected)
            val E = getStrictPrivileges(getEffect(tree))

            /*println("del delta saco que E1:", E1)
            //println("y que E es ", getEffect(tree), " y el estricto es ", E)
            //println("la diferencia es ", diff(E1, E))
            //println("inserting lazy for ", expr, "to ", newExpected)

            //println("dont adjust=?", dontAdjust.getOrElse(tree, false))*/

            /*typedDontAdjust(typed)

            if(!dontAdjust.getOrElse(tree, false))
              adjustChildren(tree, newExpected, rel)
            else
              adjustChildren(tree)
            */
            if(!dontAdjust.getOrElse(tree, false))
              insertRestrictLazy(expr, newExpected) /* do transform the tree, because we are not inserting this restrict in that tree */

            //printlnln("settung OT", expr, tree)
            expr.updateAttachment[OriginalTree](OriginalTree(tree))
            diff(E1, E)
          }
          else delta(tree)
        case _ => delta(tree)
      }


      if(!lte(deltas, bottom)){ /* if deltas is not bottom */
        //println("transformed", deltas)
        setTransformed(tree)
        //transformed += (tree -> true)
        insertHasTree(restrictedTree, deltas)
      } else restrictedTree
    } else {
      val adjustedTree = tree.attachments.get[HasLazy].map{hasLazy =>
        insertHasTree(tree, hasLazy.phi)
      }.getOrElse{tree}

      val restrictedTree = tree.attachments.get[RestrictLazy].map{newPrivileges =>
        //println("inserting restrict un tree quizas donde no deberia", tree.symbol, newPrivileges)
        val newTree = insertRestrictTree(adjustedTree, newPrivileges, tree)
        //println(newTree)
        /*remove the RestrictLazy here because adjustedTree doesn't have the same information than tree */
        tree.removeAttachment[RestrictLazy]
        newTree

      }.getOrElse{adjustedTree}

      restrictedTree
    }
  }


  def updateDontAdjust(tree: Tree, bool: Boolean = true) = {
    dontAdjust += (tree -> true)
  }
  /*mark all the tree's as transformed recursively until the `until` tree */
  def setTransformedUntil(tree: Tree, until: Tree): Unit  = {
    tree match{
      case `until` =>
      case _ => {
        setTransformed(tree)
        //transformed += (tree -> true)
        tree.children.map{ c => setTransformedUntil(c, until)}
      }
    }
  }

  val mode = 0 /* 0: classes; 1: bit vectors; 2: bit vectors + blameless*/
  def insertHasTree(tree: Tree, phi: Effect): Tree = {
    if(!lte(phi,bottom)){ /* if it's NOT bottom */
      //println("inserting has tree in ", tree, tree.pos)
      val originalTree = tree.attachments.get[OriginalTree].getOrElse(OriginalTree(tree)).tree


      tree.removeAttachment[HasLazy]
      val transformedDeltas = getStrictPrivilegesTree(phi).asInstanceOf[List[EffectDomain.this.global.Tree]]

      val newCode = mode match{
        case 0 => Apply(Select(Ident(reify{runtimePrivileges.RuntimePrivileges}.tree.symbol), newTermName("has")),
        List(Apply(Select(Select(Ident(reify{scala.Predef}.tree.symbol), newTermName("Set")), newTermName("apply")),
          transformedDeltas
        ), Literal(Constant(originalTree.toString()))
         , Literal(Constant(originalTree.pos.source.toString()))
         , Literal(Constant(originalTree.pos.line))
        ))
        case 1 => Apply(Select(Ident(reify{runtimePrivileges.RuntimePrivileges}.tree.symbol), newTermName("hasSimple")),
        List(Literal(Constant(getStrictPrivilegesSimple(phi).asInstanceOf[Int]))
        , Literal(Constant(originalTree.toString()))
         , Literal(Constant(originalTree.pos.source.toString()))
         , Literal(Constant(originalTree.pos.line))
        ))
        case _ => Apply(Select(Ident(reify{runtimePrivileges.RuntimePrivileges}.tree.symbol), newTermName("hasNoBlame")),
        List(Literal(Constant(getStrictPrivilegesSimple(phi).asInstanceOf[Int]))
        ))
      }


      setTransformed(tree)
      //transformed += (tree -> true)

      val newTree = typer.typed(Block(List(newCode), tree))
      val xi = getEffect(tree)
      val rel = getRelEffects(tree)
      //println("adjusting (has)", tree, xi)
      setTransformedUntil(newTree, tree)
      adjustChildrenRec(newTree, xi, rel)
      adjust(newTree, xi, rel)
      //println(newTree.attachments.all)
      //println("adjusted ", newTree, getEffect(newTree))
      newTree
    }
    else tree
  }
  def insertRestrictTree(tree: Tree, newPrivileges: RestrictLazy, oTree: Tree): Tree = {
    /*val c =
      Apply(
        Apply(
          Select(
            Select(Ident(reify{runtimePrivileges.RuntimePrivileges}.tree.symbol), newTermName("dyn")),
            newTermName("withValue")),
          List(
            Apply(
              Select(Select(Ident(reify{scala.Predef}.tree.symbol), newTermName("Set")), newTermName("apply")),
              newPrivileges.l
            )
          )
        ),
        List(tree))*/


    val originalTree = tree.attachments.get[OriginalTree].getOrElse(OriginalTree(tree)).tree
    val c = mode match{
      case 0 => Apply(
        Apply(
          Select(
            Ident(reify{runtimePrivileges.RuntimePrivileges}.tree.symbol),
            newTermName("restrict")),
          List(
            Apply(
              Select(Select(Ident(reify{scala.Predef}.tree.symbol), newTermName("Set")), newTermName("apply")),
              newPrivileges.l
            ), Literal(Constant(originalTree.toString()))
            , Literal(Constant(originalTree.pos.source.toString()))
            , Literal(Constant(originalTree.pos.line))

          )
        ),
        List(tree))
      case 1 =>
        val simplePhi = oTree.attachments.get[RestrictLazySimple].map{rzs => rzs.l}.getOrElse(0)
        Apply(
          Apply(
            Select(
              Ident(reify{runtimePrivileges.RuntimePrivileges}.tree.symbol),
              newTermName("restrictSimple")),
            List(
              Literal(Constant(simplePhi)),
              Literal(Constant(originalTree.toString()))
              , Literal(Constant(originalTree.pos.source.toString()))
              , Literal(Constant(originalTree.pos.line))

            )
          ),
          List(tree))
      case _ =>
      val simplePhi = oTree.attachments.get[RestrictLazySimple].map{rzs => rzs.l}.getOrElse(0)
      Apply(
        Apply(
          Select(
            Ident(reify{runtimePrivileges.RuntimePrivileges}.tree.symbol),
            newTermName("restrictNoBlame")),
          List(
            Literal(Constant(simplePhi))


          )
        ),
        List(tree))
    }




    setTransformed(tree)
    //transformed += (c -> true)

    tree.removeAttachment[RestrictLazy]
    tree.removeAttachment[RestrictLazySimple]


    val newTree = typer.typed(c)
    //printlnln("original tree", originalTree)
    //printlnln("new restricted Tree:", newTree)
    setTransformedUntil(newTree, tree)
    //println("adjusting (restrict)", tree, getEffect(tree))
    val xi = getEffect(tree)
    val rel = getRelEffects(tree)
    adjustChildrenRec(newTree, xi, rel)
    adjust(newTree, xi, rel)

    newTree.updateAttachment[OriginalTree](OriginalTree(tree))

    newTree
  }
  case class HasLazy(phi: Effect)
  def insertHasLazy(tree: Tree, phi: Effect): Tree = {
    //transformed += (tree -> true)
    if(!lte(phi,bottom)){
      //println("inserting has lazy", tree, phi)
      tree.updateAttachment[HasLazy](HasLazy(phi))
    }
    else tree
  }
  case class RestrictLazy(l: List[Tree])
  case class RestrictLazySimple(l: Int)


  def insertRestrictLazy(tree: Tree, eff: Effect): Tree = {
    /* if unknown in eff, then restrict shouldn't do anything*/
    if(!eff.unknown){
      //transformed += (tree -> true)
      val phi = getStrictPrivileges(eff)
      val newPrivileges = getStrictPrivilegesTree(phi).asInstanceOf[List[EffectDomain.this.global.Tree]]
      val newPrivilegesSimple = getStrictPrivilegesSimple(phi).asInstanceOf[Int]
      //printlnln("FOR REAL inserting restrict lazy",tree, newPrivileges)
      tree.updateAttachment[RestrictLazy](RestrictLazy(newPrivileges))
      tree.updateAttachment[RestrictLazySimple](RestrictLazySimple(newPrivilegesSimple))
    } else tree
  }

  case class TransformedTree(transformed: Boolean)
  def setTransformed(tree: Tree): Tree = {
    tree.updateAttachment[TransformedTree](TransformedTree(true))
  }
  def setTransformedRec(tree: Tree): Unit = {
    setTransformed(tree)
    tree.children.map{setTransformedRec(_)}
  }
  def isTransformed(tree: Tree): Boolean = {
    tree.attachments.get[TransformedTree].map{ t => t.transformed }.getOrElse(false)
  }



  /**
   * True there exists some effect annotations in the list `annotations`.
   */
  final def existsEffectAnnotation(annotations: List[AnnotationInfo]) = {
    annotations.exists(ann => allEffectAnnots.contains(ann.atp.typeSymbol))
  }
  final lazy val pureClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.pure"))

  final lazy val unknownClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.unknown"))

  final lazy val allEffectAnnots = pureClass :: unknownClass :: annotationClasses


  final def fromAnnotation(tpe: Type): Effect = fromAnnotationList(tpe.finalResultType.annotations)
  /**
   * The effect represented by `annotations`
   */
  final def fromAnnotationList(annots: List[AnnotationInfo]) = {
    val hasPure = annots.exists(_.atp.typeSymbol == pureClass)
    val hasUnknown = annots.exists(_.atp.typeSymbol == unknownClass)
    //println(hasUnknown)
    /*TODO: definir cual es el estado por defecto si no viene anotaciones (aca es top */
    val default = if(hasUnknown) unknown else{ if (hasPure) effectForPureAnnotated else top }
    //println(default)
    parseAnnotationInfos(annots, default)
  }

  /**
   * Read the effect from a list of annotations. If no effect annotation for this domain
   * can be found, this method should return the `default` effect.
   *
   * This method does not need to handle `@pure` annotations: this is already done by
   * `fromAnnotations`.
   */
  def parseAnnotationInfos(annots: List[AnnotationInfo], default: => Effect): Effect



  def buildPath(sym: Symbol): String = sym.fullName
  def isUncheckedFun(sym: Symbol, l: List[String]) = {
    val path = buildPath(sym)
    //println(tree, path, l)
    (false /: l)((unchecked, s) => {
      unchecked || s.r.findAllIn(path).hasNext
    })
  }

  def buildPath(tree: Tree): String = {
    val sym = tree.symbol
    //println("enclosing package:",sym.enclosingPackage, sym.enclosingPackage.isEmptyPackage)
    val path = tree match{
      case Select(qualifier, name) if qualifier.symbol != null => buildPackagePath(sym)+sym.enclClass.nameString+"."+name.decoded //println("q:",qualifier, qualifier.symbol.enclosingPackage, sym.enclClass.nameString)
      case Select(New(tpt), name) => tpt.toString+"."+name

      case Ident(name) if sym != null => buildPackagePath(sym)+sym.enclClass.nameString+"."+name.toString
      case DefDef(mods, name, tparams, vparamss, tpt, rhs) if sym != null => buildPackagePath(sym)+sym.enclClass.nameString+"."+name.toString
      case TypeApply(fun, args) =>  buildPackagePath(sym)+sym.enclClass.nameString+"."+fun
      case _ => "<error>"
    }
    path
  }
  def buildPackagePath(sym: Symbol, s: String = ""): String = {
    if(sym != null){
      val pack = sym.enclosingPackage
      if(pack.exists){
        buildPackagePath(pack, pack.nameString+"."+s)
      } else s
    } else s
  }
  def isUncheckedFun_(tree: Tree, l: List[String]) = {
    val path = buildPath(tree)
    //println(tree, path, l)
    (false /: l)((unchecked, s) => {
      unchecked || s.r.findAllIn(path).hasNext
    })
  }

  object Debugger{
    def identyfySymbol(tree: Tree, indent: Int = 0): Unit = {
      print("  "*indent)
      tree match{
        case Ident(name) =>
          println("Ident(name)")
        case Select(qualifier, selector) =>
          println("Select(qualifier, selector)")
          identyfySymbol(qualifier, indent+1)
        case Apply(fun, args) =>
          println("Apply(fun, args)")
          identyfySymbol(fun, indent+1)
          args.map{identyfySymbol(_, indent+1)}
        case TypeTree() =>
          println("TypeTree()")
        case Literal(value) =>
          println("Literal(value)")
        case This(qual) =>
          println("This(qual)")
        case ValDef(mods, name, tpt, rhs) =>
          println("ValDef(mods, name, tpt, rhs)")
        case DefDef(mods, name, tparams, vparamss, tpt, rhs) =>
          println("DefDef(mods, name, tparams, vparamss, tpt, rhs)")
        case Block(stats, expr) =>
          println("Block(stats, expr)")
        case If(cond, thenp, elsep) =>
          println("If(cond, thenp, elsep)")
        case CaseDef(pat, guard, body) =>
          println("CaseDef(pat, guard, body)")
        case TypeApply(fun, args) =>
          println("TypeApply(fun, args)")

          identyfySymbol(fun, indent+1)
          args.map{identyfySymbol(_, indent+1)}

        case AppliedTypeTree(tpt, args) =>
          println("AppliedTypeTree(tpt, args)")
        case Bind(name, body) =>
          println("Bind(name, body)")
        case Function(vparams, body) =>
          println("Function(vparams, body)")
        case Match(selector, cases) =>
          println("Match(selector, cases)")
        case New(tpt) =>
          println("New(tpt)")
        case Assign(lhs, rhs) =>
          println("Assign(lhs, rhs)")
        case AssignOrNamedArg(lhs, rhs) =>
          println("AssignOrNamedArg(lhs, rhs)")
        case Try(block, catches, finalizer) =>
          println("Try(block, catches, finalizer)")
        case EmptyTree =>
          println("EmptyTree")
        case Throw(expr) =>
          println("Throw(expr)")
        case Super(qual, mix) =>
          println("Super(qual, mix)")
        case TypeBoundsTree(lo, hi) =>
          println("TypeBoundsTree(lo, hi)")
        case Typed(expr, tpt) =>
          println("Typed(expr, tpt)")
        case Import(expr, selectors) =>
          println("Import(expr, selectors)")
        case Template(parents, self, body) =>
          println("Template(parents, self, body)")
        case ClassDef(mods, name, tparams, impl) =>
          println("ClassDef(mods, name, tparams, impl)")
        case ModuleDef(mods, name, impl) =>
          println("ModuleDef(mods, name, impl)")
        case TypeDef(mods, name, tparams, rhs) =>
          println("TypeDef(mods, name, tparams, rhs)")
        case LabelDef(name, params, rhs) =>
          println("LabelDef(name, params, rhs)")
        case PackageDef(pid, stats) =>
          println("PackageDef(pid, stats)")
        case Annotated(annot, arg) =>
          println("Annotated(annot, arg)")
        case SingletonTypeTree(ref) =>
          println("SingletonTypeTree(ref)")
        case SelectFromTypeTree(qualifier, selector) =>
          println("SelectFromTypeTree(qualifier, selector)")
        case CompoundTypeTree(templ) =>
          println("CompoundTypeTree(templ)")
        case ExistentialTypeTree(tpt, whereClauses) =>
          println("ExistentialTypeTree(tpt, whereClauses)")
        case Return(expr) =>
          println("Return(expr)")
        case Alternative(trees) =>
          println("Alternative(trees)")
        case Star(elem) =>
          println("Star(elem)")
        case UnApply(fun, args) =>
          println("UnApply(fun, args)")
        case ArrayValue(elemtpt, trees) =>
          println("ArrayValue(elemtpt, trees)")
        case ApplyDynamic(qual, args) =>
          println("ApplyDynamic(qual, args)")
        case ReferenceToBoxed(idt) =>
          println("ReferenceToBoxed(idt)")
        case _ => println("unknown")
      }
    }
  }
}

trait WithUnknownEffect{
  val unknown: Boolean = false

}

trait EffectLattice {
  val global: Global
  import global._

  type Effect <: WithUnknownEffect
  //type privilege

  implicit val ctEvidence: ClassTag[Effect]

  def top: Effect
  def bottom: Effect

  def unknown: Effect

  /**
   * This effect is used by the framework for methods that have a `@pure` annotation.
   * By default, pure methods have the `bottom` effect. Effect domains can override
   * this method to assign a different default effect to `@pure` methods.
   */
  def effectForPureAnnotated: Effect = bottom

  def join(a: Effect, b: Effect): Effect
  def meet(a: Effect, b: Effect): Effect
  def lte(a: Effect, b: Effect): Boolean
  /* Usable for static effects */
  def diff(a: Effect, b: Effect): Effect

  def joinUnknown(a: Effect, b: Effect): Boolean = a.unknown || b.unknown
  def meetUnknown(a: Effect, b: Effect): Boolean = a.unknown && b.unknown
  def diffUnknown(a: Effect, b: Effect): Boolean = !b.unknown && a.unknown
  def lteUnknown(a: Effect, b: Effect): Boolean = {
    if(a.unknown || b.unknown) true
    else false
  }

  final def joinAll(effs: Effect*): Effect = {
    if (effs.isEmpty) bottom
    else {
      var acc = bottom
      for (e <- effs) { acc = join(acc, e) }
      acc
    }
  }

  def default: Effect
  implicit class EffectOps(eff: Effect) {
    def u(other: Effect): Effect = join(eff, other)
    def n(other: Effect): Effect = meet(eff, other)
    def <=(other: Effect): Boolean = lte(eff, other)
  }
  //def check(tree: Tree, eff: Effect): Boolean

  def getStrictPrivileges(eff: Effect): Effect

  def getStrictPrivilegesTree(eff: Effect): List[scala.tools.nsc.Global#Tree]


  type OverrideEffect <: {def e: Effect}
  implicit val ctEvidenceOverrideEffect: ClassTag[OverrideEffect]
  def constructNeededEffect(e: Effect): OverrideEffect

  def getStrictPrivilegesSimple(eff: Effect): Int = 0
  
}
