package effectchecker

import scala.annotation.effects._

import scala.tools.nsc.Global
import scala.tools.nsc.plugins.Plugin

import scala.tools.nsc
import nsc.plugins.PluginComponent
import nsc.transform.Transform
import nsc.transform.InfoTransform
import nsc.transform.TypingTransformers
import nsc.ast.TreeDSL
import nsc.symtab.Flags._
import nsc.typechecker



class EffectsCheckerPlugin(val global: Global) extends Plugin{
  /** The name of this plugin. */
  val name = "effects-checker-plugin"

  /** A short description of the plugin */
  val description = "tracks side effects"

  //println("Effects Checker Plugin")

  /*val typerPlugin = new {
    val global: EffectsCheckerPlugin.this.global.type = EffectsCheckerPlugin.this.global
  } with TypeCheckerPlugin
  */
  /**
   * The compiler components that will be applied when running this plugin
   */
  //val components = Nil // List(cleanupPhase)

  val components = new EffectChecker(this, global) :: Nil
}
class EffectChecker(plugin:Plugin, val global:Global) extends PluginComponent
  with Transform with TypingTransformers with TreeDSL {
    import global._
    import analyzer.{AnalyzerPlugin, Typer, AbsTypeError, SilentResultValue, SilentTypeError}
    import analyzer.ErrorUtils.{issueNormalTypeError, issueTypeError}

    //println("AnnotationRewriter")

    val runsAfter = "typer" :: Nil
    val phaseName = "effect-checker"

    def newTransformer(unit:CompilationUnit) = new MyTransformer(unit)


    def mkDomains(names: List[String]): List[EffectDomain { val global: EffectChecker.this.global.type }] = names match {
      /*case "io" :: xs =>
        new io.IODomain {
          val global: EffectChecker.this.global.type = EffectChecker.this.global
        } :: mkDomains(xs)
      */
      case "exceptions" :: xs =>
        new exceptions.ExceptionsDomain {
          val global: EffectChecker.this.global.type = EffectChecker.this.global
          override val uncheckedFunctions = settings.unchecks
        } :: mkDomains(xs)
      
      /*case "purity" :: xs =>
        new purity.PurityDomain {
          val global: EffectChecker.this.global.type = EffectChecker.this.global
        } :: mkDomains(xs)*/

      case "ref" :: xs =>
        new ref.RefDomain {
          val global: EffectChecker.this.global.type = EffectChecker.this.global
          override val uncheckedFunctions = settings.unchecks
        } :: mkDomains(xs)

      case "example" :: xs =>
        new example.ExampleDomain {
          val global: EffectChecker.this.global.type = EffectChecker.this.global
          override val uncheckedFunctions = settings.unchecks
        } :: mkDomains(xs)
        
      case "myIO" :: xs =>
        new myio.MyIODomain {
          val global: EffectChecker.this.global.type = EffectChecker.this.global
          override val uncheckedFunctions = settings.unchecks
        } :: mkDomains(xs)
      case "DBAccess" :: xs =>
        new dbaccess.DBAccessDomain {
          val global: EffectChecker.this.global.type = EffectChecker.this.global
          override val uncheckedFunctions = settings.unchecks
        } :: mkDomains(xs)
      case "simpleIO" :: xs =>
        new simpleio.SimpleIODomain {
          val global: EffectChecker.this.global.type = EffectChecker.this.global
          override val uncheckedFunctions = settings.unchecks
        } :: mkDomains(xs)
      case "simpleSEIO" :: xs =>
        new simpleseio.SimpleSEIODomain {
          val global: EffectChecker.this.global.type = EffectChecker.this.global
          override val uncheckedFunctions = settings.unchecks
        } :: mkDomains(xs)
      case "simpleTPIO" :: xs =>
        new simpletpio.SimpleTPIODomain {
          val global: EffectChecker.this.global.type = EffectChecker.this.global
          override val uncheckedFunctions = settings.unchecks
        } :: mkDomains(xs)
      case x :: xs =>
        mkDomains(xs)
      //global.abort(s"Unknown effect domain: $x")

      case Nil => Nil

    }
    val domain = {
      mkDomains(settings.domains) match {
        case Nil =>
          global.abort(s"No effect domain specified")

        case x :: xs =>
          (x /: xs) {
            case (domain1, domain2) => new {
              val global: EffectChecker.this.global.type = EffectChecker.this.global
              override val d1 = domain1
              override val d2 = domain2
            } with BiEffectDomain
          }
      }
    }



    class MyTransformer(unit:CompilationUnit) extends TypingTransformer(unit){

      /*val domain = new {
        val global: EffectChecker.this.global.type = EffectChecker.this.global
        override val d1 = new example.ExampleDomain{val global: EffectChecker.this.global.type = EffectChecker.this.global}
        override val d2 = new ref.RefDomain{val global: EffectChecker.this.global.type = EffectChecker.this.global}
      } with BiEffectDomain*/
      val defaultXi = domain.lattice.default


      override def transform(tree: Tree): Tree = {
        //println("---------------------", tree, tree.tpe)
        //Debugger.identyfySymbol(tree)

        super.transform(checkEffect(tree))
      }


      case class Synthetic()
      def setSynthetic(tree: Tree): Unit = {
        tree.updateAttachment[Synthetic](Synthetic())
        tree.children.map{ child =>
          setSynthetic(child)
        }
      }

      var transformed =  Map[Tree, Boolean]()
      var yes = false
      def checkEffect(tree: Tree): Tree = {
        if(!yes){
          //println(tree)
          yes = true
        }

        domain.setParent(tree)
        //println("checkEffect", tree)
        //if the tree is synthetic, we dont want to check or adjust it. It is never annotated!

        //val localXi = Xi.get(tree).getOrElse(defaultXi)
        //val localXi = tree.attachments.get[domain.lattice.Effect].getOrElse(defaultXi)
        domain.preAdjust(tree) //for annotations
        if(!domain.dontCheck.getOrElse(tree, false)){
          val errors = domain.check(tree)
          errors.map{ error =>
            val pos = tree.attachments.get[domain.OriginalTree].map{ ot => ot.tree.pos}.getOrElse(tree.pos)
            unit.error(pos, error)
          }
        }
        else{
          //println("NOT CHECKING ", tree)
        }
        //println("adjusting")
        domain.adjust(tree)
        //println("adjusted")
        //println("hassing")
        val hassed = domain.insertHas(tree, currentOwner)
        //println("hassed")
        hassed
      }


    }
    object settings {
        var domains: List[String] = Nil
        var unchecks: List[String] = Nil
        val domainsStr = "domains:"
        val unchecksStr = "unchecks:"


        process(globalOptions, s => global.abort(s))

        /**
         * Read the options for this compiler plugin - see documentation on `processOptions`
         * above.
         */
        def globalOptions = {
          val namec = "effects:"
          val opts = global.settings.pluginOptions.value.filter(_ startsWith namec)
          opts.map(_ stripPrefix namec)
        }

        def process(opts: List[String], error: String => Unit) {
          opts match {
            case x :: xs if x.startsWith(domainsStr) =>
              if (domains.nonEmpty) {
                error("Multiple 'domain' settings")
              } else {
                domains = x.drop(domainsStr.length).split(":").toList
              }
              process(xs, error)
            case x :: xs if x.startsWith(unchecksStr) =>
              if (unchecks.nonEmpty) {
                error("Multiple 'uncheck' settings")
              } else {
                unchecks = x.drop(unchecksStr.length).split(":").toList.map{_.replace(".","\\.").replace("*", ".*")}
              }
              process(xs, error)
            case x :: xs =>
              error(s"Unknown setting: $x")

            case Nil =>
              ()
          }
        }
      }

}

