package effectchecker



trait Infer { self: EffectDomain =>
  import global._
  import lattice._

  case class RelEffectList(rel: List[RelEffectE])

  def isPolymorphic(sym: Symbol) = {
    //println("\n")
    //println("Is Polymorphic???", sym)
    //println("the type is ", sym.tpe.dealias)
    val fun = sym
    val paramss = fun.tpe.paramss
    //println("paramss ", sym.tpe.paramss)

    val rels = relEffects(fun).map{ _.param }

    val args = paramss.map(v => v.map( p => {
      rels contains ParamLocE(p)
    }))



    //println("latents: ",latent(fun, Map(), Map(), RelEffectList(List())))
    //println("relEffs: ")
    //println("ispoly?", args)
    //val paramLocs = paramss.map(pa => ParamLocE(pa._1))
    //println("\n")
    args
  }
  def computeLatentEffect(tree: Tree) = {
    val treeInfo.Applied(fun, _, argss) = tree
    val funSym = fun.symbol
    val (unaryArgs, repeatedArgs) = splitRepeatedParamss(funSym.paramss, argss)
    val lat = {
      if (hasRelativeEffect(fun)) bottom
      else {
        // for non-repeated parameters, we remember the argument type and location in a map and pass it to `latent`.
        // if a parameter has a relative effect, the argument type and location define its expansion.
        val paramLocs = unaryArgs.map(pa => ParamLocE(pa._1))
        //println("paramLocs = ", paramLocs)
        val thisLocMapping = fun match {
          case Select(qual, _) => Some(ThisLoc(funSym.owner) -> argTpeAndLoc(qual))
          case _ => None
        }
        val rels = getRelEffects(tree)
        //println("thisLocMapping = ", thisLocMapping)
        //println("latent args = ", funSym, paramLocs.zip(unaryArgs.map(pa => argTpeAndLoc(pa._2))).toMap ++ thisLocMapping, rels)
        val latEff = latent(funSym, paramLocs.zip(unaryArgs.map(pa => argTpeAndLoc(pa._2))).toMap ++ thisLocMapping, Map[Symbol, Effect](), rels)
        //println("el latnente es ", funSym, funSym.info, latEff)
        updateEffect(tree, latEff, tree.symbol.info.resultType)
      }
    }
    lat
  }

  /**
   * Returns the type and location of a tree which is passed to a method as argument.
   */
  private def argTpeAndLoc(arg: Tree): (Type, Option[LocE]) = {
    val tp = arg.tpe
    arg match {
      case Ident(_) if arg.symbol.owner.isMethod =>
        (tp, Some(ParamLocE(arg.symbol)))
      case This(_) =>
        (tp, Some(ThisLoc(arg.symbol)))
      case _ =>
        (tp, None)
    }
  }
  /**
   * Returns
   *   - a list of non-repeated parameter symbols
   *   - the corresponding list of arguments
   *   - a list of arguments for repeated parameters
   */
  def splitRepeatedParamss(paramss: List[List[Symbol]], argss: List[List[Tree]]): (Map[Symbol, Tree], Map[Symbol, List[Tree]]) = (paramss, argss) match {
    case (params :: paramss, args :: argss) =>
      val (unaryPairs, repeatedPairs) = params.zipAll(args, NoSymbol, EmptyTree).span(p => !definitions.isRepeated(p._1))

      val (unaryArgs, repeatedArgs) = splitRepeatedParamss(paramss, argss)
      val repeatedArg: Option[(Symbol, List[Tree])] = repeatedPairs match {
        case (sym, arg) :: ps => Some((sym, arg :: ps.map(_._2)))
        case _ => None
      }
      (unaryArgs ++ unaryPairs, repeatedArgs ++ repeatedArg)

    //      val (unaryParams, unaryArgs, repeatedArgs) = splitRepeatedParamss(paramss, argss)
    //      (unaryPairs.map(_._1) ::: unaryParams, unaryPairs.map(_._2) ::: unaryArgs, repeatedPairs.map(_._2) ::: repeatedArgs)

    case _ =>
      (Map(), Map())
  }

  /**
   * True if the method `fun` exists in the relative effect environment in `ctx`, i.e. if it is
   * a method call on a parameter of an enclosing method, and that method is effect-polymorphic in `fun`.
   */
  def hasRelativeEffect(fun: Tree): Boolean = {
    val sym = fun.symbol
    val relEffects = getRelEffects(fun)
    fun match {
      case Select(id @ Ident(_), _) =>
        val r = RelEffectE(ParamLocE(id.symbol), Some(sym))
        lteRel(List(r), relEffects.rel)

      case Select(th @ This(_), _) =>
        val r = RelEffectE(ThisLoc(th.symbol), Some(sym))
        lteRel(List(r), relEffects.rel)

      case _ => false
    }
  }

  /**
   * The latent effect of `fun`, i.e. the maximal effect that might occur when invoking `fun`.
   *
   * @TODO: fixpoint computation needed (rel effects can go in circles!)?
   * @TODO: document what happens here (in general, argtps map specifically)
   *
   * @param argtps A map from the parameters of `fun` to the actual argument types which are passed
   *               in the invocation. These types might be more specific than the parameter types of
   *               `fun`, which implements effect-polymorphism.
   * @param rel    If `fun` has itself relative effects, but those relative effects are also in the
   *               environment ctx.relEnv (i.e. the enclosing function has the same relative effect), that
   *               relative effect can be ignored.
   */
  def latent(fun: Symbol, argtps: Map[LocE, (Type, Option[LocE])],
                     byNameEffs: Map[Symbol, Effect], rel: RelEffectList): Effect = {
    defaultInvocationEffect(fun).getOrElse {
      val concrete = fromAnnotation(fun.info)
      val relEff = relEffects(fun)
      //println("concrete = ", concrete, fun, fun.info, fun.info.finalResultType.annotations)
      //println("relEff = ", relEff)

      val expandedRelEff = relEff map { r =>
        if (lteRelOne(r, rel.rel)) bottom
        else r match {
          case RelEffectE(ParamLocE(param), None) if param.isByNameParam =>
            byNameEffs.getOrElse(param, top)

          case RelEffectE(paramLoc, funOpt) if (argtps contains paramLoc) =>
            argtps(paramLoc) match {
              case (tp, Some(argLoc)) if lteRelOne(RelEffectE(argLoc, funOpt), rel.rel) =>
                // @TODO: document, example. if a parameter is forwarded to another method as parameter, detect rel effects
                bottom
              case (tp, _) =>
                val funSyms = funOpt match {
                  case Some(fun) =>
                    List(tp.member(fun.name).suchThat(m => m.overriddenSymbol(fun.owner) == fun || m == fun))
                  case None =>
                    tp.member(nme.apply).alternatives
                }
                //println("111", funSyms)
                val funEffs = funSyms.map(latent(_, Map(), Map(), rel))
                joinAll(funEffs: _*)
            }

          case RelEffectE(_, Some(fun)) =>
            latent(fun, Map(), Map(), rel)

          case RelEffectE(_, None) =>
            val applyEffs = r.applyMethods.map(latent(_, Map(), Map(), rel))
            joinAll(applyEffs: _*)
        }
      }
      concrete u (expandedRelEff :\ bottom)(_ u _)
    }
  }


  final class FixedUnderlyingSingleType(pre: Type, sym: Symbol, override val underlying: Type) extends SingleType(pre, sym)

  /**
   * Remove annotations with a type in `cls` from the type `tp`.
   */
  private def removeAnnotations(tp: Type, cls: List[Symbol]): Type = tp match {
    case AnnotatedType(annots, underlying, _) =>
      underlying.withAnnotations(annots.filterNot(ann => cls.contains(ann.atp.typeSymbol)))

    case st @ SingleType(pre, sym) =>
      val underlying = st.underlying
      val cleanUnderlying = removeAnnotations(underlying, cls)
      if (cleanUnderlying eq underlying) st
      else new FixedUnderlyingSingleType(pre, sym, cleanUnderlying)

    case ExistentialType(quantified, underlying) =>
      val cleanUnderlying = removeAnnotations(underlying, cls)
      if (cleanUnderlying eq underlying) tp
      else ExistentialType(quantified, cleanUnderlying)

    case tp => tp
  }



  /**
   * Apply operation `op` on the result type of (method) type `tp`.
   */
  private def transformResultType(tp: Type, op: Type => Type): Type = tp match {
    case MethodType(args, res) => copyMethodType(tp, args, transformResultType(res, op))
    case PolyType(targs, res) => PolyType(targs, transformResultType(res, op))
    case NullaryMethodType(res) => NullaryMethodType(transformResultType(res, op))
    case tp => op(tp)
  }

  /**
   * Remove all effect annotations from `tp`, or from its result type if it is a method type.
   */
  def removeAllEffectAnnotations(tp: Type) = {
    transformResultType(tp, removeAnnotations(_, allEffectAnnots))
  }

}
