package runtimePrivileges

import scala.tools.nsc.Global
import scala.util.DynamicVariable
import scala.annotation.effects._
import scala.reflect.internal.util.Position

import scala.reflect.runtime._
import scala.tools.reflect.ToolBox


object RuntimePrivileges {
  val cm = universe.runtimeMirror(getClass.getClassLoader)
  val tb = cm.mkToolBox()

  case class BlamePosition(originalTree: String, file: String, line: Int)

  case class EffectPrivilegeException(message: String = null, cause: Throwable = null) extends RuntimeException(message, cause)

  case class EffectPrivilegeLocalBlameException(missingEffs: Set[Effect], p1: BlamePosition, message: String = null, cause: Throwable = null) extends RuntimeException(message, cause)

  case class EffectPrivilegeOutermostBlameException(missingEffs: Set[Effect], p1: BlamePosition, blameLocal: BlamePosition, message: String = null, cause: Throwable = null) extends RuntimeException(message, cause)

  val dyn = new DynamicVariable[Set[Effect]](Set())
  val simpleDyn = new DynamicVariable[Int](0)

  var started = new DynamicVariable[Boolean](false)
  var initialSet = true

  val classNames = scala.collection.mutable.HashMap[String, String]()
  val classes = scala.collection.mutable.HashMap[String, Class[_]]()


  def hasSimple(effs: Int, originalTree: String, file: String, line: Int): Unit = {
    //println("checking ",originalTree, effs,"<",simpleDyn.value, (effs & simpleDyn.value))
    if((effs & simpleDyn.value) != effs){
      blame(Set(), BlamePosition(originalTree, file, line))
    }
  }

  def hasNoBlame(effs: Int): Unit = {
    //println("checking ", effs)
    if((effs & simpleDyn.value) != effs){
      throw new Exception("not enough privileges")
    }
  }

  def has(effs: Set[Effect], originalTree: String, file: String, line: Int): Unit = {
    //println("en "+originalTree+" tengo: ", dyn.value, "y estoy chequeando que tenga ", effs)
    val missings = missingEffects(effs)
    if(missings.size>0){

      blame(missings, BlamePosition(originalTree, file, line))
      //throwRuntimeException(missings, expr)
    }
  }
  def check(effs: Set[Effect]): Boolean = {
    (true /: effs)((cond, eff) =>
      cond && dyn.value.exists{ v => v.getClass.isAssignableFrom(eff.getClass)}
    )
  }
  def getClassByString(s: String): Class[_] = {
    classes.getOrElseUpdate(s, Class.forName(getClassNameByString(s)))
  }

  def getClassNameByString(s: String): String = {
    classNames.getOrElseUpdate(s, {
      val st = if(s=="Int") "Integer" else s
      //println("calculando de "+s)
      tb.eval(tb.parse("classOf["+st+"].getName")).asInstanceOf[String]
    })
    //"java.lang.String"
  }
  /*def lessThan(eff: EffectsWithGeneric, al: List[List[String]], bl: List[List[String]]): Boolean = {
    val lesses = (al zip bl zipWithIndex).map{case ((a, b), index) => 
      val ac: List[Class[_]] = a.map{ c => Class.forName(getClassNameByString(c))}
      val bc: List[Class[_]] = b.map{ c => Class.forName(getClassNameByString(c))}

      val v = (true /: ac)((cond, c) =>
        cond && bc.exists{ v => 
          eff.leqFunc(index, c, v) //c <: v
        }
      )
      println(ac, "<:",bc, v)
      v
    }
    (true /: lesses)((cond, c) => c && cond)
  }*/
  def missingEffects(effs: Set[Effect]): Set[Effect] = {
    val granted = dyn.value

    // check if effs <: granted
    effs.filter{ eff =>
      !granted.exists{ v =>
        //println(eff, "<:", v, eff.lte(v, getClassByString _))
        eff.lte(v, getClassByString _)
        /*val effLessThanV = eff.lte(v)//v.getClass.isAssignableFrom(eff.getClass)

        if(effLessThanV){
          //check args
          (eff, v) match{
            case (e @ EffectsWithGeneric(l1), EffectsWithGeneric(l2)) => lessThan(e, l1, l2)
            case _ => effLessThanV
          }
        } else false*/
      }
    }

    //effs.diff(dyn.value)
  }
  /*def restrict2[S](newval: Set[Effect])(thunk: => S): S = {
      dyn.withValue(newval){
        thunk
      }
  }*/
  def calculteNewRestrict(newval: Set[Effect]): Set[Effect] = {
    /*if(initialSet){
      initialSet = false
      newval
    } else{*/
      val granted = dyn.value
      newval.filter(eff => granted.exists{v => eff.lte(v, getClassByString _)})
    //}
  }
  def calculteNewRestrict(newval: Int): Int = {
   
      val granted = simpleDyn.value
      newval & granted
   
  }



  def restrictNoBlame[S](newval: Int)(thunk: => S): S = {
    if(!started.value){
      started.withValue(true){
        doRestrict(simpleDyn, newval, thunk)
      }
    } else{
      val fixedval = calculteNewRestrict(newval)
      doRestrict(simpleDyn, fixedval, thunk)
    }
  }

  def doRestrict[T,S](dd: DynamicVariable[T], priv: T, thunk: => S): S = {
    dd.withValue(priv){
        thunk
    }
  }

  def restrictSimple[S](newval: Int, originalTree: String, file: String, line: Int)(thunk: => S): S = {
    //println("restrict simple to ", newval)
    try{
      if(!started.value){
        started.withValue(true){
          doRestrict(simpleDyn, newval, thunk)
        }
      } else{
        val fixedval = calculteNewRestrict(newval)
        doRestrict(simpleDyn, fixedval, thunk)
      }
      
      
    } catch {
      case e: EffectPrivilegeLocalBlameException =>
        throw new Exception("not enough privileges")
      case e: EffectPrivilegeOutermostBlameException =>
        throw new Exception("not enough privileges")
    }
  }


  def restrict[S](newval: Set[Effect], originalTree: String, file: String, line: Int)(thunk: => S): S = {
    //println("restrict simple to ", newval)
    
    try{
      //must restrict to smaller sets
      if(!started.value){
        started.withValue(true){
          doRestrict(dyn, newval, thunk)
        }
      } else{
        val fixedval = calculteNewRestrict(newval)
        doRestrict(dyn, fixedval, thunk)
      }

      /*val fixedval = calculteNewRestrict(newval)
      //println("restringiendo "+originalTree+" dyn a "+fixedval)
      dyn.withValue(fixedval){
          thunk
      }*/
    } catch {
      case e: EffectPrivilegeLocalBlameException =>
        if(!containsAll(newval, e.missingEffs)){
          //local guilty
          //test for outermost too in case is the same restrict code
          val newE = EffectPrivilegeOutermostBlameException(e.missingEffs, e.p1, BlamePosition(originalTree,file,line))
          //println("soy el culpable local, mi padre tiene", dyn.value)
          if((containsAll(dyn.value, e.missingEffs) || dyn.value.size<=0)){
            throwEffectPrivilegeException(newE, newval, originalTree, file, line)
          } else {
            throw newE
          }
        }
        else{
          throw e //not guilty, lets move on
        }
      case e: EffectPrivilegeOutermostBlameException =>
        //println("captured, lets try")
        //we must ask for dyn.value <= 0 or the algorithm will fail if the initial restrict doesn't have the required privileges
        if((containsAll(dyn.value, e.missingEffs) || dyn.value.size<=0) && !containsAll(newval, e.missingEffs)){
          throwEffectPrivilegeException(e, newval, originalTree, file, line)
        } else {
          throw e
        }
    }
  }
  def cutLongString(s: String) = {
    val newS = s.take(100)
    if(newS.size<s.size)
      newS+"..."
    else newS
  }
  def throwEffectPrivilegeException(e: EffectPrivilegeOutermostBlameException, newval: Set[Effect],  originalTree: String, file: String, line: Int) = {
    val infoP1 =  "\tat "+cutLongString(e.p1.originalTree)+" ("+e.p1.file+":"+e.p1.line+")"
    val infoLocal =  "\tat "+cutLongString(e.blameLocal.originalTree)+" ("+e.blameLocal.file+":"+e.blameLocal.line+")"
    val infoOutermost =  "\tat "+cutLongString(originalTree)+" ("+file+":"+line+")"

    val textOutermost = if(originalTree != e.blameLocal.originalTree){
      "\n\tThe outermost restriction to \""+newval.mkString(",")+"\" did not set the required privileges"+"\n"+infoOutermost
    } else {""}

    val cause = new EffectPrivilegeException(
      "\n\tThe localized restriction to \""+newval.mkString(",")+"\" did not set all the required privileges"+"\n"+infoLocal+ textOutermost
    )
    val exception = new EffectPrivilegeException("Not enough privileges:"+e.missingEffs.mkString(",")+"\n"+infoP1, cause)
    exception.setStackTrace(e.getStackTrace)
    throw exception
  }
  //if set1 contains all elements of set2
  def containsAll[T](set1: Set[T], set2: Set[T]): Boolean = {
    set2.forall(x => set1 contains x)
  }

  def throwRuntimeException(missingEffs: Set[Effect], msg: String = null, cause: Throwable = null) = {
    throw new EffectPrivilegeException("Privileges not found:"+missingEffs+" "+msg)
  }

  def blame(missingEffs: Set[Effect], p1: BlamePosition, msg: String = null, cause: Throwable = null) = {
    throw new EffectPrivilegeLocalBlameException(missingEffs, p1, "Privileges not found:"+missingEffs, cause)
  }


  def lambdalize[T,N](f: => T => N) = {
    //println("checking the cast")
    (x: T) => f(x)
  }
}


