package effectchecker
package example
import scala.tools.nsc.Global

abstract class ExampleDomain extends EffectDomain {
import global._

lazy val lattice = new ExampleLattice {
  val global: ExampleDomain.this.global.type = ExampleDomain.this.global
}
import lattice._

lazy val annotationClasses: List[Symbol] = List(readeClass, writeeClass, alloceClass, emptyRefeClass, testClass, subClass, subwriteeClass)

lazy val barWriteeTrait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$writee"))
lazy val barAlloce1Trait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$alloce1"))
lazy val barAlloce2Trait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$alloce2"))
lazy val barTestTrait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$test"))
lazy val barSubwriteeTrait = rootMirror.getClassByName(newTypeName("scala.annotation.effects.$subwritee"))

def parseAnnotationInfos(annots: List[AnnotationInfo], default: => Effect): Effect = {
 def exceptionsOf(tp: Type): List[Type] = tp match {
    case TypeRef(_, _, args) if args.size > 0 =>
  args.flatMap(exceptionsOf)
    case tp =>
  List(tp)
  }
val anns = annots.filter(annotationClasses contains _.atp.typeSymbol)
if (anns.isEmpty) {
  default
} else {
  ((bottom) /: anns)((eff, annot) => {
val effect =
if(annot.atp.typeSymbol == readeClass){
new ExampleReade()
}
else if(annot.atp.typeSymbol == writeeClass){
val TypeRef(_, _, List(arg)) = annot.atp
val tps = exceptionsOf(arg)
new ExampleWritee(tps)
}
else if(annot.atp.typeSymbol == alloceClass){
val TypeRef(_, _, List(arg1,arg2)) = annot.atp
val tps = (exceptionsOf(arg1),exceptionsOf(arg2))
new ExampleAlloce(tps)
}
else if(annot.atp.typeSymbol == emptyRefeClass){
new ExampleEmptyRefe()
}
else if(annot.atp.typeSymbol == testClass){
val TypeRef(_, _, List(arg)) = annot.atp
val tps = exceptionsOf(arg)
new ExampleTest(tps)
}
else if(annot.atp.typeSymbol == subClass){
new ExampleSub()
}
else{
val TypeRef(_, _, List(arg)) = annot.atp
val tps = exceptionsOf(arg)
new ExampleSubwritee(tps)
}
join(eff, ExampleEffect(List(effect), default.unknown))
})
}
}

def toAnnotation(eff: Effect): List[AnnotationInfo] = {
def toAnnotationInfo(eff: ExampleType): AnnotationInfo = eff match {
case _: ExampleSub => AnnotationInfo(subClass.tpe, Nil, Nil)
case ExampleSubwritee(tps) => AnnotationInfo(typeRef(subwriteeClass.tpe.prefix, subwriteeClass, List(toType(tps, barSubwriteeTrait))), Nil, Nil)
case ExampleAlloce(tps) => AnnotationInfo(typeRef(alloceClass.tpe.prefix, alloceClass, List(toType(tps._1, barAlloce1Trait), toType(tps._2, barAlloce2Trait))), Nil, Nil)
case _: ExampleEmptyRefe => AnnotationInfo(emptyRefeClass.tpe, Nil, Nil)
case ExampleTest(tps) => AnnotationInfo(typeRef(testClass.tpe.prefix, testClass, List(toType(tps, barTestTrait))), Nil, Nil)
case _: ExampleReade => AnnotationInfo(readeClass.tpe, Nil, Nil)
case ExampleWritee(tps) => AnnotationInfo(typeRef(writeeClass.tpe.prefix, writeeClass, List(toType(tps, barWriteeTrait))), Nil, Nil)
}

def toType(eff: List[Type], barTrait: Symbol): Type = eff match {
case Nil => nothingType
case x :: Nil => x
case x :: xs => typeRef(barTrait.tpe.prefix, barTrait, List(x, toType(xs, barTrait)))
}

val l = eff.tps.map{
e => toAnnotationInfo(e)
}

if(eff.unknown)
l ++ List(AnnotationInfo(unknownClass.tpe, Nil, Nil))
else l
}
override def adjust(tree: Tree): Tree = {
super.adjust(tree)
val xi = getEffect(tree)
tree match{
case DefDef(mods, name, tparams, vparamss, tpt, rhs) =>
if("""(foo\.bar)\.read[^.]*$""".r.findAllIn(buildPath(tree)).hasNext){
adjustChildren(tree, xi u ExampleEffect(new ExampleReade()))
}
if("""withAlloc$""".r.findAllIn(buildPath(tree)).hasNext){
val params = tree.symbol.paramss.flatten.map{t => t.tpe}
val newEff = (for{
p1 <- params.lift(0) if p1 <:< definitions.ObjectClass.tpe
p2 <- params.lift(1)
}
yield(ExampleEffect(List(
new ExampleAlloce((List(p1), List(p2)))
)))).getOrElse(bottom)
adjustChildren(tree, xi u newEff)
}
if("""(foo|baz)\.removePrivilege$""".r.findAllIn(buildPath(tree)).hasNext){
val eff = ExampleEffect(revoke(xi.tps, new ExampleTest(List(definitions.NothingClass.tpe))))
adjustChildren(tree, eff)
}
case _ =>
}
tree
}

override def neededEffect(tree: Tree): Option[lattice.Effect] = {
tree match{
case Apply(fun, args) => 
 val effs1: List[Effect] = (if("""(foo\.bar)\.write[^.]*$""".r.findAllIn(buildPath(fun)).hasNext){
   fun match{
case TypeApply(f, as) => 
val genericParams = as.map{_.tpe}
(for{
g1 <- genericParams.lift(0)
}
yield List(ExampleEffect(List(
new ExampleWritee(List(g1)), new ExampleReade()
)))).getOrElse(List())
case _ => List()
}
} else List()) 
 val effs2: List[Effect] = (if("""walloc$""".r.findAllIn(buildPath(fun)).hasNext){
   fun match{
case TypeApply(f, as) => 
val params = fun.symbol.paramss.flatten.map{t => t.tpe}
val genericParams = as.map{_.tpe}
(for{
g1 <- genericParams.lift(0) if g1 <:< definitions.ObjectClass.tpe
p1 <- params.lift(0)
}
yield List(ExampleEffect(List(
new ExampleAlloce((List(g1), List(p1))), new ExampleTest(List(g1))
)))).getOrElse(List())
case _ => List()
}
} else List()) 
 val effs3: List[Effect] = (if("""scala\..*\.print[^.]*$""".r.findAllIn(buildPath(fun)).hasNext){
   fun match{
case TypeApply(f, as) => 
val params = fun.symbol.paramss.flatten.map{t => t.tpe}
(for{
p1 <- params.lift(0) if p1 <:< definitions.ThrowableClass.tpe
}
yield List(ExampleEffect(List(
new ExampleTest(List(p1))
)))).getOrElse(List())
case _ => List()
}
} else List()) 
val effs = effs1 ++ effs2 ++ effs3
if(effs.length>0){
Some((bottom /: effs)((feff, eff) =>
join(eff, feff)))
} else None
case _ => None
}
}
}
abstract class ExampleLattice extends EffectLattice {
val global: Global
import global._
lazy val readeClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.reade"))
lazy val writeeClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.writee"))
lazy val alloceClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.alloce"))
lazy val emptyRefeClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.emptyRefe"))
lazy val testClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.test"))
lazy val subClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.sub"))
lazy val subwriteeClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.subwritee"))

lazy val nothingType = definitions.NothingClass.tpe

type Effect = ExampleEffect
val ctEvidence = implicitly[scala.reflect.ClassTag[Effect]]
abstract class ExampleType{
  def lte(b: ExampleType) = {
b.getClass.isAssignableFrom(this.getClass)
}

def lteCovariant(aTp: Type, bTp: Type): Boolean = {
  aTp <:< bTp
}

def lteContravariant(aTp: Type, bTp: Type): Boolean = {
bTp <:< aTp
}

def lteInvariant(aTp: Type, bTp: Type): Boolean = {
aTp =:= bTp
}

def lteCoContravariant(aTp: Type, bTp: Type): Boolean = {
aTp <:< bTp || bTp <:< aTp
}

}
def lteOneTps[T](aTp: T, bTps: List[T], lteFunc: (T, T) => Boolean) = {
  bTps.exists(bTp => lteFunc(aTp, bTp))
}
class ExampleReade() extends ExampleType{
override def toString(): String = {
return "reade"
}
}

class ExampleWritee(val tps: List[Type]) extends ExampleType{
override def lte(b: ExampleType) = b match {
case bt: ExampleWritee =>
tps.forall(lteOneTps(_, bt.tps, lteCovariant))
case _ => false
}
override def toString(): String = {return "writee[("+(tps.map{tp => tp.toString}.mkString(","))+")]"}
}
object ExampleWritee{ def apply(tps: List[Type]) = new ExampleWritee(tps); def unapply(e: ExampleWritee): Option[List[Type]] = Some(e.tps)}
class ExampleAlloce(val tps: (List[Type], List[Type])) extends ExampleType{
override def lte(b: ExampleType) = b match {
case bt: ExampleAlloce =>
tps._1.forall(lteOneTps(_, bt.tps._1, lteCovariant)) && tps._2.forall(lteOneTps(_, bt.tps._2, lteContravariant))
case _ => false
}
override def toString(): String = {return "alloce[("+(tps._1.map{tp => tp.toString}.mkString(","))+"),("+(tps._2.map{tp => tp.toString}.mkString(","))+")]"}
}
object ExampleAlloce{ def apply(tps: (List[Type], List[Type])) = new ExampleAlloce(tps); def unapply(e: ExampleAlloce): Option[(List[Type], List[Type])] = Some(e.tps)}
class ExampleEmptyRefe() extends ExampleType{
override def lte(b: ExampleType) =  true
override def toString(): String = {
return "emptyRefe"
}
}
class ExampleTest(val tps: List[Type]) extends ExampleType{
override def lte(b: ExampleType) = b match {
case bt: ExampleTest =>
tps.forall(lteOneTps(_, bt.tps, lteContravariant))
case _ => false
}
override def toString(): String = {return "test[("+(tps.map{tp => tp.toString}.mkString(","))+")]"}
}
object ExampleTest{ def apply(tps: List[Type]) = new ExampleTest(tps); def unapply(e: ExampleTest): Option[List[Type]] = Some(e.tps)}
class ExampleSub() extends ExampleType{
override def lte(b: ExampleType) = b match {
case bt: ExampleSub => true
case bt: ExampleReade => true
case _ => false
}
override def toString(): String = {
return "sub"
}
}

class ExampleSubwritee(val tps: List[Type]) extends ExampleType{
override def lte(b: ExampleType) = b match {
case bt: ExampleSubwritee =>
tps.forall(lteOneTps(_, bt.tps, lteContravariant)) 
case bt: ExampleWritee =>
tps.forall(lteOneTps(_, bt.tps, lteCoContravariant)) 

case _ => false
}
override def toString(): String = {return "subwritee[("+(tps.map{tp => tp.toString}.mkString(","))+")]"}
}
object ExampleSubwritee{ def apply(tps: List[Type]) = new ExampleSubwritee(tps); def unapply(e: ExampleSubwritee): Option[List[Type]] = Some(e.tps)}
case class ExampleEffect(tps: List[ExampleType] = Nil, override val unknown: Boolean = false) extends WithUnknownEffect {
def this(tp: ExampleType) = this(List(tp))
override def toString() = {
tps.map{tp => "@"+tp.toString()}.mkString(" ") + (if(unknown) "+?" else "")
}
}

object ExampleEffect {
def apply(tp: ExampleType) = new ExampleEffect(tp)
}
lazy val top: Effect=
ExampleEffect(List(new ExampleReade, new ExampleWritee(List(definitions.ObjectClass.tpe)), new ExampleAlloce((List(definitions.AnyClass.tpe),List(definitions.NothingClass.tpe))), new ExampleTest(List(definitions.ThrowableClass.tpe))))

lazy val bottom: Effect = ExampleEffect(List(new ExampleEmptyRefe))

lazy val unknown: Effect = new ExampleEffect(List(new ExampleEmptyRefe), true)

def default = bottom

def lte(a: Effect, b: Effect): Boolean = {
lteUnknown(a,b) ||
a.tps.forall(lteOne(_, b))
}

def lteOne(aTp: ExampleType, b: Effect): Boolean = aTp match{
case _: ExampleEmptyRefe => true
case _ => b.tps.exists(bTp => aTp.lte(bTp))
}

def lteOneOne(aTp: ExampleType, bTp: ExampleType): Boolean = aTp.lte(bTp)

def elimSub(aTps: List[ExampleType]): List[ExampleType] = aTps match {
case Nil => Nil
case t :: ts =>
val (subtypes, others) = ts.partition(t1 => lteOneOne(t1, t))
val rest = elimSub(others)
if (rest.exists(t1 => lteOneOne(t, t1))) rest
else t :: rest
}

def revoke(aTps: List[ExampleType], tp: ExampleType): List[ExampleType] = {
aTps.filterNot(t => lteOneOne(t,tp))
}

def join(a: Effect, b: Effect): Effect = {
val tps = if(a <= bottom) b.tps
else if(b <= bottom) a.tps
else a.tps ++ b.tps
ExampleEffect(elimSub(tps), joinUnknown(a,b))
}

def meet(a: Effect, b: Effect): Effect = {
val keepA = a.tps.filter(lteOne(_, b))
val keepB = b.tps.filter(lteOne(_, a))
val aNotB = keepA.filterNot(aTp => keepB.exists(bTp => aTp.lte(bTp)))
ExampleEffect(aNotB ++ keepB, meetUnknown(a,b))
}

def diff(a: Effect, b: Effect): Effect = {
val valid = a.tps.filter(!lteOne(_, b))
ExampleEffect(elimSub(valid))
}

def getStrictPrivileges(eff: Effect): Effect = {
ExampleEffect(eff.tps)
}
def getStrictPrivilegesTree(eff: Effect): List[scala.tools.nsc.Global#Tree] = {
eff.tps.flatMap{ tp => tp match{

case _: ExampleReade => List(reify{scala.annotation.effects.reade()}.tree)

case ExampleWritee(tps) =>
val lists = List(
Apply(Ident(reify{List}.tree.symbol), tps.map{tp => Literal(Constant(tp.toString()))})
)
if(eff.tps.size>0)
List(Apply(Select(Ident(reify{scala.annotation.effects.runtimeWritee}.tree.symbol), newTermName("apply")),
List(Apply(Ident(reify{List}.tree.symbol), lists))))
else List[Tree]()

case ExampleAlloce(tps) =>
val lists = List(
Apply(Ident(reify{List}.tree.symbol), tps._1.map{tp => Literal(Constant(tp.toString()))}), Apply(Ident(reify{List}.tree.symbol), tps._2.map{tp => Literal(Constant(tp.toString()))})
)
if(eff.tps.size>0)
List(Apply(Select(Ident(reify{scala.annotation.effects.runtimeAlloce}.tree.symbol), newTermName("apply")),
List(Apply(Ident(reify{List}.tree.symbol), lists))))
else List[Tree]()

case _: ExampleEmptyRefe => List(reify{scala.annotation.effects.emptyRefe()}.tree)

case ExampleTest(tps) =>
val lists = List(
Apply(Ident(reify{List}.tree.symbol), tps.map{tp => Literal(Constant(tp.toString()))})
)
if(eff.tps.size>0)
List(Apply(Select(Ident(reify{scala.annotation.effects.runtimeTest}.tree.symbol), newTermName("apply")),
List(Apply(Ident(reify{List}.tree.symbol), lists))))
else List[Tree]()

case _: ExampleSub => List(reify{scala.annotation.effects.sub()}.tree)

case ExampleSubwritee(tps) =>
val lists = List(
Apply(Ident(reify{List}.tree.symbol), tps.map{tp => Literal(Constant(tp.toString()))})
)
if(eff.tps.size>0)
List(Apply(Select(Ident(reify{scala.annotation.effects.runtimeSubwritee}.tree.symbol), newTermName("apply")),
List(Apply(Ident(reify{List}.tree.symbol), lists))))
else List[Tree]()
}
}.toList
}

type OverrideEffect = NeededEffect
  val ctEvidenceOverrideEffect = implicitly[scala.reflect.ClassTag[OverrideEffect]]
  case class NeededEffect(e: Effect)

  def constructNeededEffect(e: Effect): OverrideEffect = NeededEffect(e)
}
