package effectchecker
package io

import scala.tools.nsc.Global

abstract class IODomain extends EffectDomain {
  import global._

  val lattice: IOLattice = new IOLattice{
    val global: IODomain.this.global.type = IODomain.this.global
  }

  import lattice._

  lazy val ioClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.io"))
  lazy val noIoClass = rootMirror.getClassByName(newTypeName("scala.annotation.effects.noIo"))

  lazy val annotationClasses: List[Symbol] = List(ioClass, noIoClass)

  def parseAnnotationInfos(annots: List[AnnotationInfo], default: => Effect): Effect =
    if      (annots.exists(_.atp.typeSymbol == ioClass)) IOEffect(true)
    else if (annots.exists(_.atp.typeSymbol == noIoClass)) IOEffect(false)
    else default

  override def adjust(tree: Tree): Tree = {
    super.adjust(tree) //adjust children
    val xi = getEffect(tree)

    tree match {
      //case when is an apply print (for example a def that do some print shouldn't be rejected)
      case Apply(fun, args) if isPrint(fun.symbol) => tree
      /*case _: Apply | _: TypeApply =>
        val treeInfo.Applied(fun, _, argss) = tree
        //println("apply, updating", fun)
        //Debugger.identyfySymbol(fun)
        adjust(fun, IOEffect(false))
        for(args <- argss; arg <- args){
          adjust(arg, xi)
        }
        tree
      */
      case _ => tree
    }
  }
  override def check(tree: Tree): List[String] = {
    val xi = getEffect(tree)
    tree match {
      case Apply(fun, args) if isPrint(fun.symbol) =>
        //println(xi)
        if(IOEffect(true) <= xi) errors
        else "Not enought privileges to print here" :: errors
      case _ => super.check(tree)
    }
  }


  private lazy val printNames = List("println", "print", "printIO").map(newTermName(_))
  private lazy val PredefMClass = definitions.PredefModule.moduleClass
  private def isPrint(sym: Symbol) = {
    // observed NoSymbol in some situations (annotation constructor invocations)
    //sym != NoSymbol && sym.owner == PredefMClass && printNames.contains(sym.name)
    sym != NoSymbol && printNames.contains(sym.name)
  }


}

abstract class IOLattice extends EffectLattice {
  val global: Global
  import global._

  type Effect = IOEffect
  val ctEvidence = implicitly[scala.reflect.ClassTag[Effect]]

  case class IOEffect(e: Boolean) extends WithUnknownEffect {
    override def toString() = {
      (if (e) "@io"
      else "@noIo") + (if(unknown) "+¿" else "")
    }
  }

  def join(a: Effect, b: Effect): Effect = IOEffect(a.e || b.e)
  def meet(a: Effect, b: Effect): Effect = IOEffect(a.e && b.e)
  def diff(a: Effect, b: Effect): Effect = if(!b.e) a else bottom
  def lte(a: Effect, b: Effect): Boolean = {
    b.e || !a.e
  }
  def default = IOEffect(true)

  def top: Effect    = IOEffect(true)
  def bottom: Effect = IOEffect(false)
  def unknown: Effect = new IOEffect(false) {override val unknown = true}

  def getStrictPrivileges(eff: Effect): Effect = {
    bottom
  }

  def getStrictPrivilegesTree(eff: Effect): List[Tree] = {
    List[Tree]()
  }

  type OverrideEffect = NeededEffect
  val ctEvidenceOverrideEffect = implicitly[scala.reflect.ClassTag[OverrideEffect]]
  case class NeededEffect(e: Effect)

  def constructNeededEffect(e: Effect): OverrideEffect = NeededEffect(e)
}