package util

import java.sql.Timestamp
import org.joda.time.{ DateTime, DateTimeZone, Period }
import scala.util.matching._
import scala.language.implicitConversions
import scala.annotation.tailrec


object DateUtil{


  def dateRange(from: DateTime, to: DateTime, step: Period): Iterator[DateTime] =Iterator.iterate(from)(_.plus(step)).takeWhile(!_.isAfter(to))
  implicit def dateTimeOrdering: Ordering[DateTime] = Ordering.fromLessThan(_ isBefore _)
  def utcNow = new DateTime(DateTimeZone.UTC)
  implicit def timestamp2DateTime(sqlTimeStamp: Timestamp) = new DateTime(sqlTimeStamp, DateTimeZone.UTC)
  implicit def dateTime2Timestamp(datetime: DateTime) = new Timestamp(datetime.getMillis)

  implicit def optionDateTime2OptionTimestamp(datetime: Option[DateTime]) = datetime.map{d => new Timestamp(d.getMillis)}


  //TODO: arregarlo
  implicit def date2Timestamp(date: java.util.Date) = new Timestamp(date.getTime())

  implicit def timestamp2Date(sqlTimeStamp: Timestamp) = {
    new java.util.Date(sqlTimeStamp.getMillis())
  }

  implicit def optionDate2OptionTimestamp(odate: Option[java.util.Date]) = odate.flatMap{ d => Option(date2Timestamp(d))}
  implicit def optionTimestamp2OptionDate(otimestamp: Option[java.util.Date]) = otimestamp.flatMap{ d => Option(timestamp2Date(d))}


  val day: Long = 86400000
  val hourRe = """(\d{1,2})\:?(\d{1,2})""".r


  def hourStringToLong(hour: String): Long = {
    val hourRe(hours,minutes) = hour
    (hours.toInt*3600 + minutes.toInt*60)*1000
  }

  def hourToDurationBuilder(hour:String) = {
    val hourRe(hours,minutes) = hour
    new Period().withHours(hours.toInt).plus(new Period().withMinutes(minutes.toInt))
  }

  /**
   * completes a data using a reference hour (00:00) and a date, the idea
   * is to take a certain date and assume that the hour corresponds to that date
   * @param hour time in format HH:mm
   * @param date a DateTime for reference
   * @return the "guessed" DateTime corresponding to the given hour
   */
  def guess(hour: String, date: DateTime = utcNow): DateTime = {

    val d1: DateTime = date.withTimeAtStartOfDay().plus(hourToDurationBuilder(hour))

    val oneDay = new Period().withDays(1)
    val candidates = List(d1,d1.plus(oneDay), d1.minus(oneDay))
    val min = (candidates map (d=>Math.abs(d.getMillis()-date.getMillis()))).zipWithIndex.min._2
    candidates(min)
  }

  /**
   * completes a data using a reference hour (00:00) and a date, the idea
   * is to take a certain date and assume that the hour corresponds to that date
   * @param ddhhmm  date in format DDHH
   * @param date a DateTime for reference
   * @return the "guessed" DateTime corresponding to the given day and hour
   */
  def guessMonth(ddhhmm: String, date: DateTime = utcNow): Option[DateTime] = {
    try{
      //debug("gussing for",ddhhmm)
      val d1 =
      if(ddhhmm.size==4)
        date.withDayOfMonth(ddhhmm.substring(0,2).toInt).withHourOfDay(ddhhmm.substring(2,4).toInt).withMinuteOfHour(0).withSecondOfMinute(0).withMillisOfSecond(0)
      else if(ddhhmm.size==6)
        date.withDayOfMonth(ddhhmm.substring(0,2).toInt).withHourOfDay(ddhhmm.substring(2,4).toInt).withMinuteOfHour(ddhhmm.substring(4,6).toInt).withSecondOfMinute(0).withMillisOfSecond(0)
      else
        throw new Exception("No se pudo parsear ddhhmm")
      val candidates = List(d1.minusMonths(1), d1,d1.plusMonths(1))
      val min = (candidates map (d=>Math.abs(d.getMillis()-date.getMillis()))).zipWithIndex.min._2
      Some(candidates(min))
    } catch {
      case a: Throwable =>
        None
    }
  }


  def guessAta(hour: String, dd: DateTime, date: DateTime = utcNow): DateTime = {
    guess(hour, date)

    //iterateUntilBiggerRec(gd, dd)
  }

  @tailrec
  private def iterateUntilBiggerRec(wantedBigger: DateTime, reference: DateTime, days: Int = 0): DateTime = {
    if(days > 2) //to many days, should never enters
      throw new Exception("For some reason one date is much much smaller that the other!!")
    if(wantedBigger.minusMinutes(10).compareTo(reference) < 0) /* 10 minutes to the past in case the clocks are desyncronized */
      iterateUntilBiggerRec(wantedBigger.plus(day), reference, days+1)
    else
      wantedBigger
  }
}

